//	file: sipApplication.cc

#include "sipApplication.h"


Define_Module(sipApplication);



/*
	initializes variables and set own address
*/
void sipApplication::initialize() {

	debug = par("debug").boolValue();
	
	// set address of host
	_localAddr = new sipAddress();
	_localAddr->setIp( "127.0.0.1" );
	_localAddr->setPort( (int)par("localPort") );
	_localAddr->setUser( opp_strdup(par("localUser").stringValue()) );
	_localUsers.push_back( opp_strdup(par("localUser").stringValue()) );

	_nodename = opp_strdup( par("nodeName").stringValue() );

	// Proxy to use
	_proxyString = opp_strdup( par("useProxy").stringValue() );
	if(!opp_strcmp(_proxyString, NULL)) {
		_useProxy = false;
		_proxy = IN_Addr();
	}
	else {
		_useProxy = true;
		_proxy = IN_Addr( _proxyString );
	}

	_isServer = par("aktAsServer").boolValue();
	// schedule INITIALIZE message
	scheduleAt(simTime(), new cMessage("INITIALIZE"));

	// Load commandfile
	_readCmdFile();
	_branchCount = 0;
	_counterChar = 'a';
};



/*
	handles received messages and call specific method for specific gate:
		- gate fromServerModule - method _handleMsgFromServerModule()
		- gate fromClientModule - method _handleMsgFromClientModule()
		- gate fromDialog - method _handleMsgFromDialog()
		- gate fromRTP - delete RTP module
	if a self message is received call _handleSelfMessage()
*/
void sipApplication::handleMessage(cMessage *msg) {

	if(msg->isSelfMessage()) {						// self message
		_handleSelfMessage( msg );
	}	
	
	else if(msg->arrivalGateId() == findGate("fromServerModule")) {		
		// message from server module
		_handleMsgFromServerModule( msg );
	}
	
	else if(msg->arrivalGateId() == findGate("fromClientModule")) {		
		// message from client module
		_handleMsgFromClientModule( msg );
	}
	
	else if(msg->arrivalGateId() == findGate("fromDialog")) {		
		// message from a dialog
		_handleMsgFromDialog( msg );
	}
	
	else if( msg->arrivalGateId() == findGate("fromRTP") ) {		
		// message from RTP module
		// at the moment only "close"-requests are accepted
		cModule *mod = simulation.module( msg->senderModuleId() );
		mod->deleteModule();		
	}

	delete msg;
	msg = NULL;
};



/*
	cleaning up: delete address and user variables
*/
void sipApplication::finish() {

	while(! _localUsers.empty()) {
		delete _localUsers.front();
		_localUsers.erase( _localUsers.begin() );
	}


	delete _localAddr;	_localAddr = NULL;
	delete _nodename;	_nodename = NULL;
};



/*
	handles initializing stuff like open port, reading routing table
*/
void sipApplication::_runtimeInitialization() {

	// send openPort() message to server module to open a port to socket layer
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	sipIntPckt->openPort( IN_Port(5060) );
	send(sipIntPckt, "toServerModule");
			
	// Determine the local ip-address(es) from RoutingTable-Module
	RoutingTable *rt = (RoutingTable *)parentModule()
->submodule("networkLayers")
->submodule("routingTable");
	if ( rt == NULL ) {
		opp_error("could not find the RoutingTable!");
		return;
	}
	_localAddr->setIp( 
		opp_strdup(rt->getInterfaceByIndex(0)->inetAddr->getString()) );		
	if ( debug ) ev << "ip is set to: " << _localAddr->getIp() << endl;
};



/*
	handles received self messages
	check message (class) name:
		- (class name:) sipRequest: a command() message is received 
			(read from the command file)
			call _handleCommandMessages()
		- (message name:) )INITIALIZE: call _runtimeIntialization() to open port 
			and read routing table
*/
void sipApplication::_handleSelfMessage(cMessage *msg) {

	if( 0 == opp_strcmp(msg->className(), "sipRequest") ) {
		sipRequest *request = (sipRequest *)msg;

		if( 0 == opp_strcmp(request->name(), "command()") ) {
			request->setName( sipRequestMethodStr[ request->getMethod()] );
			_handleCommandMessages( msg );
		}		
	}
	
	else if( 0 == opp_strcmp(msg->name(), "INITIALIZE") ) {
		_runtimeInitialization();
	}

};



/*
	handles received command() messages (read from command file)
	check method of request:
		- case INVITE: create new client dialog which sends an INVITE request
		- case BYE: send related dialog a message (sipInnerPacket) which tells 
			the dialog to send a BYE request
		- case CANCEL: send related dialog a message (sipInnerPacket) which 
			tells the dialog to send a CANCEL request
		- case REGISTER: send a REGISTER request to the used proxy 
*/
void sipApplication::_handleCommandMessages(cMessage *msg) {

	// needed variables
	sipRequest *req = (sipRequest *)msg;
	sipInnerPacket *sip = NULL;
	cModule *mod = NULL;

	
	// check method of request
	switch( req->getMethod() ) {
	
		case INVITE:

			// method is INVITE
			// create new dialog (acting as uac)
			createClientDialog( req );
			
			break;

		case BYE:

			// method is BYE
			// find related dialog and send a message to it which tells the dialog
			// to send a BYE request
			mod = _matchDialog(this, "remoteIp", req->header()->getTo().getIp());
			if(  mod != NULL ) {
				sip = new sipInnerPacket( sipInnerPacket::MT_BYE );
				sendDirect( sip, 0.0, mod, "fromApp" );
			}
			else {
				cout << "app: no subdialog for BYE-command found!" << endl;
			}

			break;

		case CANCEL:

			// method is CANCEL
			// find related dialog and send a message to it which tells the dialog
			// to send a CANCEL request
			mod = _matchDialog(this, "remoteIp", req->header()->getTo().getIp());
			if( mod != NULL ) {
				sip = new sipInnerPacket( sipInnerPacket::MT_CANCEL );
				sendDirect( sip, 0.0, mod, "fromApp" );
			}
			else {
				cout << "app: no subdialog for CANCEL-command found!" << endl;
			}

			break;
		  
		case REGISTER:

			// method is REGISTER
			// the registration has only to be sent to the proxy if 
			// the To: - IP-address does not equal the local ip-
			// address
			if( 0 != opp_strcmp(_localAddr->getIp(), 
			                    req->header()->getTo().getIp()) )
			{
				// add local Address
				req = req->dup();
				req->header()->addContact( _localAddr->dup() );
				if( debug ) ev << "requesting registration for " 
					<< req->header()->getTo().getUser() << "@"
					<< req->header()->getTo().getIp() << " to "
					<< req->header()->getContact(0)->getIp() 
					<< endl;
                	        
				_sendRequest( req, "REGISTER Request");
			}
			else {
				_localUsers.push_back( 
					opp_strdup(req->header()->getTo().getUser()) );			
				ev << req->header()->getTo().toChar() 
				   << " locally registered." << endl;
			}
		
			break;	

		default:
			break;
	}
};



/*
	handles received messages from the server module (requests)
*/
void sipApplication::_handleMsgFromServerModule(cMessage *msg) {

	// needed variables
	sipRequest *request = NULL;
	sipResponse *response = NULL;
	// decapsulate sipInterfacePacket from message
	sipInterfacePacket *sipIntPckt = (sipInterfacePacket *)msg;

	
	// check action of the sipInterfacePacket
	switch(sipIntPckt->action()) {
	
		case sipInterfacePacket::SA_TIMEOUT_RESPONSE: 
		
			// timeout occured
			// send timeout message to dialog
			response = sipIntPckt->getResponse();
			handleServerTimeout( response );
			delete response;
			break;
		
		case sipInterfacePacket::SA_REQUEST: 	// request received
		case sipInterfacePacket::SA_ACK:	// ACK received
		
			// if user is found call handleRequest() else 
			// send error message (404 Not found) to sender
			
			// decapsulate request from sipInterfacePacket
			request = sipIntPckt->getRequest();
			
			// had to be changed - for BYE requests, we give no username
			if( request->header()->getTo().getUser() == NULL || checkUser( request) )
				handleRequest( request );
			else {
				if( debug) ev << "addressed user of request not found." << endl;
				sendProvFinResponse( 404, "Not Found", false, request);
			}		
		
			delete request;		
			break;
		
		default:
			break;
	}	
};



/*
	handles received messages from the client module (responses)
*/
void sipApplication::_handleMsgFromClientModule(cMessage *msg) {

	// needed variables
	sipRequest *request = NULL;
	sipResponse *response = NULL;
	// decapsulate sipInterfacePacket from message
	sipInterfacePacket *sipIntPckt = (sipInterfacePacket *)msg;
	
	
	// check action of received sipInterfacePacket
	switch( sipIntPckt->action() ) {
	
		case sipInterfacePacket::SA_TIMEOUT_REQUEST: 

			// timeout occured
			// send timeout message to dialog
			request = sipIntPckt->getRequest();
			handleClientTimeout( request );
			delete request;
			break;

		case sipInterfacePacket::SA_RESPONSE: 
		
			// response received
			// decapsulate response and call handleResponse()
			response = sipIntPckt->getResponse();
			handleResponse( response );
			delete response;
			break;
		
		default:
			break;
	}
};




/*
	handles received messages from dialogs (sipInterfacePacket)
*/
void sipApplication::_handleMsgFromDialog(cMessage *msg) {

	// decapsulate received sipInterfacePacket from message
	sipInterfacePacket *sipIntPckt = (sipInterfacePacket *)msg;


	// check action of received sipInterfacePacket
	switch(sipIntPckt->action()) {
	
		case sipInterfacePacket::SA_CLOSE_SESSION: 
		
			// closeSession() request from a dialog received
			// the dialog reached the state TERMINATED and has to be deleted
			
			// find sender module and delete it
			simulation.module(msg->senderModuleId())->deleteModule();
			
			if( debug ) ev << "sipApp " << _isServer << " : dialog deleted" << endl;
			break;

		case sipInterfacePacket::SA_RESPONSE: 
		
			// a dialog generated a response to a received request
			// forward response to server module for transmission
			send(sipIntPckt->dup(), "toServerModule");
			break;

		case sipInterfacePacket::SA_REQUEST: 	// request received
		case sipInterfacePacket::SA_ACK:	// ACK request received
		
			// a dialog generated a request 
			// forward request to client module for transmission
			send(sipIntPckt->dup(), "toClientModule");
			break;
	
		default:
			break;
	}
};



/*
	handles received responses from client module
	- if a response to a REGISTER request is received and the registration 
		did not fail, the user name is stored 
	- if a response is received (except REGISTER) the related dialog has to 
		be found; the response is forwarded to the dialog
	  - if no dialog is found the response is ignored
*/
void sipApplication::handleResponse( sipResponse *response ) {

	// needed variables
	cModule *_mod;
	sipInterfacePacket *sipIntPckt;
	char *id, *localTag, *remoteTag;


	// check method of response
	switch( response->getMethod() ) {
	
		case REGISTER:
		
			// method is REGISTER
			if( 200 <= response->getStatusCode() < 300 ) {
				// registration succeeded
				if( debug ) ev << "registration of user '"
<< response->header()->getTo().getUser() << "' succeeded!" << endl;
				_localUsers.push_back( 
opp_strdup(response->header()->getTo().getUser()) );
			} 
			else {
				// registration failed
				if( debug ) ev << "registration failed!" << endl;
			}
			break;
		
		default:
		
			// method is not REGISTER
			// find related dialog and forward response to it
			
			// set id and localTag from request
			id = response->header()->getCallId().id;
			localTag = response->header()->getFrom().getTag();

			// if host is acting as UAC the dialog id is 'id + localTag'
			// if host is acting as UAS the dialog id is 'id + localTag + remoteTag'
			
			if(! _isServer)		// UAC
				_mod = simulation.module( findSubmodule(opp_concat(id, localTag)) );
			else {			// UAS
				remoteTag = response->header()->getTo().getTag();
				_mod = simulation.module( findSubmodule(
					opp_concat(id, localTag, remoteTag)) );
			}
        		
			if( _mod != NULL ) {
				// related dialog is found - forward response to it
				sipIntPckt = new sipInterfacePacket();
				sipIntPckt->respond( response->dup() );
				sendDirect(sipIntPckt, 0.0, _mod, "fromApp");
			}
			else {	
				// no related dialog is found - ignore response
				ev << "sipApp: no dialog to received msg" << endl;
			}
			
			break;
	}	
};



/*
	handles received request from server module
	
	If no related dialog exists it has to be checked if the request is sent 
	inside a dialog. If yes a error message is sent: 481 Call/Transaction 
	Does Not Exist. The dialog is not re-build. If the request is sent outside 
	a dialog it has to be checked if a new server dialog can be created. If no 
	dialog exists at this host a new server dialog is created. This is because 
	of RTP. Since RTP can not use different Client-/Serverports it does not seem 
	to be good to work with more than one RTP-Session. In this case a 
	'486 Busy Here' error message is sent. Else a new server dialog is created.
	If a dialog exists or a new one is created, the response is forwarded to 
	the dialog. 
*/
void sipApplication::handleRequest( sipRequest *request ) {

	// needed variables
	cModule *_mod;
	sipInterfacePacket *sipIntPckt;
	sipResponse *errorMsg = NULL;
	char *id, *localTag, *remoteTag;
	sipRequest *sendRequest = request->dup();


	// set id, localTag and remoteTag of dialog which has to be found
	id = opp_strdup(request->header()->getCallId().id);
	localTag = opp_strdup(request->header()->getTo().getTag());
	remoteTag = opp_strdup(request->header()->getFrom().getTag());


	// find the dialog module
	if( (request->getMethod() == BYE) && (!_isServer))
		_mod = simulation.module( findSubmodule(opp_concat(id, localTag, NULL)));
	else 
		_mod = simulation.module( findSubmodule(
			opp_concat(id, localTag, remoteTag)) );


	if(_mod == NULL) {
		// no dialog is found
		if(!opp_strcmp(localTag, NULL) && (request->getMethod() == INVITE)) {
			if( _subModCount(this, "sipDialog") != 0 ) {
				// at least one dialog is found at this host 
				// send '486 Busy Here' response
				errorMsg = new sipResponse( 486, "Busy Here" );
				errorMsg->header()->setCallId(request->header()->getCallId());
				errorMsg->header()->setTo( *request->header()->getTo().dup() );
				errorMsg->header()->setFrom( *request->header()->getFrom().dup() );
				errorMsg->header()->setViaList(request->header()->getViaList());
				errorMsg->header()->setCSeq(request->header()->getCSeq());
			
				errorMsg->setDestination( 
					opp_strdup(request->header()->getFrom().getIp()),
request->header()->getFrom().getPort()
				);
			}
			else {
				// no dialog is found - create new dialog and forward request to it
				createServerDialog( request );
				_mod = simulation.module(findSubmodule(
					opp_concat(id, currentTag, remoteTag)));

				if( _mod == NULL ) {
					opp_error("Could not create 'ServerDialog'!");
					return;
				}
			
				sipIntPckt = new sipInterfacePacket();
				sipIntPckt->request( sendRequest );
				sendDirect(sipIntPckt, 0.0, _mod, "fromApp");
			}
		}
		else {
			// dialog crashed - send '481 Call/Transaction Does Not Exist' 
			// error message
			errorMsg = new sipResponse(481, "Call/Transaction Does Not Exist");
			errorMsg->header()->setCallId(request->header()->getCallId());
			errorMsg->header()->setTo( *request->header()->getTo().dup() );
			errorMsg->header()->setFrom( *request->header()->getFrom().dup() );
			errorMsg->header()->setViaList(request->header()->getViaList());
			errorMsg->header()->setCSeq(request->header()->getCSeq());
			
			errorMsg->setDestination( 
				opp_strdup(request->header()->getFrom().getIp()),
request->header()->getFrom().getPort()
			);
		}
                
		if( errorMsg != NULL ) {
			// send error message if one was generated
			sipIntPckt = new sipInterfacePacket();
			sipIntPckt->respond( errorMsg );
			send(sipIntPckt, "toServerModule");
		}
	}

	else {
		// dialog is found - forward request to it	
		sipIntPckt = new sipInterfacePacket();
		sipIntPckt->request( sendRequest );
		sendDirect(sipIntPckt, 0.0, _mod, "fromApp");
	}
};



/*
	handles timeout messages received from server module
	if related dialog is found, forward timeout message to dialog
	else ignore timeout
*/
void sipApplication::handleServerTimeout( sipResponse *response ) {

	// needed variables
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	cModule *_mod;
	char *_id, *_localTag, *_remoteTag;
	
	// set id, local tag and remote tag of related dialog
	_id = opp_strdup( response->header()->getCallId().id);
	_localTag = opp_strdup( response->header()->getTo().getTag());
	_remoteTag = opp_strdup( response->header()->getFrom().getTag());
	
	// find dialog module
	_mod = simulation.module( findSubmodule( 
		opp_concat( _id, _localTag, _remoteTag)));


	// if dialog module is found, forward timeout message to dialog
	// else ignore timeout 
	if( _mod != NULL ) {
		sipIntPckt->timeoutResponse( response->dup());
		sendDirect( sipIntPckt, 0.0, _mod, "fromApp");
	}
	else {
		if( debug) {
			ev << "--------------------------------------------------------" << endl;
			ev << "--- sipApplication::handleServerTimeout" << endl;
			ev << "--- ERROR: could not match timeoutmessage to anything!!!" << endl;
			ev << "--- expected modulename: '" 
				<< opp_concat(_id, _localTag, _remoteTag) << "'" << endl;
			ev << "--------------------------------------------------------" << endl;
		}
	}
};



/*
	handles timeout messages received from the client module
	if related dialog is found, forward timeout message to dialog
	else ignore timeout	
*/
void sipApplication::handleClientTimeout( sipRequest *request ) {

	// needed variables
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	cModule *_mod;
	char *_id, *_localTag, *_remoteTag; 
	
	// set id, local tag and remote tag of related dialog
	_id = opp_strdup( request->header()->getCallId().id);
	_localTag = opp_strdup( request->header()->getFrom().getTag());
	_remoteTag = opp_strdup( request->header()->getTo().getTag());
	
	// find dialog module
	_mod = simulation.module(findSubmodule(
		opp_concat(_id, _localTag, _remoteTag)));
	
	
	// if dialog module is found, forward timeout message to dialog
	// else ignore timeout 
	if( _mod != NULL ) {
		sipIntPckt->timeoutRequest( request->dup() );
		sendDirect( sipIntPckt, 0.0, _mod, "fromApp");
	}
	else {
		cout << "--------------------------------------------------------" << endl;
		cout << "--- sipApplication::handleClientTimeout" << endl;
		cout << "--- ERROR: could not match timeoutmessage to anything!!!" << endl;
		cout << "--- expected modulename: '" 
<< opp_concat(_id, _localTag, _remoteTag) << "'" << endl;
		cout << "--------------------------------------------------------" << endl;
	}
};



/*
	creates a new server dialog
	parameters of a dialog are added as parameter to the module and 
	the values are copied from the request
*/
void sipApplication::createServerDialog( sipRequest *request )  {

	// needed varialbes
	cModuleType *_modType;
	cModule *_mod;
	char *id, *lTag, *rTag;

	// create tag of new dialog
	// tag is part of dialog id
	createTag();
	
	// set id, local tag and remote tag of dialog which has to be created
	// values are copied from the request
	id = opp_strdup(request->header()->getCallId().id);
	lTag = opp_strdup(createdTag);
	rTag = opp_strdup(request->header()->getFrom().getTag());
	currentTag = opp_strdup(lTag);

	// find module type and create module
	_modType = findModuleType("sipDialog");
	_mod = _modType->create(opp_concat(id, lTag, rTag), this);
	
	
	// add necessary parameters to module and set values
	_mod->par("ip")  = _localAddr->getIp();
	_mod->par("port") = _localAddr->getPort();
	_mod->par("callId") = id;
	_mod->par("localTag") = opp_strdup( currentTag );
	_mod->par("remoteIp") = opp_strdup( request->header()->getFrom().getIp() );
	_mod->par("proxy") = opp_strdup( _proxy );
	_mod->par("remoteTag") = rTag;
	_mod->par("debug")  = par("debug");
	_mod->par("aktAsServer") = true;
	_mod->par("user")  = _localAddr->getUser();
	_mod->par("rUser") = opp_strdup( request->header()->getFrom().getUser() );
	_mod->par("nodeName") = par("nodeName");
	// dialog timer
	_mod->par("timer183") = par("timer183");
	_mod->par("timer200") = par("timer200");
	// set RTP parameters
	_mod->par("portNumberRTP").setLongValue( _getRtpPort() );
	_mod->par("profileNameRTP") = par("profileNameRTP");
	_mod->par("bandwidthRTP") = par("bandwidthRTP");
	_mod->par("fileNameRTP") = par("fileNameRTP");
	_mod->par("payloadTypeRTP") = par("payloadTypeRTP");
	_mod->par("startTransmissionDelayRTP") = par("startTransmissionDelayRTP");
	_mod->par("enterSessionDelayRTP") = par("enterSessionDelayRTP");
	_mod->par("leaveSessionDelayRTP") = par("leaveSessionDelayRTP");
	
	
	// call remaining OMNeT++ methods for creating new modules
	_modType->buildInside(_mod);
	_mod->callInitialize();
	_mod->scheduleStart(simTime());

	if( debug ) ev << "sipApp: ServerDialog '" << _mod->name() << "' - no: " 
		<< _mod->id() << " created." << endl;	
};



/*
	creates a new client dialog
	parameters of a dialog are added as parameter to the module and 
	the values are copied from the request or created
*/
void sipApplication::createClientDialog( sipRequest *request ) {

	// needed variables
	cModuleType *_modType;
	cModule *_mod;

	// new id and tag are created
	createId();
	createTag();
	currentId = opp_strdup( createdId );
	currentTag = opp_strdup( createdTag );
	
	// find module type and create new module
	_modType = findModuleType("sipClientDialog");
	_mod = _modType->create(opp_concat(currentId, currentTag, NULL), this);
	
	
	// add dialog parameter as parameters to module and set values 
	// from the request
	_mod->par("ip") = opp_strdup( _localAddr->getIp() );
	_mod->par("port") = _localAddr->getPort();
	_mod->par("callId") = currentId;
	_mod->par("localTag") = currentTag;
	_mod->par("remoteIp") = opp_strdup( request->getTo().getIp() );
	_mod->par("rUser") = opp_strdup( request->getTo().getUser() );
	_mod->par("proxy") = opp_strdup( _proxyString );
	_mod->par("debug") = par("debug");
	_mod->par("aktAsServer") = false;
	_mod->par("user") = opp_strdup( _localAddr->getUser() );
	_mod->par("nodeName") = par("nodeName");
	_mod->par("branch") = _createBranch();
	// dialog timer 
	_mod->par("timer183") = par("timer183");
	_mod->par("timer200") = par("timer200");
	// set RTP parameters
	_mod->par("portNumberRTP").setLongValue( _getRtpPort() );
	_mod->par("profileNameRTP") = par("profileNameRTP");
	_mod->par("bandwidthRTP") = par("bandwidthRTP");
	_mod->par("fileNameRTP") = par("fileNameRTP");
	_mod->par("payloadTypeRTP") = par("payloadTypeRTP");
	_mod->par("startTransmissionDelayRTP") = par("startTransmissionDelayRTP");
	_mod->par("enterSessionDelayRTP") = par("enterSessionDelayRTP");
	_mod->par("leaveSessionDelayRTP")  = par("leaveSessionDelayRTP");


	// call remaining OMNeT++ methods for creating a new module	
	_modType->buildInside(_mod);
	_mod->callInitialize();
	_mod->scheduleStart(simTime());
	
	if( debug ) ev << "sipApp: ClientDialog '" << _mod->name() << "' - no: " << 
		_mod->id() << " created." << endl;
};



/*
	creates a new tag for a dialog (address parameter)
*/
void sipApplication::createTag() {
	int i;
	srand(time(NULL));

	for (i = 0; i < 4; i++) {
		do {
			createdTag[i] = (char) ((63.0 * rand()) / RAND_MAX) + 65;
		} while (! isalpha(createdTag[i]));
	}
	createdTag[4] = _counterChar++;
	createdTag[5] = 0;
};



/*
	creates a new id for a dialog (part of dialog id)
*/
void sipApplication::createId() {
	int i;
	srand(time(NULL));

	for (i = 0; i < 9; i++) {
		do {
			createdId[i] = (char) ((63.0 * rand()) / RAND_MAX) + 65;
		} while (! isalpha(createdId[i]));
	}
	createdId[9] = _counterChar++;
	createdId[10] = 0;
};



/*
	checks if the addressed user exists or is registered at this host
	if yes, it returns true
*/
bool sipApplication::checkUser( sipRequest *request ) {
	for(unsigned int i = 0; i < _localUsers.size(); i++) {
		if( 0 == opp_strcmp(_localUsers[i], 
			request->header()->getTo().getUser()) ) {
			return true;
		}
	}
	return false;
};



/*
	sends a (provisional / final) response to the server module 
	it uses the given status code and reason phrase and the received request
	for which the response was generated
*/
void sipApplication::sendProvFinResponse(int status, char *reason, bool inv, 
	sipRequest *request) {

	// needed variables
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	sipResponse *lastResponseToApp = new sipResponse(status, reason);

	// set destination address
	if( _useProxy = false)
		lastResponseToApp->setDestination(IN_Addr(
			request->header()->getFrom().getIp()), 
request->header()->getFrom().getPort());
	else 
		lastResponseToApp->setDestination( _proxy, 
			request->header()->getFrom().getPort());
		
	// set header fields
	lastResponseToApp->header()->setTo( sipAddress( request->header()->getTo()));
	lastResponseToApp->header()->setFrom( sipAddress( 
request->header()->getFrom()));
	lastResponseToApp->header()->setCallId( opp_strdup( 
request->header()->getCallId().id));
	lastResponseToApp->header()->setMaxForwards(70);
	lastResponseToApp->header()->addContact( _localAddr->dup());
	lastResponseToApp->header()->setViaList(request->header()->getViaList());
	lastResponseToApp->header()->setCSeq(request->getMethod(), 
request->header()->getCSeq().seqNumber);  
        

	// encapsulate response and send it to server module
	sipIntPckt->respond(lastResponseToApp);
	send( sipIntPckt, "toServerModule");
};



/*
	sends the given request and sets the sipInterfacePacket name to name
*/
void sipApplication::_sendRequest(sipRequest *request, char *name) {

	// set destination
	if( _useProxy ) {
		request->setDestination( opp_strdup(_proxy), 5060 );
	} 
	else {
		request->setDestination( opp_strdup(request->getRequestURI()->getIp()), 
5060 );	
	}
	
	// set via
	if( request->header()->getVia() == NULL ) {
		sipTagVia *via = new sipTagVia( Socket::UDP, _localAddr->getIp(), 
_localAddr->getPort() );
		via->setBranch( _createBranch() );
		request->header()->addVia( via );
	}
	
	// send request to client module 
	sipInterfacePacket *sIfPkt = new sipInterfacePacket( name );
	sIfPkt->request( request );
	send( sIfPkt, "toClientModule" );
};



/*
	reads the given command file and schedule command() self message 
	with action which was read
*/
void sipApplication::_readCmdFile() {

	if( debug ) 
		cout << "Reading commandfile '" << par("cmdFileName").stringValue() 
<< "'" << endl;

	// needed variables	
	char c;
	string conv;
	unsigned int t = 0, line = 1;	
	simtime_t time;
	sipRequestMethod method;
	string toUser;
	string toIp;
	string optArg;
	ifstream cmdstream;
	cmdstream.open(par("cmdFileName").stringValue());

	// check if command file exists	
	if (!cmdstream) {
		error("SIP-commandfile %s not found!", par("cmdFileName").stringValue());
	}
	
	// read command file
	while (cmdstream.get(c)) {
		if ( c == '#' ) {
			// leave out comments
			line++;
			while ( cmdstream.get(c) && (c != '\n') ) {}
		}
		else if ( (c == ' ') || (c == '\t') ) {
			while ( cmdstream.get(c) && ((c == ' ') || (c == '\t')) ) {}
		}
		if (c != '\n')
			do { 
				conv += c;
			} while ( cmdstream.get(c) && (c != ' ') && (c != '\t') && (c != '\n') );
		if  ( 0 != conv.compare("") ) {
			//cout << "Conv: " << conv << "\n";
			switch ( t ) {

				case 0:
					time = strToSimtime( conv.c_str() );
					break;

				case 1: 
					for( unsigned int i = 0; i < 
						(sizeof(sipRequestMethodStr)/sizeof(char *)); i++ )
						if( 0 == conv.compare(sipRequestMethodStr[i]) )
							method = (sipRequestMethod)i;
					break;	
				
				case 2:
					toUser = conv;
					break;
				
				case 3:	 
					toIp = conv;
					break;
					
				case 4: 
					optArg = conv;
					break;
				
				default: 
					error ("To many parameters in %s line %d", 
						par("cmdFileName").stringValue(), line);
			}		 
			
			conv = ""; 
			t++;
		}
		
		// create command() self message
		if ( (c == '\n') && (t != 0) ) {
			if( debug ) {
				cout << "Loaded dataset: t=" << time
				     << ",m=" << sipRequestMethodStr[method]
				     << ",u=" << toUser
				     << ",i=" << toIp
				     << ",o=" << optArg
				     << endl;
			}

			sipRequest *r = new sipRequest();			

			switch( method ) {
				/*
				in the "REGISTER" case the given username and ip-address are those
				values which form the address whose registration is to be changed.
				The fifth argument has to be the ip-address of the registrar!
				*/
				case REGISTER:
					if( t < 4 ) error("not enough parameter for REGISTER command!"); 
					
					r->sipRegister( 
sipAddress(optArg.c_str(), 5060),		// registrar
sipAddress(toUser.c_str(), toIp.c_str(), 5060)	// address to map
					);
					break;
				
				/*
				for "INVITE" requests we only need the destination username 
				and (ip-)address.
				*/
				case INVITE:
					r->invite( sipAddress(toUser.c_str(), toIp.c_str(), 5060) );
					break;
	
				/*
				for "BYE" request we only need the ip address of the communication 
				partner
				*/
				case BYE:	
					r->bye( opp_strdup( toIp.c_str()) );
					break;
				
				/*
				for "CANCEL" request we only need the ip address of the communication 
				partner
				*/
				case CANCEL:
					r->cancel( opp_strdup( toIp.c_str()) );
					break;
	
				/*
				not implemented
				*/
				case OPTIONS:
					break;

				// Unknown method -> delete temporary request
				default:
					delete r;
					r = NULL;
					break;
			}
			
			// change Messagename to Command. THIS IS NECESSARY!!!!!!!
			r->setName( "command()" );
			if ( r != NULL ) scheduleAt( simTime() + time, r );

			// Finally reset values
			t = 0; 
			line++;
			toUser = ""; toIp =""; optArg = "";
		}
	}
	cmdstream.close();
};



/*
	creates a branch id
	branch id = SIP_MAGIC_COOKIE + ip + name of host + random char
*/
char *sipApplication::_createBranch() {
	char buf[255];
	sprintf( buf, "%s-%d", _nodename, _branchCount++ );
	return opp_strdup( opp_concat(SIP_MAGIC_COOKIE, _localAddr->getIp(), buf) );
};




/*
	finds related dialog for a given char value
	if the given parameter value is equal to the value of the module parameter
	the module pointer is returned
	else NULL is returned
*/
cModule *sipApplication::_matchDialog(cModule *parent, char *parameter, 
	char *value) {

	// needed variable
	cModule *mod = NULL;
	
	// look for dialog
	// if the given parameter value is equal to the value of the module parameter
	// the module pointer is returned
	// else NULL
	for( cSubModIterator iter(*parent); !iter.end(); iter++) {
		mod = iter();
		
		if( (0 == opp_strcmp(mod->className(), "sipDialog")) &&
			_matchParameter(mod, parameter, value) ) 
		{
			return mod;
		}
		
		else if( 0 == opp_strcmp(mod->className(), "sipClientDialog") ) {
		
			if( _matchParameter(mod, parameter, value) ) {
				return mod;
			}
			else if( ((0 == opp_strcmp(parentModule()->className(), "sipApplication"))
				&& (mod = _matchDialog(mod, parameter, value)) != NULL) ) 
			{
				return mod;
			}
		}
	}

	return NULL;
};


/*
	finds related dialog for a given long value
	if the given parameter value is equal to the value of the module parameter
	the module pointer is returned
	else NULL is returned
*/
cModule *sipApplication::_matchDialog(cModule *parent, char *parameter, 
	long value) {

	// needed variable
	cModule *mod = NULL;

	// look for dialog
	// if the given parameter value is equal to the value of the module parameter
	// the module pointer is returned
	// else NULL
	for( cSubModIterator iter(*parent); !iter.end(); iter++) {
		mod = iter();
		
		if( (0 == opp_strcmp(mod->className(), "sipDialog")) &&
		    _matchParameter(mod, parameter, value) ) 
		{
			return mod;
		}
		
		else if( 0 == opp_strcmp(mod->className(), "sipClientDialog") ) {
		
			if( _matchParameter(mod, parameter, value) ) {
				return mod;
			}
			else if( ((0 == opp_strcmp(parentModule()->className(), "sipApplication"))
				&& (mod = _matchDialog(mod, parameter, value)) != NULL) ) 
			{
				return mod;
			}
		}
	}

	return NULL;
};



/*
	checks if the given (char) value of the parameter is equal to the module 
	parameter value
	if yes, return true
*/
bool sipApplication::_matchParameter(cModule *mod, char *parameter, 
	char *value) {

	if( mod->hasPar( parameter ) &&
	    opp_strcmp( mod->par( parameter ).stringValue(),
value ) == 0 )
	{
		return true;
	}
	
	return false;
};


/*
	checks if the given (long) value of the parameter is equal to the module 
	parameter value
	if yes, return true
*/
bool sipApplication::_matchParameter(cModule *mod, char *parameter, 
	long value) {

	if( mod->hasPar( parameter ) &&
mod->par( parameter ).longValue() == value )
	{
		return true;
	}
	
	return false;
};



/*
	returns the value of the RTP port
*/
int sipApplication::_getRtpPort() {

	int testPort = 5004;
	
	while( _matchDialog(this, "portNumberRTP", testPort) != NULL ) {
		testPort += 2;
	}
	
	return testPort;
};



/*
	creates a char value from a given int parameter
*/
char *sipApplication::_intToChar(int i) {

	char buf[33];
	sprintf( buf, "%d", i );
	return opp_strdup( buf );
};


/*
	counts submodules whose className matches given name
*/
int sipApplication::_subModCount( cModule *parent, char *name ) {

	int count = 0;

	for(cSubModIterator iter(*parent); !iter.end(); iter++) {
		if( 0 == opp_strcmp(iter()->className(), name) ) {
			count++;
		}
	}
	return count;
};


