/*!
	simulation of SIP
	author: Susann Geisler
	last modified from Susann Geisler on 2003-08-19

	\file   sipApplication.h
	\author Susann Geisler
*/

#ifndef SIPAPPLICATION_H
#define SIPAPPLICATION_H

// for gcc3.x
using namespace std;

// C++-stuff (for fileoperations)
#include <fstream>
#include <string>

// OMNeT++-stuff
#include <omnetpp.h>

// IP-Suite (to determine IP-address)
#include "RoutingTable.h"

// SIP-stuff
#include "sipServerModule.h"
#include "sipClientModule.h"
#include "sipDialog.h"
#include "../Messages/sipInterfacePacket.h"
#include "../Messages/sipInnerPacket.h"
#include "../Messages/sipConstants.h"
#include "../Messages/sipResponse.h"
#include "../Messages/sipRequest.h"
#include "../Messages/sipHeader.h"


/*!

	\class  sipApplication sipApplication.h
	sipApplication is the main module of a SIP host. The
	module receives messages from the transaction layer (\sa sipServerModule,
	\sa sipClientModule) and forwards them to the related dialogs. The module
	creates new dialogs (\sa sipDialog, \sa sipClientDialog) if no related
	dialog exists and no error takes place. Messages received from a dialog
	are forwarded to the sipClientModule (request) and sipServerModule
	(response). 
	
	\author Susann Geisler
	\sa sipClientDialog , \sa sipDialog
*/


class sipApplication: public cSimpleModule {

	Module_Class_Members(sipApplication, cSimpleModule, 0);
        
	private:
	
		// parameters
	
		//! address of the host
		sipAddress *_localAddr;
                
		//! indicates which proxy should be used for communication
		IN_Addr _proxy;
		char *_proxyString;

		//! indicates if a proxy should be used for communication
		bool _useProxy;
                
		//! indicates if the application acts as server
		bool _isServer;
                
		//! IDs of last created dialog
		/*! 
			\li currentTag dialog tag parameter
			\li currentId dialog id parameter
		*/
		char *currentTag, *currentId;
	
		/*! \sa currentTag */
			char createdId[11];
	
		/*! \sa currentId */
			char createdTag[5];

                
		//! variable for creating tags
		char _counterChar;
		
                
		//! indicates debug mode
		bool debug;
		
		//! variable for local branch-parameter-creation
		int _branchCount;
		
		//! local nodename
		char *_nodename;
				
		//! stores registered users at this host
		vector<char *> _localUsers;		

	public:
	
		// OMNeT++ methods
	
		//! initializes variables
		void initialize();


		//! handles received messages
		void handleMessage(cMessage *msg);

		
		//! cleaning up
		void finish();


	protected:

		//! manages initializing stuff
		void _runtimeInitialization();


	
		//! handles received self messages
		/*!
			\param msg received message
		*/
		void _handleSelfMessage(cMessage *msg);
				
		
		//! handles received command() messages
		/*!
			check method of request:
			\li INVITE create new client dialog which sends an INVITE request
			\li BYE send related dialog a message (sipInnerPacket) which tells 
				the dialog
				to send a BYE request
			\li CANCEL send related dialog a message (sipInnerPacket) which 
				tells the dialog to send a CANCEL request
			\li REGISTER send a REGISTER request to the used proxy 		
			
			\param msg received message
		*/
		void _handleCommandMessages(cMessage *msg);
		
				
		//! handles received messages from the server module
		/*!
			\param msg received message
		*/		
		void _handleMsgFromServerModule(cMessage *msg);
		
		
		//! handles received messages from the client module
		/*!
			\param msg received message
		*/
		void _handleMsgFromClientModule(cMessage *msg);


		//! handles received messages from dialogs
		/*!
			\param msg received message
		*/
		void _handleMsgFromDialog(cMessage *msg);
		
		
		
		//! handles received responses from transaction layer (clientModule)
		/*!
			If a response to a REGISTER request is received and the registration 
			did not fail, the user name is stored. If a response is received 
			(except REGISTER) the related dialog has to be found. The response 
			is forwarded to the dialog. If no dialog is found the response is 
			ignored.
			\param response received response
		*/
		void handleResponse( sipResponse *response);


		//! handles received requests from transaction layer (serverModule)
		/*!	If no related dialog exists it has to be checked if the request is 
				sent inside a dialog. If yes a error message is sent: 
				481 Call/Transaction Does Not Exist. The dialog is not re-build. 
				If the request is sent outside a dialog it has to be checked if a new 
				server dialog can be created. If no dialog exists at this host a new 
				server dialog is created. This is because of RTP. Since RTP can not 
				use different Client-/Serverports it does not seem to be good to work 
				with more than one RTP-Session. In this case a '486 Busy Here' error 
				message is sent. Else a new server dialog is created. If a dialog 
				exists or a new one is created, the response is forwarded to the dialog.
				\param request received request
		*/
		void handleRequest( sipRequest *request);



		//! handles timeout messages from transaction layer
		/*!
			If related dialog is found, forward timeout message to dialog. 
			Else ignore timeout. 
			\param response response which caused the timeout while transmission
		*/
		void handleServerTimeout( sipResponse *response);


		//! handles timeout messages from transaction layer
		/*!
			If related dialog is found, forward timeout message to dialog. 
			Else ignore timeout. 
			\param request request which caused the timeout while transmission
		*/
		void handleClientTimeout( sipRequest *request);



		//! creates a server dialog of class sipDialog
		/*!   
			All parameters of the dialog are set. The values are copied from the 
			received request.
			\param request request which is used to create a new dialog 
				(parameter values)
                */
		void createServerDialog( sipRequest *request);


		//! creates a client dialog of class sipClientDialog
		/*! 
			All important parameters of the dialog are set except tag from
			communication partner. This tag is NULL (used to guarantee forking).
			\param request request which is used to create a new dialog 
				(parameter values)
		*/
		void createClientDialog(sipRequest *request);


		//! creates an unique tag for the address parameter
		void createTag();


		//! creates an unique ID for dialog (Call-ID parameter)
		void createId();
				
		
		
		//! checks if addressed user exists or is registered at this host
		//! returns true if user exists
		bool checkUser( sipRequest *request);		
		
		
		
		//! creates and sends a prov. or final response with status and reason
		// inv is true, if a cancel msg to INVITE is sent (487 Request Terminated)
		void sendProvFinResponse(int status, char *reason, bool inv, 
			sipRequest *request = NULL);	

		
		//! sends the given request and sets the sipInterfacePacket name to name
		void _sendRequest(sipRequest *request, char *name);	


                
		//! reads the given command file and schedule command() self message with 
		//!	action which was read
		void _readCmdFile();
				


		//! creates an unique branch id
		char *_createBranch();

		//! finds related dialog for a given char value
		/*! If the given parameter value is equal to the value of the module 
				parameter the module pointer is returned. Else NULL is returned. 
		*/
		cModule *_matchDialog(cModule *parent, char *parameter, char *value);
		
		//! finds related dialog for a given long value
		/*! If the given parameter value is equal to the value of the module 
				parameter the module pointer is returned. Else NULL is returned. 
		*/
		cModule *_matchDialog(cModule *parent, char *parameter, long value);
		
		//! checks if the given (char) value of the parameter is equal to the 
		//! module parameter value
		//! if yes, return true
		bool _matchParameter(cModule *mod, char *parameter, char *value);
		
		//! checks if the given (long) value of the parameter is equal to the 
		//!	module parameter value
		//! if yes, return true
		bool _matchParameter(cModule *mod, char *parameter, long value);
		
		//! returns the RTP port
		int _getRtpPort();
		
		//! creates a char value from a given int parameter
		char *_intToChar(int i);
		
		//! counts submodules whose className matches name
		int _subModCount( cModule *parent, char *name );
};


#endif
