//      file: sipClientDialog.cc

#include "sipClientDialog.h"

Define_Module(sipClientDialog);


/*
	initializes finite state machine variable, addresses and
	all other dialog parameters from the parameter of the 
	created  module
*/
void sipClientDialog::initialize() {

	// finite state machine
	 _fsm.setName("_fsm");
	FSM_Goto(_fsm, UNDEF);
        
	// schedule self message to send INVITE request
	scheduleAt(simTime(), new cMessage("INVITE"));

	 // set own address and received parameters
	localUri = new sipAddress();
	remoteUri = new sipAddress();
	remoteTarget = new sipAddress();

	callId = opp_strdup(par("callId").stringValue());
	localTag = opp_strdup(par("localTag").stringValue());
	remoteTag = opp_strdup( NULL);

	// own address (user, ip address, port, tag)                
	localUri->setIp((char *)par("ip").stringValue());
	localUri->setUser((char *)par("user").stringValue());
	localUri->setPort((int)par("port"));
	localUri->setTag(localTag);

	// remote address (user, ip address, port, tag)
	remoteUri->setIp((char *)par("remoteIp").stringValue());
	remoteUri->setUser((char *)par("rUser").stringValue());
	remoteUri->setPort(5060);
	remoteUri->setTag(remoteTag);

	// set proxy variable
	proxyString = opp_strdup((const char *)par("proxy"));
	if(!opp_strcmp(proxyString, NULL)) {
		useProxy = false;
		 proxy = "";
	}
	else {
		useProxy = true;
		proxy = IN_Addr(proxyString);
	}

	// set default values of dialog identifiers
	secure = false;
	localCSeq = 0;
	remoteCSeq = 0;

	inviteRequest = NULL;
	lastRequest = NULL;

	_debug = par("debug").boolValue();
};



/*
	cleaning up: deleting class variables
*/
void sipClientDialog::finish() {
	delete inviteRequest; inviteRequest = NULL;
	delete remoteUri; remoteUri = NULL;
	delete localUri; localUri = NULL;
	delete remoteTarget; remoteTarget = NULL;
	delete lastRequest; lastRequest = NULL;
};



/*
	handles received messages
	check gate on which the message was received and call
		specific method
*/
void sipClientDialog::handleMessage(cMessage *msg) {

	// self message
	if(msg->isSelfMessage())  
		handleSelfMsg(msg);
	
	// message from application
	else if(msg->arrivalGateId() == findGate("fromApp"))
		handleMsgFromApp(msg);

	// message from subdialog
	else if(msg->arrivalGateId() == findGate("fromDialog")) 
		handleMsgFromDialog(msg);
		
	delete msg;
	msg = NULL;
};



/*
	handles received self message: an INVITE request is sent
*/
void sipClientDialog::handleSelfMsg(cMessage *msg) {

	if( 0 == opp_strcmp(msg->name(), "INVITE")) {
		// call method for creating and sending an INVITE request
		sendInvite(); 
	}
};



/*
	handles timeout messages
	a timeout occured while transmission of a request
	sends a timeout message to the dialog which sent the request
*/
void sipClientDialog::handleClientTimeout( sipRequest *request) {
	
	// needed variables
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	cModule *_mod;
	char *_id, *_localTag, *_remoteTag; 

	// set id, local tag and remote tag variables
	_id = opp_strdup( request->header()->getCallId().id);
	_localTag = opp_strdup( request->header()->getFrom().getTag());
	_remoteTag = opp_strdup( request->header()->getTo().getTag());
	
	// find subdialog
	_mod = simulation.module(findSubmodule(
		opp_concat(_id, _localTag, _remoteTag)));
	
        
	if( _mod == NULL) {
		// no subdialogs exist
		// send closeSession() to app
		sipIntPckt->closeSession();
		sendDirect( sipIntPckt, 0.0, parentModule(), "fromDialog");
	}
	else {
		// subdialog exist
		// forward timeout to subdialog
		sipIntPckt->timeoutRequest( request->dup());
		sendDirect( sipIntPckt, 0.0, _mod, "fromApp");
	}
};



/*
	Handles received messages from the application. If a sipInnerPacket was
	received, a sipInnerPacket has to be sent to the subdialog. It tells the
	subdialog to send a BYE or CANCEL request. If a sipInterfacePacket was 
	received, the action is checked and the message is handled. 
*/
void sipClientDialog::handleMsgFromApp(cMessage *msg) {

	// needed variables
	sipInterfacePacket *sipIntPckt;
	sipResponse *response = NULL;
	sipRequest *request = NULL;
	cModule *_mod;
	char *_id, *_lTag, *_rTag;
	sipInnerPacket *sip = NULL;


	if( 0 == opp_strcmp(msg->className(), "sipInnerPacket") ) {
		// a sipInnerPacket was received
		// this was created when a command() message in the 
		// sipApplication was received
		sip = (sipInnerPacket *)msg;
		
		if( sip->msgType() == sipInnerPacket::MT_BYE ) {
			// the client dialog tells the subdialog to send a BYE request
			if( _debug ) ev << "Generating BYE" << endl;
			
			// find subdialog and send BYE-sipInnerPacket
			for( cSubModIterator iter(*this); !iter.end(); iter++) {
				if( !opp_strcmp( iter()->className(), "sipDialog")) {
					if(! opp_strcmp( remoteUri->getIp(), 
((sipDialog *)iter())->par("remoteIp").stringValue())) {
						sip = new sipInnerPacket( sipInnerPacket::MT_BYE );
						sendDirect( sip, 0.0, simulation.module( iter()->id()), "fromApp");
					}
				}
				else 
					ev << "Client Dialog (" << id() << "): no subdialog for 
						BYE found" << endl;
			}		
		}
		else if( sip->msgType() == sipInnerPacket::MT_CANCEL ) {
			// the client dialog tells the subdialog to send a CANCEL request
			if( _debug ) ev << "Generating CANCEL" << endl;
			
			// find subdialog and send CANCEL-sipInnerPacket
			for( cSubModIterator iter(*this); !iter.end(); iter++) {
				if( !opp_strcmp( iter()->className(), "sipDialog")) {
					if(! opp_strcmp( remoteUri->getIp(), 
((sipDialog *)iter())->par("remoteIp").stringValue())) {
						sip = new sipInnerPacket( sipInnerPacket::MT_CANCEL );
						sendDirect( sip, 0.0, simulation.module( iter()->id()), "fromApp");
					}
				}
			}
		}
		else {
			ev << "Unknown InnerPacket-Type!" << endl;
		}
	}

	else if( 0 == opp_strcmp(msg->className(), "sipInterfacePacket") ) {
		// sipInterfacePacket was received
		sipIntPckt = (sipInterfacePacket *)msg;

		// check action of message
		switch(sipIntPckt->action()) {
		
			case sipInterfacePacket::SA_TIMEOUT_REQUEST:
			
				// timeout occured
				request = sipIntPckt->getRequest();
				handleClientTimeout( request);
				break;
				
			case sipInterfacePacket::SA_RESPONSE:
			
				// response was received
				response = sipIntPckt->getResponse();
				handleResponse( response);
				break;
				
			case sipInterfacePacket::SA_REQUEST:
			
				// request was received
				request = sipIntPckt->getRequest();
				sipIntPckt = new sipInterfacePacket();
				_id = opp_strdup(request->header()->getCallId().id);
				_lTag = opp_strdup(request->header()->getTo().getTag());
				_rTag = opp_strdup(request->header()->getFrom().getTag());
				_mod = simulation.module(findSubmodule(opp_concat(_id, _lTag, _rTag)));
				sipIntPckt->request( request->dup());
				sendDirect( sipIntPckt, 0.0, _mod, "fromApp");
				break;
				
			default:
				break;
		}
	}
	
	else {
		ev << "Unknown Message-Type!" << endl;
	}

	delete response; 
	delete request;
};



/*
	handles received messages from a subdialog
	The action of the message is checked. If a closeSession() message is received
	the sender module is deleted. If no more subdialogs exist, the client dialog 
	sends	a closeSession() message to the sipApplication. Then it will be 
	deleted. If a request or response is received, the message is forwarded to 
	sipApplication. If a terminate() message is received, all subdialogs which 
	are in state EARLY have to be deleted. 
*/
void sipClientDialog::handleMsgFromDialog(cMessage *msg) {

	// needed variables
	sipInterfacePacket *sipIntPckt = (sipInterfacePacket *)msg;
	sipInterfacePacket *sendSipIntPckt;
	int no = 0;


	// check action of sipInterfacePacket
	switch(sipIntPckt->action()) {

		case sipInterfacePacket::SA_CLOSE_SESSION:
                
			// closeSession() message was received
			// delete sender module
			// if no other subdialog exists, send closeSession()
			// message to sipApplication
			simulation.module(msg->senderModuleId())->deleteModule();
			for(cSubModIterator iter(*this); !iter.end(); iter++) {
				no++;
			}
			if(no == 0) {
				sendSipIntPckt = new sipInterfacePacket();
				sendSipIntPckt->closeSession();
				sendDirect(sendSipIntPckt, 0.0, parentModule(), "fromDialog");
			}
			break;

		case sipInterfacePacket::SA_REQUEST:    // request was received
		case sipInterfacePacket::SA_ACK:        // ACK request was received
		case sipInterfacePacket::SA_RESPONSE:   // response was received
		
			// forward message to sipApplication
			sendDirect( sipIntPckt->dup(), 0.0, parentModule(), "fromDialog");
			break;

		case sipInterfacePacket::SA_TERMINATE:
			// all subdialogs which are in the state EARLY should be deleted
			break;
			
		default:
			break;
	}
};



/*
	handles received responses from the application
*/
void sipClientDialog::handleResponse( sipResponse *response ) {

	// needed variables
	sipInterfacePacket *sipIntPckt;
	sipRequest *invReq;
	cModule *_mod;
	char *_id, *_lTag, *_rTag;
	int _status;
   
	// set important variables
	_id = opp_strdup(response->header()->getCallId().id);
	_lTag = opp_strdup(response->header()->getFrom().getTag());
	_rTag = opp_strdup(response->header()->getTo().getTag());
	_status = response->getStatusCode();
	_mod = simulation.module(findSubmodule(opp_concat(_id, _lTag, _rTag)));
	
	// responses with status code 100 are ignored
	// dialogs are only created from other responses (because of the remote tag)
	if(_status != 100) {
		if(!generalResponseCheck()) {
			switch(response->getMethod()) {
                        
				case INVITE:
                                
					// if no subdialog exists, a new one is created
					// the response is forwarded to the dialog
					if(_mod == NULL) {
						createDialog( response);
						_mod = simulation.module( findSubmodule( 
opp_concat( _id, _lTag, _rTag)));
						invReq = inviteRequest->dup();
						sipIntPckt = new sipInterfacePacket("INVITE");
						sipIntPckt->request( invReq);
						sendDirect( sipIntPckt, 0.0, _mod, "fromApp");
					}
					remoteTarget = response->header()->getContact(0);
					if( remoteTarget != NULL ) remoteTarget = remoteTarget->dup();
                                        
				case CANCEL:
				case BYE:
                                
					// response to CANCEL or BYE request is forwarded to the subdialog
					if(_mod != NULL) {
						sipIntPckt = new sipInterfacePacket();
						sipIntPckt->respond( response->dup());
						sendDirect( sipIntPckt, 0.0, _mod, "fromApp");
					}
					else 
						ev << "C-Dialog (" << id() << "): no dialog exists for 
							response" << endl;
					break;
                                        
				default:
					break;
			}	
		}
	}
};



/*
	creates and sends INVITE request
*/
void sipClientDialog::sendInvite() {

	sipTagVia *via = new sipTagVia();
	via->setBranch(  opp_strdup(par("branch").stringValue()) );
	via->setIp( opp_strdup(localUri->getIp()) );
	via->setPort( localUri->getPort() );
	
	localCSeq++;
	// set headerfields
	inviteRequest = new sipRequest();
	inviteRequest->invite(*localUri, *remoteUri, opp_strdup(callId), 
localCSeq, via);
	inviteRequest->header()->setMaxForwards(70);
	inviteRequest->header()->addContact( localUri->dup() );
	
	// set RTP port
	inviteRequest->body()->setRtpPort( (int)par("portNumberRTP").longValue() );

	if(useProxy == false)
		inviteRequest->setDestination( opp_strdup(remoteUri->getIp()), 5060);
	else 
		inviteRequest->setDestination( opp_strdup(proxyString), 5060);
        
        
	if(lastRequest != NULL && lastRequest != inviteRequest) delete lastRequest;
		lastRequest = inviteRequest->dup();
        
	// send INVITE request
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();	
	sipIntPckt->request( inviteRequest->dup() );
	sendDirect( sipIntPckt, 0.0, parentModule(), "fromDialog" );
};



/*
        creates new dialog of type sipDialog;
        all parameter values are set 
*/
void sipClientDialog::createDialog( sipResponse *response) {

	// needed variables
	cModuleType *_modType;
	cModule *_mod;
	char *id, *lTag, *rTag;
        
	// set variable values
	id = opp_strdup(response->header()->getCallId().id);
	lTag = opp_strdup(response->header()->getFrom().getTag());
	rTag = opp_strdup(response->header()->getTo().getTag());

	// find module type and create new dialog module
	_modType = findModuleType("sipDialog");
	_mod = _modType->create(opp_concat(id, lTag, rTag), this);
        
	// add dialog parameter and set values
	_mod->par("ip")                         = localUri->getIp();
	_mod->par("port")                       = localUri->getPort();
	_mod->par("callId")                     = id;
	_mod->par("localTag")                   = lTag;
	_mod->par("remoteIp")                   = remoteUri->getIp();
	_mod->par("proxy")                      = proxy;
	_mod->par("remoteTag")                  = rTag;
	_mod->par("debug")                      = par("debug");
	_mod->par("aktAsServer")                = false;
	_mod->par("CSeq")                       = localCSeq;
	_mod->par("user")                       = localUri->getUser();
	_mod->par("rUser")                      = remoteUri->getUser();
	_mod->par("nodeName")                   = par("nodeName");
	// dialog timer
	_mod->par("timer183")                   = par("timer183");
	_mod->par("timer200")                   = par("timer200");
	// set RTP parameters
	_mod->par("portNumberRTP")	        = par("portNumberRTP");
	_mod->par("profileNameRTP")      	= par("profileNameRTP");
	_mod->par("bandwidthRTP")	        = par("bandwidthRTP");
	_mod->par("fileNameRTP")	        = par("fileNameRTP");
	_mod->par("payloadTypeRTP")	        = par("payloadTypeRTP");
	_mod->par("startTransmissionDelayRTP")	= par("startTransmissionDelayRTP");
	_mod->par("enterSessionDelayRTP")	= par("enterSessionDelayRTP");
	_mod->par("leaveSessionDelayRTP")	= par("leaveSessionDelayRTP");
	
	// call remaining OMNeT++ methods to create new module
	_modType->buildInside(_mod);
	_mod->callInitialize();
	_mod->scheduleStart(simTime());
};

/*
	checks if a received response is correct 
	if no error was occured false is returned
	not implemented yet
*/
bool sipClientDialog::generalResponseCheck() {
	return false;
};



/*
	writes the state of the client dialog with the parameter values
*/
void sipClientDialog::writeState() {
	ev << "--------------- DIALOG STATE --------------------------" << endl;
	ev << "state of dialog " << this->name() << " (" << id() << ") of class " 
<< this->className() << endl;
	ev << "created by " << parentModule()->className() << endl;
	ev << "callID: " << callId << endl;
	ev << "localTag: " << localTag << " , remoteTag: " << remoteTag << endl;
	ev << "localUri: " << localUri->getIp() << " , remoteUri: " << 
remoteUri->getIp() << endl;
	ev << "remoteTarget: " << remoteTarget->getIp() << endl;
	ev << "localCSeq: " << localCSeq << " , remoteCSeq: " << remoteCSeq << endl;
	ev << "secure: " << secure << endl;
	ev << "state: " << _fsm.stateName() << endl;
	ev << "--------------------------------------------------------" << endl;
};



/*
	public method to create and send INVITE request
*/
void sipClientDialog::sipInvite() {
	sendInvite();
};



/*
	public method to create and send a REGISTER request
*/
void sipClientDialog::sipRegister() {

};



/*
	returns the original INVITE request
*/
sipRequest *sipClientDialog::getRequest() const {
	return inviteRequest;
};

