/*
	simulation of SIP
	author: Susann Geisler
	last modified from Susann Geisler on 2003-08-19

  \file sipClientDialog.h
	\author Susann Geisler
*/

#ifndef SIPCLIENTDIALOG_H
#define SIPCLIENTDIALOG_H

// OMNeT++ stuff
#include <omnetpp.h>

// C++ stuff
#include <list.h>

// SIP stuff
#include "SocketInterfacePacket.h"
#include "../Messages/sipRequest.h"
#include "../Messages/sipResponse.h"
#include "../Messages/sipInnerPacket.h"
#include "../Messages/sipInterfacePacket.h"
#include "sipDialog.h"


/*
	\class sipClientDialog sipClientDialog.h
	
	sipClientDialog represents a client dialog. This class is necessary to 
	guarantee forking. A dialog on the client side can receive more than one
	response to a sent INVITE	request. For each response a new dialog is 
	created. The (complete) dialog is of class \sa sipDialog. It has one more
	parameter than a client dialog: the tag of the remote	address. It is the
	tag of the server side which is created then a server responds to	a 
	request. \n
	A client dialog has a dialog id. It consits call id, the local tag and 
	the remote tag. The	remote tag is NULL. 
	
	\author Susann Geisler
	\sa sipApplication, \sa sipDialog
*/

class sipClientDialog: public cSimpleModule {

	Module_Class_Members(sipClientDialog, cSimpleModule, 0);

	public:
		
		//! states of client dialog: UNDEF, EARLY, CONFIRMED, TERMINATED, FAILURE
		enum FSM_STATES {
			UNDEF      = 0,
			EARLY      = FSM_Steady(1),
			CONFIRMED  = FSM_Steady(2),
			TERMINATED = FSM_Steady(3),
			FAILURE    = FSM_Steady(4),
		};
                
		
		// OMNeT++ methods
 	          
		//! initializes variables
		void initialize();
		
		//! handles received messages
		void handleMessage(cMessage *msg);
		
		//! cleaning up
		void finish();
                
		
		
		//! starts creating and sending of an INVITE request
		void sipInvite();
		
		//! starts creating and sending of a REGISTER request
		void sipRegister();
		
		//! returns the original sent INVITE request	
		sipRequest *getRequest() const;
		
               
	protected:
	
		// parameters
	
	
		//! variable for finite state machine
		cFSM _fsm;
                
		//! indicates debug mode
		bool _debug;

                
		//! stores INVITE request and last sent request 
		sipRequest *inviteRequest, *lastRequest;
                
		//! parameters which identify dialog uniquely
		/*!
			the dialog parameters are  remote URI (remoteUri),
			local URI (localUri), remote target (remoteTarget),
			call-ID (callId), tag of own address (localTag), tag of
			remote address (remoteTag), local last used sequence number
			(localCSeq), remote last used sequence number (remoteCSeq), 
			secure mode (secure) and route (route) 
		*/
		sipAddress *remoteUri, *localUri, *remoteTarget;
		char *callId, *localTag, *remoteTag; // all together = dialog id
		int localCSeq, remoteCSeq;
		bool secure;
		list<sipTagRoute *> route;

		//! indicates which proxy should be used if a proxy should be used
		IN_Addr proxy; 
		//! see IN_Addr proxy
		char *proxyString;
                
		//! indicates if a proxy should be used
		bool useProxy;
                
                
		// methods


		//! handles received responses from the client module
		/*!
			\param response received response
		*/
		void handleResponse( sipResponse *response);
                
                
		//! handles received self messages
		/*!
			\param msg received message
		*/
		void handleSelfMsg(cMessage *msg);
                
                
		//! handles received messages from sipApplication
		/*!
			Handles received messages from the application. If a sipInnerPacket was
			received, a sipInnerPacket has to be sent to the subdialog. It tells the
			subdialog to send a BYE or CANCEL request. If a sipInterfacePacket was 
			received, the action is checked and the message is handled.                 
				\param msg received message
		*/
		void handleMsgFromApp(cMessage *msg);
                
                
		//! handles received messages from a (complete) dialog
		/*!
			The action of the message is checked. If a closeSession() message is 
			received the sender module is deleted. If no more subdialogs 
			exist, the client dialog sends a closeSession() message to the 
			sipApplication. Then it will be deleted. If a request or response is 
			received, the message is forwarded to sipApplication. If a terminate() 
			message	is received, all subdialogs which are in state EARLY have to 
			be deleted. 
				\param msg received message
		*/
		void handleMsgFromDialog(cMessage *msg);


		//! a timeout occured while transmission of a request
		/*!
			A timeout occured while transmission of a request.
			Send a timeout message to the dialog which sent the request.
				\param request timeout occured for the given request
		*/
		void handleClientTimeout( sipRequest *request);



		//! general checks for responses
		/*! 
			return 0 if no error was detected \n
			not implemented
		*/
		bool generalResponseCheck();
                
                
                
		//! creates and sends an INVITE request
		void sendInvite();
                
                
                
		//! creates new dialog with same parameters
		/*!
			A new dialog is created if a non-100 response to an
			INVITE request was received.
			The new dialog has the same parameter values like this. 
			Added is the parameter remoteTag which value is taken from the
			response. The new dialog is an instance of class sipDialog. 
		*/
		void createDialog( sipResponse *response);
                
                
                
		//! writes state of dialog
		void writeState();

};


#endif
