/*!
	\file	sipClientTransaction.cc
	\author	jAN bURGHARD
	\version 0.6
	
	This module represents a sip-transaction. This includes:
	connecting a socket, handle the transaction till it 
	reaches the "terminated"-state and finally close the socket.
*/
#include "sipClientTransaction.h"


Define_Module(sipClientTransaction);

//
// methods inherited from cSimpleModule
//
/*
inherited from OMNeT++ - initializes variables.
*/
void sipClientTransaction::initialize() {

	_socketFd = Socket::FILEDESC_UNDEF;
	_debug = par("debug");
	_cseq = 0;
	
	_fsm.setName("State Machine");
	FSM_Goto( _fsm, INITIALIZE );

	_T1 = strToSimtime("500ms");
	_T2 = strToSimtime("4s");
	_T4 = strToSimtime("5s");

	// INVITE - Timers
	_timerAstart = _T1;
	_timerBstart = 64 * _T1;
	_timerDstart = strToSimtime("32s");
	
	// non-INVITE - Timers
	_timerEstart = _T1;
	_timerFstart = 64 * _T1;
	_timerKstart = _T4;
	
	_timerA = NULL;
	_timerB = NULL;
	_timerD = NULL;
	_timerE = NULL;
	_timerF = NULL;
	_timerK = NULL;
	
	_sendCount = 0;
	_request = NULL;
	
	setPhase( sipClientTransaction::INITIALIZE );
	
	// do not become owner of any messages...
	takeOwnership( false );
};


/*
inherited from OMNeT++ - handles incoming messages.
*/
void sipClientTransaction::handleMessage(cMessage *msg) {
	if ( _debug ) {
		myEv << endl
			<< "- state: " << _fsm.state() << endl
			<< "- received msg -> " << msg->name() << endl;
	}
	
	if ( msg->isSelfMessage() ) {
		_handleSelfMessage( msg );
	}
	
	else if (msg->arrivalGateId() == findGate("fromApp")) {
		_handleMessageFromApp(msg);
	} 

	else if (msg->arrivalGateId() == findGate("fromSocketLayer")) {
		_handleMessageFromSocketLayer(msg);
	} 

	else {
		myEv << "handleMessage(): Message from unknown Gate." << endl;
//		opp_error("Message from unknown Gate.");
	}

	delete msg;
};






/*
method for changing the state of the fsm
*/
void sipClientTransaction::setPhase(const FSM_STATES s) {
	if (_debug) myEv << "moving from phase " 
		<< _fsm.state() << " to phase " << s << endl;
	
	// this is only cosmetically and the same as "FSM_Goto( _fsm, s);"
	
	switch( s ) {
	case INITIALIZE:
		FSM_Goto( _fsm, INITIALIZE );
		break;
	case INV_CALLING:
		FSM_Goto( _fsm, INV_CALLING );
		break;
	case INV_PROCEEDING:
		FSM_Goto( _fsm, INV_PROCEEDING );
		break;
	case INV_COMPLETE:
		FSM_Goto( _fsm, INV_COMPLETE );
		break;
	case TERMINATED:
		FSM_Goto( _fsm, TERMINATED );
		break;
	case NON_INV_TRYING:
		FSM_Goto( _fsm, NON_INV_TRYING );
		break;
	case NON_INV_PROCEEDING:
		FSM_Goto( _fsm, NON_INV_PROCEEDING );
		break;
	case NON_INV_COMPLETE:
		FSM_Goto( _fsm, NON_INV_COMPLETE );
		break;
	case ACK_ONLY:
		FSM_Goto( _fsm, ACK_ONLY );
		break;
	}
	
	// on entering a new state, we might have something to do
	scheduleAt( simTime(), new cMessage("new state") );
};





/*
handels a message received from the application
*/
void sipClientTransaction::_handleMessageFromApp(cMessage *msg) {
	// we want sipInterfacePackets
	sipInterfacePacket *sifpkt = (sipInterfacePacket *)msg;
	int _action = sifpkt->action();
	
	// needed objects
	SocketInterfacePacket *sockipack;
	
	switch (_action) {
	case sipInterfacePacket::SA_ACK:
		if ( _fsm.state() != INITIALIZE ) {
//			opp_error("unexpected ACK request!");
			myEv << "unexpected ACK request!" << endl;
		} else {
			_request = sifpkt->getRequest();
			setPhase( sipClientTransaction::ACK_ONLY );

			_destination.setIp( _request->getDestIp() );
			_destination.setPort( _request->getDestPort() );

			// connect a socket
			sockipack = new SocketInterfacePacket("SocketCall");
			sockipack->socket(Socket::AF_INET, Socket::SOCK_DGRAM, Socket::UDP);
			send(sockipack, "toSocketLayer");
		}
		break;
		
	case sipInterfacePacket::SA_REQUEST:
		if ( _fsm.state() != INITIALIZE ) {
			opp_error("unexpected SIP request!");
			myEv << "unexpected SIP request!" << endl;
		} else {
			_request = sifpkt->getRequest();
			
			if ( _request->getMethod() == INVITE ) {
				setPhase( sipClientTransaction::INV_CALLING );
			} else {
				setPhase( sipClientTransaction::NON_INV_TRYING );
			}
			_destination.setIp( _request->getDestIp() );
			_destination.setPort( _request->getDestPort() );

			// connect a socket
			sockipack = new SocketInterfacePacket("SocketCall");
			sockipack->socket(Socket::AF_INET, Socket::SOCK_DGRAM, Socket::UDP);
			send(sockipack, "toSocketLayer");
		}
		break;
		
	case sipInterfacePacket::SA_TERMINATE:
		if (_debug) myEv << "terminating (sipClientTransaction)." << endl;
		_closeSession();
		break;
		
	default:
		myEv << " ignoring action " << _action << endl;
		break;
	}
};










/*
handels a message received from the socketlayer
*/
void sipClientTransaction::_handleMessageFromSocketLayer(cMessage *msg) {
	SocketInterfacePacket *sockipack = (SocketInterfacePacket *)msg;
	
	// needed objects
	sipResponse *response;
	
	switch( sockipack->action() ) {
	case SocketInterfacePacket::SA_SOCKET_RET:
		_socketFd = sockipack->filedesc();

		if ( _debug ) {
			myEv << "got socketdescriptor " << _socketFd << endl
				<< "connecting socket to " 
				<< _destination.getIp() << endl;
		}
		
		// now, we can send the request
		sockipack = new SocketInterfacePacket("ConnectCall");
		sockipack->connect(_socketFd, _destination.getIp(), _destination.getPort());
		send(sockipack, "toSocketLayer");
		break;
		
	case SocketInterfacePacket::SA_CONNECT_RET:
		if ( _debug ) myEv << "got socketconnection - sending request." << endl;

		// now having a socket we can send the request
		_sendRequest();

		// ... and in case of ACK_ONLY we've to terminate this transaction
		if ( _fsm.state() == ACK_ONLY ) setPhase( TERMINATED );
		break;
		
	case SocketInterfacePacket::SA_READ_RET:
		if ( _debug ) myEv << "read packet from socketlayer" << endl;
		response = (sipResponse *)sockipack->decapsulate();
		
		switch ( _fsm.state() ) {
		case INV_CALLING:
		case INV_PROCEEDING:
		case INV_COMPLETE:
			_handleInviteTransaction( response );
			break;
			
		case NON_INV_TRYING:
		case NON_INV_PROCEEDING:
		case NON_INV_COMPLETE:
			_handleNonInviteTransaction( response );
			break;
			
		default:
			opp_error("undefined FSM state!");
			break;
		}
		break;
		
	default:
		myEv << " ignoring unknown socket action." << endl;
		break;
	}
	
};




/*
prepares the transaction for being destroyed. The socket will be closed,
timers will be halted and as the stored request deleted.
*/
void sipClientTransaction::_closeSession() {
	// close the socket
	if ( _socketFd != Socket::FILEDESC_UNDEF ) {
		SocketInterfacePacket *sockipack = new SocketInterfacePacket("shutdownCall");
		sockipack->close( _socketFd );
		send( sockipack, "toSocketLayer" );
	}
	
	// deleting the copy of the sipRequest
	delete _request;


	// canceling all timers
	_timerA = _stopTimer( _timerA );
	_timerB = _stopTimer( _timerB );
	_timerD = _stopTimer( _timerD );
	_timerE = _stopTimer( _timerE );
	_timerF = _stopTimer( _timerF );
	_timerK = _stopTimer( _timerK );
	
	
	sipInterfacePacket *sIfPkt = new sipInterfacePacket();
	sIfPkt->closeSession();
	sendDirect(sIfPkt, 0, parentModule(), "fromClientTrans");
};


/*
sets up a new timer
*/
cMessage *sipClientTransaction::_startTimer(char *name, simtime_t delay) {
	if ( delay > 0 ) {
		cMessage *timerX = new cMessage( name );
		scheduleAt( simTime() + delay, timerX );
	
		return timerX;
	} else {
		return NULL;
	}
};


/*
cancels the given timer
*/
cMessage *sipClientTransaction::_stopTimer(cMessage *timer) {
	if ( timer != NULL ) {
		if( timer->isScheduled() ) {
			cancelEvent( timer );
		}
		delete timer;
	}
	return NULL;
};


/*
sends stored request to the socketlayer
*/
void sipClientTransaction::_sendRequest() {
	if ( _socketFd != Socket::FILEDESC_UNDEF ) {
		// copy the request
		sipRequest *copy = new sipRequest( *_request );
//		copy->setOwner( _request->owner() );
		
		SocketInterfacePacket *sockipack = new SocketInterfacePacket("writeCall");
		sockipack->write( _socketFd, copy );
		send(sockipack, "toSocketLayer");
	} else {
		myEv << "could not send request - no socket available" << endl;
	}
	_sendCount++;
};


/*
sends received response to the application (via sipClientModule)
*/
void sipClientTransaction::_sendResponse(sipResponse *r, char *pktName) {

//	r->setOwner( _request->owner() );

	sipInterfacePacket *sIfPkt = new sipInterfacePacket( pktName );
	sIfPkt->respond( r );
	sendDirect( sIfPkt, 0, parentModule(), "fromClientTrans" );
};


/*

*/
void sipClientTransaction::_sendACK( sipResponse *response ) {
	sipRequest *ack = new sipRequest();
	ack->ack( _request, &response->header()->getTo() );
//	ack->setOwner( _request->owner() );
	
	if ( _socketFd != Socket::FILEDESC_UNDEF ) {
		SocketInterfacePacket *sockipack = new SocketInterfacePacket("writeCall");
		sockipack->write( _socketFd, ack );
		send(sockipack, "toSocketLayer");
	} else {
		myEv << "could not send request - no socket available" << endl;
	}
	_sendCount++;
};


/*
handels the response as an invite transaction
*/
void sipClientTransaction::_handleInviteTransaction(sipResponse *r) {
	char *text = _intToChar( r->getStatusCode() );
	
	switch( _fsm.state() ) {
	case INV_CALLING:
		// stop retransmission timers
		_timerA = _stopTimer( _timerA );
		_timerB = _stopTimer( _timerB );
		
		if ( r->getStatusCode() < 200 ) {
			// fwd response
			text = opp_concat( text, " - provisional" );
			_sendResponse( r, text );
			
			// goto PROCEEDING
			setPhase( INV_PROCEEDING );
		}
		
		else if ( r->getStatusCode() < 300 ) {
			text = opp_concat( text, " - final" );
			_sendResponse( r, text );
			
			// goto PROCEEDING
			setPhase( TERMINATED );
		}
		
		else {
			text = opp_concat( text, " - final" );
			_sendResponse( r, text );
			_sendACK( r );
			
			// goto COMPLETE
			setPhase( INV_COMPLETE );
		}
		break;
		
	case INV_PROCEEDING:
		if ( r->getStatusCode() < 200 ) {
			// fwd response
			text = opp_concat( text, " - provisional" );
			_sendResponse( r, text );
		}
		
		else if ( r->getStatusCode() < 300 ) {
			text = opp_concat( text, " - final" );
			_sendResponse( r, text );
			
			// goto PROCEEDING
			setPhase( TERMINATED );
		}
		
		else {
			text = opp_concat( text, " - final" );
			_sendResponse( r, text );
			_sendACK( r );
			
			// goto COMPLETE
			setPhase( INV_COMPLETE );
		}
		break;
		
	case INV_COMPLETE:
		if ( r->getStatusCode() < 200 ) {
			// do nothing
		}
		
		else if ( r->getStatusCode() < 300 ) {
			text = opp_concat( text, " - final" );
			_sendResponse( r, text );
			
			// goto TERMINATED
			setPhase( TERMINATED );
		}
		
		else {
			_sendACK( r );
		}
		break;
		
	default:
		opp_error("unknown FSM state");
		break;
	}
};


/*
handels the response as a non-invite transaction
*/
void sipClientTransaction::_handleNonInviteTransaction(sipResponse *r) {
	char *text = _intToChar( r->getStatusCode() );
	
	switch( _fsm.state() ) {
	case NON_INV_TRYING:
		if ( r->getStatusCode() < 200 ) {
			text = opp_concat( text, " - provisional" );
			_sendResponse( r, text );
			
			// goto PROCEEDING
			setPhase( NON_INV_PROCEEDING );
		}
		
		else {
			text = opp_concat( text, " - final" );
			_sendResponse( r, text );

			// goto COMPLETE
			setPhase( NON_INV_COMPLETE );
		}
		break;
		
	case NON_INV_PROCEEDING:
		if ( r->getStatusCode() < 200 ) {
			text = opp_concat( text, " - provisional" );
			_sendResponse( r, text );
		}
		
		else {
			text = opp_concat( text, " - final" );
			_sendResponse( r, text );

			// goto COMPLETE
			setPhase( NON_INV_COMPLETE );
		}
		break;
		
	case NON_INV_COMPLETE:
		// do nothing
		break;
		
	default:
		opp_error("transition to unexpected FSM state!");
		break;
	}
};


/*
executes for state-changes and firing timers
*/
void sipClientTransaction::_handleSelfMessage(cMessage *msg) {

	sipInterfacePacket *sIfPkt;

	if ( 0 == opp_strcmp(msg->name(), "new state") ) {
		// we just entered a new state
		switch ( _fsm.state() ) {
		case INV_CALLING:
			// start timer A
			_timerA = _startTimer( "timer A", _timerAstart );
			// start timer B
			_timerB = _startTimer( "timer B", _timerBstart );
			break;
			
	     	case INV_PROCEEDING:
			break;
					
		case INV_COMPLETE:
			// start timer D
			_timerD = _startTimer( "timer D", _timerDstart );
			break;
			
		case TERMINATED:
			_closeSession();
			break;
		
		case NON_INV_TRYING:
			// start timer E
			_timerE = _startTimer( "timer E", _timerEstart );
			// start timer F
			_timerF = _startTimer( "timer F", _timerFstart );
			break;
		
		case NON_INV_PROCEEDING:
			break;
		
		case NON_INV_COMPLETE:
			// stop timers E and F
			_timerE = _stopTimer( _timerE );
			_timerF = _stopTimer( _timerF );
			// start timer K			
			_timerK = _startTimer( "timer K", _timerKstart );
			break;
			
		case ACK_ONLY:
			break;
		
		default:
			// do nothing
			// opp_error("unexpected new STATE!");
			break;
		}
	} 
	
	else if ( 0 == opp_strcmp(msg->name(), "timer A") ) {
		if ( _fsm.state() == INV_CALLING ) {
			if (_debug) myEv << msg->name() << " fired" << endl;
			// timer A fires
			// resend INVITE
			_sendRequest();
			
			// restart timer A
			_timerAstart *= 2;
			_timerA = _startTimer( "timer A", _timerAstart );
		} else {
			_timerA = NULL;	
		}
	}
	
	else if ( 0 == opp_strcmp(msg->name(), "timer B") ) {
		if ( _fsm.state() == INV_CALLING ) {
			if (_debug) myEv << msg->name() << " fired" << endl;
			// timer B fires
			// inform TU (Application)
			sIfPkt = new sipInterfacePacket("TIMEOUT");
			sIfPkt->timeoutRequest( new sipRequest(*_request) );
			
			// via sipClientModule to Application
			sendDirect( sIfPkt, 0, parentModule(), "fromClientTrans");
			
			// switch to TERMINATED
			setPhase( sipClientTransaction::TERMINATED );
		}
		_timerB = NULL;
	}
	
	else if ( 0 == opp_strcmp(msg->name(), "timer D") ) {
		if ( _fsm.state() == INV_COMPLETE ) {
			if (_debug) myEv << msg->name() << " fired" << endl;
			// timer D fires
			// move to TERMINATED
			setPhase( sipClientTransaction::TERMINATED );
		}
		_timerD = NULL;
	}
	
	
	
	else if ( 0 == opp_strcmp(msg->name(), "timer E") ) {
		if ( _fsm.state() == NON_INV_TRYING || _fsm.state() == NON_INV_PROCEEDING ) {
			if (_debug) myEv << msg->name() << " fired" << endl;
			// timer E fires
			// resend INVITE
			_sendRequest();
			
			// restart timer E
			_timerEstart = min( _T2, _sendCount * 2 * _T1);
			_timerE = _startTimer( "timer E", _timerEstart );
		} else {
			_timerE = NULL;	
		}
	}
	
	else if ( 0 == opp_strcmp(msg->name(), "timer F") ) {
		if ( _fsm.state() == NON_INV_TRYING || _fsm.state() == NON_INV_PROCEEDING ) {
			if (_debug) myEv << msg->name() << " fired" << endl;
			// timer F fires
			// inform TU (Application)
			sIfPkt = new sipInterfacePacket("TIMEOUT");
			sIfPkt->timeoutRequest( new sipRequest(*_request) );
			
			// via sipClientModule to Application
			sendDirect( sIfPkt, 0, parentModule(), "fromClientTrans");
			
			// switch to TERMINATED
			setPhase( sipClientTransaction::TERMINATED );
		}
		_timerF = NULL;
	}
	
	else if ( 0 == opp_strcmp(msg->name(), "timer K") ) {
		if ( _fsm.state() == NON_INV_COMPLETE ) {
			if (_debug) myEv << msg->name() << " fired" << endl;
			// timer K fires
			// move to TERMINATED
			setPhase( sipClientTransaction::TERMINATED );
		}
		_timerK = NULL;
	}
};


/*
convert int-values to a char *-string
*/
char *sipClientTransaction::_intToChar(int i) {
	char buf[33];
	sprintf( buf, "%d", i );
	return opp_strdup( buf );
};

