/*!
	\file	sipClientTransaction.h
	\author	jAN bURGHARD
	\version 0.6
	
*/
#ifndef __SIPCLIENTTRANSACTION_H
#define __SIPCLIENTTRANSACTION_H

// for gcc3.x
using namespace std;

// OMNeT++
#include <omnetpp.h>

// IP-Suite
#include "sockets.h"
#include "SocketInterfacePacket.h"

// SIP-Stuff
#include "../Messages/sipConstants.h"
#include "../Messages/sipRequest.h"
#include "../Messages/sipResponse.h"
#include "../Messages/sipInterfacePacket.h"


//! SIP client-transaction
/*!
	\class sipClientTransaction sipClientTransaction.h
	\author Jan Burghard
	
	representation of client-transactions as defined in RFC 3261.
	Basically this class waits for a sipRequest from a sipApplication.
	If one is received it connects a socket to the request's destination
	address, sends the request over it and waits for responses.
*/
class sipClientTransaction : public cSimpleModule {

public:
	/*!
	enumeration of the finite state machine
	
	We have 9 states: 1 to 4 for INVITE requests,
	5, 6, 7 and 4 for non-INVITE requests and
	8 for ACK-only sending
	*/
	enum FSM_STATES {
	     INITIALIZE		= 0,
	     INV_CALLING	= FSM_Steady(1),
	     INV_PROCEEDING	= FSM_Steady(2),
	     INV_COMPLETE	= FSM_Steady(3),
	     TERMINATED		= FSM_Steady(4),	     
	     NON_INV_TRYING	= FSM_Steady(5),
	     NON_INV_PROCEEDING	= FSM_Steady(6),
	     NON_INV_COMPLETE	= FSM_Steady(7),
	     ACK_ONLY		= FSM_Steady(9)
	};
	
	//! OMNeT++ macro
	/*!
	will create constructors and OMNeT++ specific stuff.
	*/
	Module_Class_Members(sipClientTransaction, cSimpleModule, 0);
	
	//! inherited from OMNeT++
	/*!
	initializes the class.
	*/
	virtual void initialize();
	
	//! inherited from OMNeT++
	/*!
	handles incoming messages.
	*/
	virtual void handleMessage(cMessage *msg);
	

protected:
	//! method for changing the state of the fsm
	/*!
	it mainly schedules the "new state" message.
	\param s the new state out of FSM_STATES
	*/
	virtual void setPhase(const FSM_STATES s);
	

private:	
	//! the request to be sent
	sipRequest *_request;	
	
	//! debugmode
	bool _debug;
	
	//! CSeq number for this transaction
	int _cseq;
	
	//! state of the transaction
	cFSM _fsm;
	
	//! socket filedescriptor
	Socket::Filedesc _socketFd;
	
	//! destination address 
	sipAddress _destination;
	
	//! initial value for transaction timer A
	simtime_t _timerAstart;
	//! initial value for transaction timer B
	simtime_t _timerBstart;
	//! initial value for transaction timer D
	simtime_t _timerDstart;
	//! initial value for transaction timer E
	simtime_t _timerEstart;
	//! initial value for transaction timer F
	simtime_t _timerFstart;
	//! initial value for transaction timer K
	simtime_t _timerKstart;
	
	//! transaction timer A
	cMessage *_timerA;
	//! transaction timer B
	cMessage *_timerB;
	//! transaction timer D
	cMessage *_timerD;
	//! transaction timer E
	cMessage *_timerE;
	//! transaction timer F
	cMessage *_timerF;
	//! transaction timer K
	cMessage *_timerK;
	
	//! timer multiplier T1 (see also RFC 3261)
	simtime_t _T1;
	//! timer multiplier T2 (see also RFC 3261)
	simtime_t _T2;
	//! timer multiplier T4 (see also RFC 3261)
	simtime_t _T4;
	
	//! used for timer calcualtions
	int _sendCount;
	
	//! prepares the transaction for being destroyed. 
	/*!
	The socket will be closed, timers will be halted and as the stored
	request - deleted.
	*/
	void _closeSession();
	
	//! sets up a new timer
	/*!
	this includes creation, scheduling and returning the pointer.
	\param name name of the timer
	\param delay time from now (simTime()) when the timer fires
	\return pointer to the new timer
	*/
	cMessage *_startTimer(char *name, simtime_t delay);
	
	//! deletes the given timer
	/*!
	it also checks if the timer(-message) is scheduled. If it is,
	it also removes it from the FES.
	\param timer pointer to the timer
	\return NULL
	*/
	cMessage *_stopTimer(cMessage *timer);
	
	//! sends stored request to the socketlayer
	/*!
	expects that the socket is already connected. If not it only produces 
	a failure-message to the "ev"-environment.
	*/
	void _sendRequest();
	
	/*!
	sends received response to the application (via sipClientModule)
	\param r the sipResponse
	\param pktName name for the sipInterfacePacket (we don't really need it ;-)
	*/
	void _sendResponse(sipResponse *r, char *pktName);
	
	/*! 
	\sa _sendResponse
	\param response the sipResponse
	*/
	void _sendACK(sipResponse *response);
	
	//! handels a message received from the application
	/*!
	Accepts sipInterfacePackets only and processes them dependent on
	their action()-value. Only SA_ACK, SA_REQUEST and SA_TERMINATE cause
	actions.
	\param msg sipInterfacePacket as cMessage	
	*/
	void _handleMessageFromApp(cMessage *msg);

	//! handels a message received from the socketlayer
	/*!
	this includes the message's cast to a SocketInterfacePacket and it's 
	following processing due to it's action()-value.
	\param msg SocketInterfacePacket as cMessage
	*/
	void _handleMessageFromSocketLayer(cMessage *msg);
	
	//! handels the response as an invite transaction
	/*!
	Inspects the response's status code and causes necessary transitions
	and actions.
	\param r as sipResponse
	*/
	void _handleInviteTransaction(sipResponse *r);

	//! handels the response as a non-invite transaction
	/*!
	Inspects the response's status code and causes necessary transitions
	and actions.
	\param r as sipResponse
	*/
	void _handleNonInviteTransaction(sipResponse *r);
	
	//! handels self-messages such as "timers" and "state changes"
	/*!
	Message-types are recognized by messagename. Accepted messagenames
	are "timer A", "timer B", "timer D", "timer E", "timer F", "timer K"
	ans "new state". Messages with other namestrings will cause no 
	action.
	\param msg message as cMessage
	*/
	void _handleSelfMessage(cMessage *msg);
	
	//! convert int-values to a char *-string
	/*!
	Uses sprintf and allocates memory. Has to be deleted later.
	*/
	char *_intToChar(int i);
};

#endif
