//	file: sipDialog.cc

#include "sipDialog.h"

Define_Module(sipDialog);

/*
	initializes class variables
*/
void sipDialog::initialize() {

	// finite state machine variable
	_fsm.setName("_fsm");
	FSM_Goto(_fsm, UNDEF);
        
	debug = par("debug");
        
	// set own address and received parameters
	localUri = new sipAddress();
	remoteUri = new sipAddress();
	remoteTarget = new sipAddress();

	callId = opp_strdup((const char *)par("callId").stringValue());
	localTag = opp_strdup((const char *)par("localTag").stringValue());
	remoteTag = opp_strdup((const char *)par("remoteTag").stringValue());
        
	// set own address (user, ip address, port, tag)
	localUri->setIp((char *)par("ip").stringValue());
	localUri->setPort((int)par("port"));
	localUri->setTag(opp_strdup( localTag));
	localUri->setUser((char *)par("user").stringValue());
        
	// set remote address (user, ip address, port, tag)
	remoteUri->setIp((char *)par("remoteIp").stringValue());
	remoteUri->setPort(5060);
	remoteUri->setTag(remoteTag);
	remoteUri->setUser((char *)par("rUser").stringValue());

	remoteTarget = remoteUri->dup();

	// set proxy variables
	proxyString = opp_strdup((const char *)par("proxy"));
	if(!opp_strcmp(proxyString, NULL)) {
		useProxy = false;
		proxy = IN_Addr();
	}
	else {
		useProxy = true;
		proxy = IN_Addr(proxyString);
	}        

	isServer = par("aktAsServer").boolValue();

	// set default values of dialog identifier
	secure = false;
	localCSeq = 0;
	remoteCSeq = 0;
        
	// initializes self messages
	sendProgress = NULL;
	sendOk = NULL;
	sendExpires = NULL;
	cancel = NULL;
	bye = NULL;

	inviteRequest = NULL;
        
	// vectors for simulation results
	_incoming_v = new cOutVector( opp_concat("sipDialog-", localUri->getIp(), "-incoming"));
	_outgoing_v = new cOutVector( opp_concat("sipDialog-", localUri->getIp(), "-outgoing"));
	

	// RTP-based parameters
	_enableRTP              = true;
	_rtpPortNumber  	= (int)par("portNumberRTP").longValue();
	_profileName    	= opp_strdup( par("profileNameRTP").stringValue() );
	_bandwidth	        = (int)par("bandwidthRTP").longValue();
	_payloadType    	= (int)par("payloadTypeRTP").longValue();
	_fileName       	= opp_strdup( par("fileNameRTP").stringValue() );
	_nodeName       	= opp_strdup( par("nodeName").stringValue() );	
	_rtpLayer               = NULL;
	_rtpTransmissionStarted = false;
	_rtpSessionEntered      = _rtpClosing = false;	
	_rtpEnterSession        = NULL;
	_rtpStartTransmission   = NULL;
	_rtpLeaveSession        = NULL;
};



/*
	handles received messages
	checks arrival gate and call specific method
*/
void sipDialog::handleMessage(cMessage *msg) {
	
	// self message
	if(msg->isSelfMessage()) 
		handleSelfMsg(msg);
        
	// message from application
	else if(msg->arrivalGateId() == findGate("fromApp"))  
		handleMsgFromApp(msg);

	// message from RTP module
	else if(msg->arrivalGateId() == findGate("fromRTP")) 
		_handleMessageFromRTP( msg );

	delete msg;
};



/*
	cleaning up: delete class variables
*/
void sipDialog::finish() {
	delete inviteRequest; inviteRequest = NULL;
	delete sendProgress; sendProgress = NULL;
	delete sendOk; sendOk = NULL;
	delete sendExpires; sendExpires = NULL;
	delete cancel; cancel = NULL;
	delete bye; bye = NULL;
	delete remoteUri; remoteUri = NULL;
	delete localUri; localUri = NULL;
	delete remoteTarget; remoteTarget = NULL;
	delete _incoming_v;
	delete _outgoing_v;	
};






/*
	handles received messages from the application
*/
void sipDialog::handleMsgFromApp(cMessage *msg) {

	// needed variables
	sipInterfacePacket *sipIntPckt;
	sipResponse *response = NULL;
	sipRequest *request = NULL;
	sipInnerPacket *sip = NULL;
	
	if( 0 == opp_strcmp(msg->className(), "sipInnerPacket") ) {
		// a sipInnerPacket is received
		sip = (sipInnerPacket *)msg;
		
		if( sip->msgType() == sipInnerPacket::MT_BYE ) {
			// BYE
			// create and send BYE request
			if( debug ) ev << "Generating BYE" << endl;
			sendBye();
		}
		else if( sip->msgType() == sipInnerPacket::MT_CANCEL ) {
			// CANCEL
			// create and send CANCEL request
			if( debug ) ev << "Generating CANCEL" << endl;
			if( (sendProgress != NULL) && (sendProgress->isScheduled()) ) {
				delete cancelEvent( sendProgress );
				sendProgress = NULL;
			}
			sendCancel();
		}
		else {
			ev << "Unknown InnerPacket-Type!" << endl;
		}
	}
	
	else if( 0 == opp_strcmp(msg->className(), "sipInterfacePacket") ) {
		// sipInterfacePacket is received
		sipIntPckt = (sipInterfacePacket *)msg;
		
		if((!opp_strcmp(msg->name(), "INVITE")) && (_fsm.state() == UNDEF)) {
			// if dialog is in state UNDEF and this message is received
			// INVITE request from the client dialog is received
			// original INVITE request is necessary for creating a CANCEL or 
			// BYE request
			if(inviteRequest != NULL) delete inviteRequest;
			sipIntPckt = (sipInterfacePacket *)msg;
			inviteRequest = sipIntPckt->getRequest();
		}
			
		else {	// check method
			switch(sipIntPckt->action()) {
			
				case sipInterfacePacket::SA_REQUEST:
					
					// request is received
					request = sipIntPckt->getRequest();
					handleRequest( request );
					break;
                
				case sipInterfacePacket::SA_RESPONSE:

					// response is received
					response = sipIntPckt->getResponse();
					handleResponse( response );
					break;
		
				case sipInterfacePacket::SA_TIMEOUT_REQUEST:
				case sipInterfacePacket::SA_TIMEOUT_RESPONSE:
                                        // a timeout occured - handle timeout
					handleTimeout();
					break;
                
				default:
					break;
			}
		}
	}
	
	delete request;
	delete response;
};



/*
	handles occured timeouts
	if a timeout occured a closeSession() message has to be sent to the 
	parent module
	the module will be deleted by the parent module
*/
void sipDialog::handleTimeout() {
	
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	sipIntPckt->closeSession();
	sendDirect( sipIntPckt, 0.0, parentModule(), "fromDialog");
};



/*
	handles received requests
*/
void sipDialog::handleRequest( sipRequest *request ) {

	// needed variable
	RTPInterfacePacket *rtpIntPckt;

	if(!generalRequestCheck()) {

		_incoming_v->record( (double)request->getMethod());

		// check method of received request
		switch(request->getMethod()) {
                
			case INVITE:
				
				// INVITE request is received
				if(_fsm.state() == UNDEF) {
					// if state is UNDEF, go to state EARLY,
					// set dialog parameter and send 
					// '180 Ringing' repsonse
					FSM_Goto(_fsm, EARLY);
					sendProgress = new cMessage("timer 183");
					sendOk = new cMessage("timer 200");
					remoteCSeq = request->header()->getCSeq().seqNumber;
					inviteRequest = new sipRequest(*request);
					_rtpPortNumber = request->body()->getRtpPort();

					sendProvFinResponse(180, "Ringing", true);
					// timer for sending a '183 Session Progess' and '200 OK' response
					scheduleAt(simTime() + par("timer183").doubleValue(), sendProgress);
					scheduleAt(simTime() + par("timer200").doubleValue(), sendOk);
				}
				break;
				
			case ACK:
			
				// ACK request received
				// go to state CONFIRMED and start RTP session
				if(_fsm.state() == EARLY) {
					 FSM_Goto(_fsm, CONFIRMED);

					if( _enableRTP )
						_rtpLayer = _createRtpModule( opp_strdup(this->name()) );
					if ( _rtpLayer != NULL ) {
						_rtpEnterSession = new cMessage("enterRtpSession");
						scheduleAt( simTime() + par("enterSessionDelayRTP").doubleValue(),
							    _rtpEnterSession );
					}
				}
				break;
				
			case CANCEL:
				
				// CANCEL request received
				// CANCEL request is only possible if no final response has been 
				// received
				// so the dialog is in state EARLY
				// a '487 Request Terminated' response is sent to the INVITE request
				// a '200 OK' response is sent to the CANCEL request
				// dialog terminates
				if(_fsm.state() == EARLY) {
					sendProvFinResponse(487, "Request Terminated", true);
					sendProvFinResponse(200, "OK", false, request);
					goToStateTerminated();
				}
				break;
				
			case BYE:
				
				// BYE request received
				// a BYE request is only possible if a final response has been received
				// so the dialog is in state CONFIRMED; else the BYE is ignored
				// the RTP session is closed, then the dialog terminates
				if(_fsm.state() == CONFIRMED){
				
				if( _rtpLayer != NULL) {
						_rtpClosing = true;

						// delete pointer and messages
						if ( _rtpEnterSession != NULL) {
							delete cancelEvent( _rtpEnterSession);
							_rtpEnterSession = NULL;
						}
						if( _rtpStartTransmission != NULL) {
							delete cancelEvent( _rtpStartTransmission);
							_rtpStartTransmission = NULL;
						}
						if( _rtpLeaveSession != NULL) {
							delete cancelEvent( _rtpLeaveSession );
							_rtpLeaveSession = NULL;
						}
						
						// stop RTP
						if( _rtpTransmissionStarted ) {
							rtpIntPckt = new RTPInterfacePacket("stopRTP");
							rtpIntPckt->stopTransmission(_ssrc);
						
							send( rtpIntPckt, "toRTP");
						}
						else if( _rtpSessionEntered ) {
							_rtpLeaveSession = new cMessage( "leaveRtpSession");
							scheduleAt( simTime() + 
								par("leaveSessionDelayRTP").doubleValue(), _rtpLeaveSession);
						}
						else goToStateTerminated();
					}
					else goToStateTerminated(); 

					// send a '200 OK' response to the BYE request					
					sendProvFinResponse(200, "OK", false, request);
				}
				break;

			default:
				break;
		}
	}
};



/*
	handles received responses
*/
void sipDialog::handleResponse( sipResponse *response ) {
        
	// needed variables
	int status = response->getStatusCode();
	RTPInterfacePacket *rtpIntPckt;


	if(!generalResponseCheck()) {
	
		_incoming_v->record( (double)status);
		// check method of received response
		switch(response->getMethod()) {
		
			case INVITE:
			
				// response to INVITE request received
				// if provisional response is received, go to state EARLY,
				// else go to state CONFIRMED
				// set dialog parameter
				if(status > 299) {
					// response with status code > 299 received
					if(status < 400)
						handle3xxResponse( response );
					else if(status < 500) {
						handle4xxResponse( response );
					}
					else goToStateTerminated();     // status code: 5xx - 6xx
				}
				else if((_fsm.state() == UNDEF) && (status < 200)) {
					// provisional response (status code 1xx)
					localCSeq = response->header()->getCSeq().seqNumber;
					remoteTarget = response->header()->getContact(0)->dup();
					FSM_Goto(_fsm, EARLY);
				}
				else {
					if(status > 199) {
						// status code 2xx
						remoteTarget = response->header()->getTo().dup();
 						route = response->header()->getRouteList();
						FSM_Goto(_fsm, CONFIRMED);
						sendAck();
                                                
						// start RTP
						if( _enableRTP )
							_rtpLayer = _createRtpModule( opp_strdup(this->name()) );					
						if ( _rtpLayer != NULL ) {
							_rtpEnterSession = new cMessage("enterRtpSession");
							scheduleAt( simTime() + par("enterSessionDelayRTP").doubleValue(),
_rtpEnterSession );
						}					
					}
				}
				break;
                                
			case CANCEL:
                        
				// response to a CANCEL request received
				// if state is FAILURE and final response received, go to state 
				// TERMINATED
				// provisional responses are ignored
				if((_fsm.state() == FAILURE) && (status > 199)) 
					goToStateTerminated();
				break;
                                
			case BYE:
                                
				// response to a BYE request received
				// stop RTP and go to state TERMINATED
				if((_fsm.state() == CONFIRMED) && (status > 199)) {
					if( _rtpLayer != NULL) {
						_rtpClosing = true;

						if ( _rtpEnterSession != NULL) {
							delete cancelEvent( _rtpEnterSession);
							_rtpEnterSession = NULL;
						}
						if( _rtpStartTransmission != NULL) {
							delete cancelEvent( _rtpStartTransmission);
							_rtpStartTransmission = NULL;
						}
						if( _rtpLeaveSession != NULL) {
							delete cancelEvent( _rtpLeaveSession );
							_rtpLeaveSession = NULL;
						}
						
						if( _rtpTransmissionStarted ) {
							rtpIntPckt = new RTPInterfacePacket("stopRTP");
							rtpIntPckt->stopTransmission(_ssrc);
							send( rtpIntPckt, "toRTP");
						}
						else if( _rtpSessionEntered ) {
							_rtpLeaveSession = new cMessage( "leaveRtpSession");
							scheduleAt( simTime() + 
								par("leaveSessionDelayRTP").doubleValue(), _rtpLeaveSession);
						}
						else goToStateTerminated();
					}
					else goToStateTerminated(); 
				}
				break;
			default:
				break;
		}
	}
};



/*
	checks correctness of a received request
	not implemented
	returns false if no error was found
*/
bool sipDialog::generalRequestCheck() {
	return false;
};



/*
	sends a response with the given status code and reason phrase
	the header fields are copied from the received request
*/
void sipDialog::sendProvFinResponse(int status, char *reason, bool inv, 
sipRequest *request) {

	// needed variables
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	sipResponse *lastResponseToApp = new sipResponse(status, reason);

	// set response values
	if(useProxy = false)
		lastResponseToApp->setDestination(IN_Addr(remoteUri->getIp()), 
			remoteUri->getPort());
	else lastResponseToApp->setDestination(proxy, remoteUri->getPort());
	lastResponseToApp->header()->setTo( *localUri->dup() );
	lastResponseToApp->header()->setFrom( *remoteUri->dup() );
	lastResponseToApp->header()->setCallId(callId);
	lastResponseToApp->header()->setMaxForwards(70);
	lastResponseToApp->header()->addContact( localUri->dup() );
	if(inv == true) {
		lastResponseToApp->header()->setViaList(
			inviteRequest->header()->getViaList());
		lastResponseToApp->header()->setCSeq(inviteRequest->getMethod(), 
			remoteCSeq);
	}
	else {
		lastResponseToApp->header()->setViaList(request->header()->getViaList());
		lastResponseToApp->header()->setCSeq(request->getMethod(), remoteCSeq);                
	}
        
	_outgoing_v->record( (double)status );
	
	// send response
	sipIntPckt->respond(lastResponseToApp);
	sendDirect( sipIntPckt, 0.0, parentModule(), "fromDialog");
};



/*
	handles received self messages
*/
void sipDialog::handleSelfMsg(cMessage *msg) {

	if(!opp_strcmp(msg->name(), "timer 183")) { // UAS
		// send a '183 Session Progress' response and schedule timer again
		sendProvFinResponse(183, "Session Progress", true);
		sendProgress = new cMessage("timer 183");
		scheduleAt(simTime() + par("timer183").doubleValue(), sendProgress);
	}

	else if(!opp_strcmp(msg->name(), "timer 200")) { // UAS
		// send a '200 OK' response and delete progress timer
		if(sendProgress != NULL && sendProgress->isScheduled()) {
			delete cancelEvent(sendProgress);
			sendProgress = NULL;
		}
		sendProvFinResponse(200, "OK", true);
		sendOk = NULL;
	}

	// RTP
	else if ( 0 == opp_strcmp(msg->name(), "enterRtpSession") && !_rtpClosing ) {
		// enter a RTP session
		_rtpLeaveSession = NULL;
		_rtpEnterSession = NULL;
		_rtpSessionEntered = true;
		RTPInterfacePacket *rIfPkt = new RTPInterfacePacket();
		rIfPkt->enterSession( opp_strdup(opp_concat(_nodeName, "-rtp")),
opp_strdup(_profileName),	_bandwidth, 
IN_Addr( opp_strdup(remoteTarget->getIp()) ),	IN_Port(_rtpPortNumber) );
		send( rIfPkt, "toRTP" );
	}
	
	else if( 0 == opp_strcmp(msg->name(), "startRtpTransmission") ) {    
		// start RTP transmission
		_rtpStartTransmission = NULL;
		if ( 0 == opp_strcmp(_fileName, "") ) {
			// a leaveSession() could be called
		} else {
			RTPInterfacePacket *rIf = new RTPInterfacePacket("startTransmission()");
			rIf->startTransmission( _ssrc, _payloadType, opp_strdup(_fileName) );
			send( rIf, "toRTP" );
			_rtpTransmissionStarted = true;
		}
	}
	
	else if ( 0 == opp_strcmp(msg->name(), "leaveRtpSession") ) {  
		// leave a RTP session
		RTPInterfacePacket *rIf = new RTPInterfacePacket("leaveSession()");
		rIf->leaveSession();
		send(rIf, "toRTP");
	}
	
	else if( 0 == opp_strcmp(msg->name(), "gotoTermination") ) { 
		// delete RTP module and go to state TERMINATED
		_rtpLayer->deleteModule();
		_rtpLayer = NULL;
		goToStateTerminated();
	}
};



/*
	creates and sends a CANCEL request
	 the header fields are copied from the INVITE request accept the method
	method is CANCEL
*/
void sipDialog::sendCancel() {

	// needed variables
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket( "CANCEL" );
	sipRequest *request = new sipRequest();

	request->cancel( inviteRequest );
	request->header()->setTo( *remoteUri );

	_outgoing_v->record( (double)request->getMethod() );

	// copy request as last sent request 
	lastRequest = request->dup();

	// sent CANCEL request only if the dialog is in state EARLY
	// because a provisional response has been received so that the
	// CANCEL request is not received by the communication partner before
	// the INVITE request
	sipIntPckt->request(request->dup());
	if(_fsm.state() == EARLY)
		sendDirect(sipIntPckt, 0.0, parentModule(), "fromDialog");
	else cout << "Dialog (" << id() << "): CANCEL not sent because in not in 
		state EARLY" << endl;
};



/*
	creates and sends a BYE request
	the header fields are copied from the INVITE request accept the method
	method is BYE
*/
void sipDialog::sendBye() {

	// needed variables
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	sipRequest *request = new sipRequest();


	request->bye(inviteRequest->dup());

	// set header fields which have to be re-set
	request->header()->setFrom( *localUri->dup() );
	request->header()->setTo( *remoteUri->dup());
	request->header()->getTo().setUser( remoteUri->getUser());
	request->header()->getFrom().setUser( localUri->getUser());
	request->header()->setUri( *remoteUri->dup());
	request->header()->setCallId(callId);
	request->setDestination(remoteTarget->getIp(), 5060);

	_outgoing_v->record( (double)request->getMethod() );

	// store request as last sent request
	lastRequest = request->dup();

	// sent BYE request only if the dialog is in state CONFIRMED
	// because a final response has been received
	sipIntPckt->request(request);
	if(_fsm.state() == CONFIRMED)
		sendDirect(sipIntPckt, 0.0, parentModule(), "fromDialog");
	else ev << "Dialog (" << id() << "): BYE not sent because no etablished 
		dialog" << endl;
};



/*
	handles if a dialog goes to the state TERMINATED:
	stop timer and send a closeSession() message to the parent module
*/
void sipDialog::goToStateTerminated() {

	// needed variable
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();

	// delete timer if they are scheduled
	if( (sendProgress != NULL) && (sendProgress->isScheduled()) ) {
		delete cancelEvent( sendProgress );
		sendProgress = NULL;
	}
	if( ( sendOk != NULL) && ( sendOk->isScheduled())) {
		delete cancelEvent( sendOk );
		sendOk = NULL;
	}

	// go to state TERMINATED and send closeSession() message
	FSM_Goto(_fsm, TERMINATED);
	sipIntPckt->closeSession();
	_outgoing_v->record( (double)(REC_SIPINTERFACE_ACTION +sipIntPckt->action()));
	sendDirect(sipIntPckt, 0.0, parentModule(), "fromDialog");
};



/*
	creates and sends an ACK request
	the header fields are copied from the INVITE request accept the method
	method is ACK
*/
void sipDialog::sendAck() {

	// needed variables
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket( "ACK");
	sipRequest *request = new sipRequest();

	// create and send ACK
	request->ack(inviteRequest, remoteUri);
	request->header()->setMaxForwards(70);
	request->setDestination(IN_Addr(remoteTarget->getIp()), 
		remoteTarget->getPort());
	_outgoing_v->record( (double)request->getMethod() );
	sipIntPckt->ack(request);
	sendDirect(sipIntPckt, 0.0, parentModule(), "fromDialog");
};



/*
	handles received responses with status code 3xx
	the request is sent once more with a new address
*/
void sipDialog::handle3xxResponse( sipResponse *response ) {

	// needed variables
	int status = response->getStatusCode();
	sipAddress *newAddress;
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	sipRequest *newRequest = new sipRequest( *lastRequest );
        
	// create new request
	newAddress = response->header()->getContact(0)->dup();
	newRequest->header()->setTo( *newAddress); 
	newRequest->setDestination( newAddress->getIp(), lastRequest->getDestPort());
	newRequest->header()->setUri( *newAddress);
        
        
	if( status == 301) {                    // Moved Permanently
		delete remoteUri;
		delete remoteTarget;
		remoteUri =  newAddress->dup();
		remoteTarget = remoteUri->dup();
	}
        
	_outgoing_v->record( (double)response->getStatusCode() );
	
	
	delete lastRequest;
	lastRequest = newRequest->dup();

	// send request once more with re-set header fields
	sipIntPckt->request( newRequest );
	sendDirect( sipIntPckt, 0.0, parentModule(), "fromDialog");
};



/*
	handles received responses with status code 4xx
*/
void sipDialog::handle4xxResponse( sipResponse *response ) {
        
	// needed variable
	int status = response->getStatusCode();

	// if the status code is '487' go to state FAILURE
	// a '487 Request Terminated' response to an INVITE request is received
	if((status == 487) && (_fsm.state() == EARLY)) {
		FSM_Goto(_fsm, FAILURE);
	}
	
	// else terminate
	else goToStateTerminated();
};



/*
	checks if a received response is correct
	if yes, returns false
	not implemented
*/
bool sipDialog::generalResponseCheck() {
	return false;
};



/*
	public method which starts creation and sending of an CANCEL request
*/
void sipDialog::sipCancel() {
	sendCancel();
};



/*
	public method which starts creation and sending of a BYE request
*/
void sipDialog::sipBye() {
	sendBye();
};



/*
	writes the state of the dialog with parameter values
*/
void sipDialog::writeState() {
	ev << "--------------- DIALOG STATE --------------------------" << endl;
	ev << "state of dialog " << this->name() << "(" << id() << ") of class " << 
this->className() << endl;
	ev << "created by " << parentModule()->className() << endl;
	ev << "callID: " << callId << endl;
	ev << "localTag: " << localTag << " , remoteTag: " << remoteTag << endl;
	ev << "localUri: " << localUri->getIp() << " , remoteUri: " << 
remoteUri->getIp() << endl;
	ev << "remoteTarget: " << remoteTarget->getIp() << endl;
	ev << "localCSeq: " << localCSeq << " , remoteCSeq: " << remoteCSeq << endl;
	ev << "secure: " << secure << endl;
	ev << "state: " << _fsm.stateName() << endl;
	ev << "--------------------------------------------------------" << endl;
};



/*
	initializes an RTP-Process module
*/
cModule *sipDialog::_createRtpModule( char *name ) {

	// find and create submodule
	cModuleType *moduleType = findModuleType("RTPLayer");
	cModule *mod = moduleType->create(opp_concat("RTP-", name), this);
	
	// set parameters
	mod->par("rtpModuleType")	= opp_strdup( "RTPEndsystemModule" );
	mod->par("rtcpModuleType")	= opp_strdup( "RTCPEndsystemModule" );
	
	// initialize submodule
	moduleType->buildInside(mod);
	mod->callInitialize();
	mod->scheduleStart(simTime());



	// -------------------------------------------------------------------------
	// locate free SocketLayer links and connect them to rtpLayer
	// 
	// find SocketLayer Module
	cModule *socketMod = parentModule();
	int recursion = 0; // count recursions;
	while( (0 != opp_strcmp(socketMod->className(), "sipApplication")) && 
		(recursion < 6) ) {
		socketMod = socketMod->parentModule();
		recursion++;
	}
	if( 0 != opp_strcmp(socketMod->className(), "sipApplication") ) {
		opp_error("Could not find gate to socketlayer!");
		return NULL;
	}
	socketMod = socketMod->gate("toServerModule")->toGate()->ownerModule()->
		gate("toSocketLayer")->toGate()->ownerModule();
	int j = 0;

	while (socketMod->gate("from_appl",j)->isConnected()) {
		j++;
		if (socketMod->gate("from_appl")->size() == j) 
			opp_error("sipDialog: SocketLayer gateSize() to small.");
	}
	// set links from and to SocketLayerRTP
	mod->gate("toSocketLayerRTP")->setTo(socketMod->gate("from_appl", j));
	mod->gate("fromSocketLayerRTP")->setFrom(socketMod->gate("to_appl", j));
	socketMod->gate("from_appl", j) -> setFrom(mod->gate("toSocketLayerRTP"));
	socketMod->gate("to_appl", j) -> setTo(mod->gate("fromSocketLayerRTP"));

	// locate free SocketLayer link
	j = 0;
	while (socketMod->gate("from_appl",j)->isConnected()) {
		j++;
		if (socketMod->gate("from_appl")->size() == j) 
			opp_error("sipDialog: SocketLayer gateSize() to small.");
	}
	// set links from and to SocketLayerRTCP	
	mod->gate("toSocketLayerRTCP")->setTo(socketMod->gate("from_appl", j));
	mod->gate("fromSocketLayerRTCP")->setFrom(socketMod->gate("to_appl", j));
	socketMod->gate("from_appl", j)->setFrom(mod->gate("toSocketLayerRTCP"));
	socketMod->gate("to_appl", j)->setTo(mod->gate("fromSocketLayerRTCP"));
	
	// App-Gates
	mod->gate("toApp")->setTo( gate("fromRTP") );
	gate("fromRTP")->setFrom( mod->gate("toApp") );
	
	mod->gate("fromApp")->setFrom( gate("toRTP") );
	gate("toRTP")->setTo( mod->gate("fromApp") );
	//
	// gates connected
	// ----------------------
 
	if( debug ) ev << "sipApp: RTPProcess '" << mod->name() << "' - no: " << 
		mod->id() << " created." << endl;
	return mod;
};



/*
        handles received messages from a RTP module
*/
void sipDialog::_handleMessageFromRTP(cMessage *msg) {
	if ( 0 != opp_strcmp(msg->className(), "RTPInterfacePacket") ) {
		opp_error("_handleMessageFromRTP can only handle messages of type 
			'RTPInterfacePacket'!");
		return;
	}
	
	RTPInterfacePacket *rtpmsg = (RTPInterfacePacket *)msg;

	switch( rtpmsg->type() ) {
	case RTPInterfacePacket::RTP_IFP_SESSION_ENTERED:
		// the RTP session was entered
		// start transmission
		_ssrc = rtpmsg->ssrc();
		
		// start transmission
		if( !_rtpClosing ) {
			_rtpStartTransmission = new cMessage("startRtpTransmission");
			scheduleAt( simTime() + par("startTransmissionDelayRTP").doubleValue(),
_rtpStartTransmission );
		}
		break;
	
	case RTPInterfacePacket::RTP_IFP_TRANSMISSION_STARTED:
		// RTP transmission has been started
		if ( debug ) ev << "RTP-transmission (" << _ssrc << ") started." << endl;
		break;
		
	case RTPInterfacePacket::RTP_IFP_TRANSMISSION_FINISHED:
		// the RTP transmission has been finished
		// leave RTP session
		_rtpTransmissionStarted = false;
		_rtpLeaveSession = new cMessage("leaveRtpSession");
		scheduleAt( simTime() + par("leaveSessionDelayRTP").doubleValue(), 
_rtpLeaveSession );
		break;
		
	case RTPInterfacePacket::RTP_IFP_TRANSMISSION_STOPPED:
		// the RTP transmission has been stopped
		// leave RTP session
		_rtpTransmissionStarted = false;
		if ( debug ) ev << "RTP-transmission (" << _ssrc << ") stopped." << endl;
		_rtpLeaveSession = new cMessage("leaveRtpSession");
		scheduleAt( simTime() + par("leaveSessionDelayRTP").doubleValue(), 
_rtpLeaveSession );
		break;
		
	case RTPInterfacePacket::RTP_IFP_SESSION_LEFT:
		// the RTP session has been left
		// go to state TERMINATED
		_rtpSessionEntered = false;
		if ( debug ) ev << "RTP-session left." << endl;
		scheduleAt( simTime() + strToSimtime("30s"), 
new cMessage("gotoTermination") );
		break;		
	
	default:
		if ( debug ) ev << simtimeToStr(simTime()) << " - Received RTP-message 
of type:" << rtpmsg->type() << "." << endl;
		break;
	}

};
