/*
	simulation of SIP
	author: Susann Geisler
	last modified from Susann Geisler on 2003-08-19


	\file sipDialog.h
	\author Susann Geisler
*/

#ifndef SIPDIALOG_H
#define SIPDIALOG_H

// C++-stuff
#include <list>
#include <vector>

// OMNeT++-stuff
#include <omnetpp.h>

// IP-Suite-stuff
#include "SocketInterfacePacket.h"

// SIP-stuff
#include "../Messages/sipRequest.h"
#include "../Messages/sipResponse.h"
#include "../Messages/sipMessageBody.h"
#include "../Messages/sipInnerPacket.h"
#include "../Messages/sipInterfacePacket.h"
#include "sipClientDialog.h"


// RTP-stuff
#include "RTPInterfacePacket.h"

/*!
	\class sipDialog sipDialog.h
        
	sipDialog represents a SIP dialog. It has a id which consists of the call-id,
	the local tag and remote tag. \sa sipClientDialog for more details. This 
	dialog also contains the behavior of a server dialog. It is the complete 
	dialog. 
        
	\author Susann Geisler
	\sa sipApplication, \sa sipClientDialog
*/

class sipDialog: public cSimpleModule {

	Module_Class_Members(sipDialog, cSimpleModule, 0);

	public:
	
		//! finite state machine variable
		/*!
			possible states are UNDEF, EARLY, CONFIRMED, TERMINATED and FAILURE
		*/	
		enum FSM_STATES {
			UNDEF      = 0,
			EARLY      = FSM_Steady(1),
			CONFIRMED  = FSM_Steady(2),
			TERMINATED = FSM_Steady(3),
			FAILURE    = FSM_Steady(4),
		};
		
                
		// OMNeT++ methods
	
		//! initializes variables
		void initialize();
                
		//! handles received messages
		/*!
			\param msg received message
		*/
		void handleMessage(cMessage *msg);
                
		//! cleaning up
		void finish();
		
		
                
		//! starts creation and sending of an INVITE request
		void sipInvite();
                
		//! starts creation and sending of a CANCEL request
		void sipCancel();
                
		//! starts creation and sending of a BYE request
		void sipBye(); 
                
		//! starts creation and sending of a REGISTER request
		void sipRegister();
                		
                
	private:
        
		// parameters
        
		//! finite state machine variable
		cFSM _fsm;
	
		//! original INVITE request and last sent request
		sipRequest *inviteRequest, *lastRequest;
                
		//! parameters which identify dialog uniquely, \sa sipClientDialog
		sipAddress *remoteUri, *localUri, *remoteTarget;
		char *callId, *localTag, *remoteTag; 
		int localCSeq, remoteCSeq;
		bool secure;
		list<sipTagRoute *> route;

		//! indicates which proxy should be used if a proxy should be used
		IN_Addr proxy; 
                
		//! see IN_Addr proxy
		char *proxyString;
                
		//! indicates if a proxy should be used
		bool useProxy;		
	
		//! timer messages for sending a progress / OK response 
		cMessage *sendProgress, *sendOk, *sendExpires;
		//! timer messages for sending a CANCEL / BYE request
		cMessage *cancel, *bye;

		//! indicates if the host acts as server
		bool isServer;
                
		//! indicates the debug mode
		bool debug;
		
		//! if RTP shall be used this has to be "true"
		bool _enableRTP;
		
		//! RTP-Module
		cModule *_rtpLayer;

		//! RTP-ssrc identifier
		u_int32 _ssrc;

		//! RTP-(server-)port of remote Host
		int _rtpPortNumber;
		
		//! RTP-Profile
		char *_profileName;
	
		//! RTP-bandwidth
		int _bandwidth;
	
		//! RTP-payload type
		int _payloadType;
	
		//! file to transport via RTP
		char *_fileName;
	
		//! name of node
		char *_nodeName;
		
		//! indicate if the RTP transmission started, stopped or were closed
		bool _rtpTransmissionStarted, _rtpSessionEntered, _rtpClosing;
		
		//! messages which indicate if the RTP session was entered / left or the 
		//! transmission were started
		cMessage *_rtpEnterSession, *_rtpStartTransmission, *_rtpLeaveSession;
		
		
		//! vector to record simulationtime of incoming messages  
		cOutVector *_incoming_v;
		//! vector to record simulationtime of outgoing messages 
		cOutVector *_outgoing_v;
		
                
		
		
		// methods
	
		//! handles received requests
		/*!
			Handles received requests depending on the current dialog state. A 
			request can be ignored, depending on the state. 
		*/
		void handleRequest( sipRequest *request);
                
		//! handles received responses
		/*!
			handles received responses depending on the current state of the dialog. 
			The dialog can go to an other state or the response is ignored. 
		*/
		void handleResponse( sipResponse *response);
                
		//! handles received self messages
		/*!
			Depending on which self message is received the provisional or final 
			response is sent or the RTP session / transmission is started. 
		*/
		void handleSelfMsg(cMessage *msg);
                
		//! handles received messages from the application
		/*!
			These messages can be responses and requests, depending on if the host 
			acts as server or not. 
		*/
		void handleMsgFromApp(cMessage *msg);
		
		//! handles occured timeouts
		/*!
			The dialog sends a closeSession() message to the parent module and go to 
			the state TERMINATED. The dialog will be deleted by the parent module. 
		*/
		void handleTimeout();
                
		//! handles received responses of class 3xx
		/*!
			The request is sent once more with re-set header fields. These header 
			fields are address header fields of the communication partner
		*/
		void handle3xxResponse( sipResponse *response);
                
		//! handles received responses of class 4xx
		/*!
			If a '487 Request Terminated' response is received, it is a response to 
			an canceled INVITE request. The dialog goes to the state FAILURE. In 
			every other case the dialog terminates. 
		*/
		void handle4xxResponse( sipResponse *response);
                
		
		//! general checks for requests - not implemented
		/*! 
			return false if no error was detected
		*/
		bool generalRequestCheck();
		
		// general checks for and responses - not implemented
		/*! 
			return false if no error was detected
		*/
		bool generalResponseCheck();
                
                
		//! creates and sends a prov. or final response with given status code 
		//! and reason phrase
		// inv is true, if a cancel msg to INVITE is sent (487 Request Terminated)
		void sendProvFinResponse(int status, char *reason, bool inv, 
sipRequest *request = NULL);
 
 
		//! creates and sends a CANCEL request
		/*!
			The values of the original INVITE request are copied accept the method 
			value. The method is CANCEL. The CANCEL request is only send if the 
			dialog is in state EARLY. The dialog has not been established. A 
			provisional response has been received, so that the CANCEL request is 
			not received before the INVITE request. If a CANCEL request is sent
			after a final response has been received, the request is ignored by the 
			communication partner.
		*/
		void sendCancel();
                
		//! creates and sends a BYE request
		/*!
			The values of the original INVITE request are copied accept the method 
			value. The method is BYE. The BYE request is only send if the dialog is 
			in state CONFIRMED. This means a final response has been received. 
		*/
		void sendBye();
                
		//! creates and sends an ACK request
		/*!
			The values of the original INVITE request are copied accept the method 
			value. The method is ACK.
		*/
		void sendAck();
                
                
		//! handles actions when a dialog goes to state TERMINATED
		/*!
			sends a closeSession() request to the parent module and will be deleted
		*/
		void goToStateTerminated();
		
                
		//! creates a RTP module
		cModule *_createRtpModule(char *name);
 
 		//! handles received messages from the RTP module
		void _handleMessageFromRTP(cMessage *msg);
		
          
		// writes state of dialog
		void writeState();		
};


#endif
