//	file: sipServerModule.cc

#include "sipServerModule.h"

Define_Module(sipServerModule);



/*
	initializes socket, name and debug
*/
void sipServerModule::initialize() {

	_socketFd = Socket::FILEDESC_UNDEF;
        
	_name     = opp_strdup( par("commonName").stringValue());
	_debug    = par( "debug").boolValue();
};



/*
	check the gate on which the message was received and 
	call specific method
	messages are only on the gates 'fromSocketLayer', 'fromApp'
	and 'fromServerTransaction' received
*/
void sipServerModule::handleMessage(cMessage *msg) {

	if (msg->arrivalGateId() == findGate("fromSocketLayer")) {
		// gate fromSocketLayer
		handleMessageFromSocketLayer(msg);
	}

	else if (msg->arrivalGateId() == findGate("fromApp")) {
		// gate fromApp
		handleMessageFromApp(msg);
	}

	else if (msg->arrivalGateId() == findGate("fromServerTransaction")) {
		// gate fromServerTransaction
		handleMessageFromTransaction(msg);
	}
	
	delete msg; msg = NULL;
};



/*
	cleaning up: delete class variables
*/
void sipServerModule::finish() {

	delete _name;
	_name = NULL;
};



/*
	check the action of the message (SocketInterfacePacket):
	- SA_SOCKET_RET: send bind() request
	- SA_READ_RET: call handleRequest() 
*/
void sipServerModule::handleMessageFromSocketLayer(cMessage *msg) {

	// decapsulate SocketInterfacePacket from message
	SocketInterfacePacket *sockIntPckt = (SocketInterfacePacket *)msg;
        
	// needed variables
	sipRequest *request;
	SocketInterfacePacket *bindMessage;
	
	// set address and port of communication partner (Socket)
	remote_addr = sockIntPckt->fAddr();
	remote_port = sockIntPckt->fPort();


	// check action of SocketInterfacePacket
	switch(sockIntPckt->action()) {
	
		case SocketInterfacePacket::SA_SOCKET_RET:
		
			// socket return
			if(_debug) ev << "serverModule: SA_SOCKET_RET from socketLayer 
received." << endl;
                        
			// set socket variable
			_socketFd = sockIntPckt->filedesc();
                        
			// send bind() message
			bindMessage = new SocketInterfacePacket("BindCall");
			bindMessage->bind(_socketFd, remote_addr, 5060);
			send(bindMessage, "toSocketLayer");
                        
			break;
                
		case SocketInterfacePacket::SA_READ_RET:
		
			// read return
			if(_debug) ev << "serverModule: SA_READ_RET from socketLayer 
received." << endl;
                        
			// decapsulate request from SocketInterfacePacket and call 
			// handleRequest()
			request = (sipRequest *)sockIntPckt->decapsulate();
			handleRequest(request);
                        
			break;

		default:
			if(_debug) ev << "serverModule: action not supported." << endl;
                        
			break;
	}
};



/*
	check the action of the message (sipInterfacePacket)
	- SA_OPEN_PORT: send socket() request to socket layer to open a port
	- SA_RESPONSE:  forward response to related transaction
*/
void sipServerModule::handleMessageFromApp(cMessage *msg) {
        
	// decapsulate sipInterfacePacket from message
	sipInterfacePacket *sipIntPckt = (sipInterfacePacket *)msg;

	// needed varialbes
	SocketInterfacePacket *socketMessage;
	sipInterfacePacket *messageForTransaction;
	sipResponse *_response;
	cModule *_mod;


	// check action
	switch(sipIntPckt->action()) {
	
		case sipInterfacePacket::SA_OPEN_PORT:

			// open port
			if(_debug) ev << "serverModule: SA_OPEN_PORT from app received." << endl;
			
			// send socket() message to socket layer
			socketMessage = new SocketInterfacePacket("SocketCall");
			socketMessage->socket(Socket::AF_INET, Socket::SOCK_DGRAM, Socket::UDP);
			send(socketMessage, "toSocketLayer");
			
			break;

		case sipInterfacePacket::SA_RESPONSE:

			// response
			if( _debug) ev << "serverModule: SA_RESPONSE from app received." << endl;
			
			// decapsulate response from sipInterfacePacket
			_response = sipIntPckt->getResponse();

			// set the module id of the related transaction
			_mod = simulation.module(findSubmodule(_response->responseKey()));
			

			// if no related transaction exists a new one has to be created
			// case RESPOND_ONLY
			if ( _mod == NULL) {
				_createSession( _response->responseKey());
				_mod = simulation.module( findSubmodule( _response->responseKey()));
			}
		
			// forward response to related transaction
			messageForTransaction = new sipInterfacePacket();
			messageForTransaction->respond(_response->dup());
			sendDirect(messageForTransaction, 0.0, _mod, "fromServerModule");
			
			break;
        
		case sipInterfacePacket::SA_REQUEST:
		
			// request: ignored
			if( _debug) ev << "serverModule: SA_REQUEST received and ignored" << endl;
			break;
        
		default:
			if(_debug) ev << "serverModule: action not supported." << endl;
			break;
	}
};



/*
	check the action of the received message (sipInterfacePacket)
	- SA_REQUEST: forward request to app
	- SA_CLOSE_SESSION: the transaction reached the state TERMINATED and has to 
		be deleted
	- SA_TIMEOUT_RESPONSE: a timeout took place and is forwarded  to 
		sipApplication
*/
void sipServerModule::handleMessageFromTransaction(cMessage *msg) {

	// decapsulate sipInterfacePacket from message
	sipInterfacePacket *sipIntPckt = (sipInterfacePacket *)msg;
	
	// needed variables
	int id = msg->senderModuleId();
	cModule *mod;


	// check the action of the sipInterfacePacket
	switch(sipIntPckt->action()) {
	
		case sipInterfacePacket::SA_REQUEST:
		
			// request
			if(_debug) ev << "serverModule: SA_REQUEST from transaction 
received." << endl;
			
			// forward request to sipApplication
			send(sipIntPckt->dup(), "toApp");
			
			break;

		case sipInterfacePacket::SA_CLOSE_SESSION:

			// close session - delete transaction
			if(_debug) ev << "serverModule: SA_CLOSE_SESSION from transaction 
received." << endl;
                        
			// find the senders module id and delete module
			mod = simulation.module(id);
			deleteTransaction(mod);
			
			break;
			
		case sipInterfacePacket::SA_TIMEOUT_RESPONSE:
			
			// timeout - forward to sipApplication
			send( sipIntPckt->dup(), "toApp");
			break;
        
		case sipInterfacePacket::SA_RESPONSE:
			
			// response: ignored
			if( _debug) ev << "serverModule: response from transaction received 
and ignored" << endl;
			break;
		
		default:
			if(_debug) ev << "serverModule: action not supported." << endl;
			break;
	}
};



/*
	looks for the related transaction and forward request to it
	if no related transaction exists, a new one is created
*/
void sipServerModule::handleRequest(sipRequest *request) {

	// needed variable
	cModule *_mod = NULL;
	

	// if method of the request is not ACK and no transaction exists, a new 
	// transaction is created
	if((simulation.module(findSubmodule(request->requestKey())) == NULL) && 
		(request->getMethod() != ACK))
		_createSession(request->requestKey());


	// find the module of the related transaction
	
	// the method is ACK and via is used (not NULL)
	if((request->getMethod() == ACK) && (request->header()->getVia() != NULL)) {
		_mod = simulation.module(findSubmodule(opp_concat(
request->header()->getVia()->getBranch(),	"INVITE")));
	}

	// method is ACK and via is not used (NULL)
	else if((request->getMethod() == ACK) && 
(request->header()->getVia() == NULL)) {
		_mod = simulation.module(findSubmodule("INVITE"));
	}

	// the method is not ACK
	else {
		_mod = simulation.module(findSubmodule(request->requestKey()));
	}


	// setReceived() in via fiedl
	request->header()->getVia()->setReceived( remote_addr, remote_port );

	// If the method is ACK and no transaction module exists anymore, a 2xx 
	// response to the INVITE request was received and the transaction module 
	// was deleted. The ACK request is sent to the application directly. Else 
	// the ACK request is sent to the transaction module. Every request where 
	// the method is not ACK is sent to the application. The request is 
	// encapsulated in a sipInterfacePacket. 
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	if((request->getMethod() == ACK) && (_mod == NULL)) {
		sipIntPckt->ack( request);
		send(sipIntPckt->dup(), "toApp");
	}
	else {
	sipIntPckt->request( request);
		sendDirect(sipIntPckt->dup(), 0.0, _mod, "fromServerModule");
	}
};


/*
	deletes a transaction module 
*/
void sipServerModule::deleteTransaction(cModule *mod) {

	if(_debug) ev << "serverModule: transaction " << mod->name() << 
		" is deleted." << endl;
	mod->deleteModule();
};


/*
	creates a transaction module
	name of the module is the requestKey
	gates to socketLayer are connected
*/

void sipServerModule::_createSession(const char *name) {

	// needed variable
	cModuleType *_moduleType;
	cModule *_mod, *socketMod;

	if(_debug) ev << "serverModule: a new transaction called " << name 
		<< " is created." << endl;

	// find module type and create module
	_moduleType = findModuleType("sipServerTransaction");
	_mod = _moduleType->create( opp_strdup(name), this );

	// add parameter to the new module
	_mod->par("debug") = par("debug");
	_mod->par("remote_addr") = remote_addr;
	 _mod->par("remote_port") = remote_port; // par("remote_port");

	// call the remaining OMNeT++ methods for creating a module
	_moduleType->buildInside(_mod);
	_mod->callInitialize();
	_mod->scheduleStart(simTime());



	// The following code was taken from Arndt Buschmann's RTSP code - thanks
        
	// find SocketLayer Module
	socketMod = gate("toSocketLayer")->toGate()->ownerModule();
	// find free gate at SocketLayer
	int j = 0;
	while (socketMod->gate("from_appl", j)->isConnected()) {
		j++;
		if (socketMod->gate("from_appl")->size() == j) 
			opp_error("SocketLayer gateSize() to small.");
	}
	// set links from and to SocketLayer
	_mod->gate("toSocketLayer") -> setTo(socketMod->gate("from_appl", j));
	_mod->gate("fromSocketLayer") -> setFrom(socketMod->gate("to_appl", j));
	socketMod->gate("from_appl", j) -> setFrom(_mod->gate("toSocketLayer"));
	socketMod->gate("to_appl", j) -> setTo(_mod->gate("fromSocketLayer"));
};

