/*!
	simulation of SIP
	author: Susann Geisler
	last modified from Susann Geisler on 2003-08-19


	\file sipServerModule.h
	\author Susann Geisler
*/

#ifndef SIPSERVERMODULE_H
#define SIPSERVERMODULE_H

// C++ stuff
#include <stdio.h>
#include <stdlib.h>

// OMNeT++ stuff
#include <omnetpp.h>

// IP-Suite stuff
#include "SocketInterfacePacket.h" 
#include "sockets.h"

// SIP stuff
#include "sipServerTransaction.h"
#include "../Messages/sipRequest.h"
#include "../Messages/sipResponse.h"
#include "../Messages/sipInterfacePacket.h"


//! SIP server module
/*!
	\class sipServerModule sipServerModule.h
	This module represents the main module of a SIP server. It manages
	transactions: their creation and destruction. It forwards messages to the
	related transaction. If no related transaction module exists, a new one is
	created.
	
	\author Susann Geisler
	\sa sipServerTransaction
*/

class sipServerModule: public cSimpleModule {

	Module_Class_Members(sipServerModule, cSimpleModule, 0);
	
	public:
		// OMNeT++ methods
		
		//! initializes variables
		void initialize();
		
		//! handles received messages
		/*!
			\param msg received message
		*/
		void handleMessage(cMessage *msg);
		
		//! cleaning up
		void finish();
		
	private:
	
		// parameters
	
		//! name of host
		char *_name;
		
		//! UDP port number
		int _udpPort;
		
		//! indicates debug mode
		bool _debug;
                
		
		//! address of communication partner
		IN_Addr remote_addr;
		
		//! port of communication partner
		IN_Port remote_port;
                
		
		//! used socket
		Socket::Filedesc _socketFd;

               
		// methods
	       
	       
		//! handles received messages from the socket layer
		/*!
			the action of the message (SocketInterfacePacket) is checked:
			\li SA_SOCKET_RET: a response to a socket() request is received and 
					 a bind() request is sent
			\li SA_READ_RET: a message from an other host is received - 
				handleRequest() is called (the socket has been bound)

			\param msg received message
		*/
		void handleMessageFromSocketLayer(cMessage *msg);
		
		
		//! handles received message from sipApplication
		/*!
			the action of the message (sipInterfacePacket) is checked:
			\li SA_OPEN_PORT: it is the first message from sipApplication and tells 
				the sipServerModule to open a port to the socket layer on which 
					  it listens for new messages
			\li SA_RESPONSE:  a response to a forwarded request is received and has 
				to be sent to the transaction
		
			\param msg received message
		*/
		void handleMessageFromApp(cMessage *msg);
		
		
		//! handles received messages from transactions
		/*!
			the action of the received message (sipInterfacePacket) is checked:
			\li SA_REQUEST: forward request to app
			\li SA_CLOSE_SESSION: the transaction reached the state TERMINATED 
				and has to be deleted
			\li SA_TIMEOUT_RESPONSE: a timeout took place and is forwarded  to 
				sipApplication
			
			\param msg received message
		*/
		void handleMessageFromTransaction(cMessage *msg);
                
		
		
		//! handles received requests
		/*!
			Looks for the related transaction. If no transaction exists, a new 
				one is created. Then the request is forwarded to it. 
			\param request received request
		*/
		void handleRequest(sipRequest *request);
		
		
		
		//! creates new transactions
		/*!
			\param name name of the new transaction
		*/
		void _createSession(const char *name);
		
		
		
		//! deletes transactions
		/*!
			\param mod id of the module which has to be deleted
		*/
		void deleteTransaction(cModule *mod);
};


#endif 
