//	file: sipServerTransaction.cc

#include "sipServerTransaction.h"

Define_Module(sipServerTransaction);


/*
	set phase, variables and timers
*/
void sipServerTransaction::initialize() {

	// set socket
	_socketFd = Socket::FILEDESC_UNDEF;
        
	lastResponseToSocket = NULL;
        
	// set variable of finite state machine
	_fsm.setName("_fsm");
	FSM_Goto(_fsm, INIT);

	// timer messages
	timeout_G_msg = NULL;
	timeout_H_msg = NULL;
	timeout_I_msg = NULL;
	timeout_J_msg = NULL;

	// values of timer
	_timerG = strToSimtime("500ms"); // T1 = 500ms
	_timerH = strToSimtime("32s"); // = 64 * T1 = 32 000ms
	_timerI = strToSimtime("5s"); // T4 = 5s
	_timerJ = strToSimtime("32s"); // = 64 * T1 = 32s
	_T2 = strToSimtime("4s"); // T2 = 4s
	
	// do not become owner of any messages...
	takeOwnership( false );
};



/*
	handles received messages: call specific method for specific gate
	if received message is a self message (timer) call handleSelfMessage()
	possible gates:
		- gate fromSocketLayer   -  method handleMessageFromSocketLayer()
		- gate fromServerModule  -  method handleMessageFromServerModule()
		
*/
void sipServerTransaction::handleMessage(cMessage *msg) {

	if (msg->arrivalGateId() == findGate("fromSocketLayer"))
		handleMessageFromSocketLayer(msg);
        
	else if (msg->arrivalGateId() == findGate("fromServerModule")) 
		handleMessageFromServerModule(msg);
        
	else if(msg->isSelfMessage())
		handleSelfMessage(msg);
		
	delete msg;
};



/*
	cleaning up: delete messages
*/
void sipServerTransaction::finish() {

	delete lastResponseToSocket; lastResponseToSocket = NULL;
	delete timeout_G_msg; timeout_G_msg = NULL;
	delete timeout_H_msg; timeout_H_msg = NULL;
	delete timeout_I_msg; timeout_I_msg = NULL;
	delete timeout_J_msg; timeout_J_msg = NULL;
};



/*
	handles messages which were received from the socketLayer
	the action of the message (SocketInterfacePacket) is checked:
		- SOCKET_RET: a response to a socket() request is received and a connect() 
			request is sent to the socket layer
		- CONNECT_RET: the socket is bound now and all responses which have not 
			been sent are sent now
		- READ_RET: a request from an other host is received, call handleRequest()
*/
void sipServerTransaction::handleMessageFromSocketLayer(cMessage *msg) {

	// decapsulate SocketInterfacePacket from message
	SocketInterfacePacket *sockIntPckt = (SocketInterfacePacket *)msg;
	
	// needed variables
	SocketInterfacePacket *connectMessage;
	sipRequest *request = NULL;


	// check action of the SocketInterfacePacket
	switch(sockIntPckt->action()) {
	
		case SocketInterfacePacket::SA_SOCKET_RET:
		
			// socket return
			connectMessage = new SocketInterfacePacket("connectCall");
			if(_debug) ev << "serverTransaction: SA_SOCKET_RET from socketLayer 
received." << endl;

			// set socket variable
			_socketFd = sockIntPckt->filedesc();

			// send connect() request to socket layer
			connectMessage->connect(_socketFd, IN_Addr( 
opp_strdup(par("remote_addr").stringValue()) ),
IN_Port((int)(par("remote_port").longValue()))); 
			send(connectMessage, "toSocketLayer");
        
			break;
	
		case SocketInterfacePacket::SA_CONNECT_RET:
		
			// connect return - socket is bound now
			if(_debug) ev << "serverTransaction: SA_CONNECT_RET from socketLayer 
received." << endl;
       
			// send all responses which are in the to-send queue
			while( !_sendQueue.empty() ) {
				sendResponseToSocket( (sipResponse *)_sendQueue.pop() );
			}
			
			// if we are in state RESPOND_ONLY - we can exit here
			if ( _fsm.state() == RESPOND_ONLY ) {
				goToStateTerminated();
			}
		
			break;
		
		case SocketInterfacePacket::SA_READ_RET:
		
			// read return - request received
			if(_debug) ev << "serverTransaction: SA_READ_RET from socketLayer 
				received." << endl;
        
			// decapsulate request from SocketInterfacePacket and call 
			// handleRequest()
			request = (sipRequest *)sockIntPckt->decapsulate();
			handleRequest( request);
			delete request;
        
			break;

		default:
			break;
	}
	
};



/*
	the action of the message (sipInterfacePacket) is checked:
	- SA_REQUEST: a request from an other host was received
		if socket has not been bound, send socket() call; call handleRequest()
	- SA_RESPONSE: a response from the application is received
		call handleResponse()
*/
void sipServerTransaction::handleMessageFromServerModule(cMessage *msg) {

	// decapsulate received sipInterfacePacket from message
	sipInterfacePacket *sipIntPckt = (sipInterfacePacket *)msg;
	
	// needed variables
	SocketInterfacePacket *socketMessage;
	sipRequest *request = NULL;
	sipResponse *response = NULL;


	// check action of sipInterfacePacket
	switch(sipIntPckt->action()) {
	
		case sipInterfacePacket::SA_REQUEST:
		
			// request was received
			if(_debug) ev << "serverTransaction: SA_REQUEST from serverModule 
				received." << endl;
                        
			// decapsulate request from sipInterfacePacket
			request = sipIntPckt->getRequest();

			// if socket has not been bound send socket() request
			if(_socketFd == Socket::FILEDESC_UNDEF) {
				socketMessage = new SocketInterfacePacket("socketCall");
				socketMessage->socket(Socket::AF_INET, Socket::SOCK_DGRAM, Socket::UDP);
				send(socketMessage, "toSocketLayer");
			}
        		
			// call handleRequest()
			handleRequest( request);
			delete request;
	
			break;
	
		case sipInterfacePacket::SA_RESPONSE:
		
			// response was received
			if(_debug) ev << "serverTransaction: SA_RESPONSE from serverModule 
				received." << endl;
        
			// decapsulate response from sipInterfacePacket and call handleResponse()
			response = sipIntPckt->getResponse();
			handleResponse(response);
			delete response;
        
			break;
	
		default:
			break;
	}
	
};



/*
	handles received self messages (timer)
*/
void sipServerTransaction::handleSelfMessage(cMessage *msg) {

	if(msg == timeout_G_msg)	// timer G fired
		handleTimerG();

	else if(msg == timeout_H_msg)	// timer H fired
		handleTimerH();

	else if(msg == timeout_I_msg)	// timer I fired
		handleTimerI();

	else if(msg == timeout_J_msg) {	// timer J fired
		handleTimerJ();
	}

	else
		if(_debug) ev << "received self message not understood." << endl;
        
};



/*
	handles received requests
	check the state of the transaction and call specific method
*/
void sipServerTransaction::handleRequest( sipRequest *request) {

	FSM_Switch(_fsm) {
		case FSM_Exit(INIT):				// state INIT
			handleRequestStateInit( request);
			break;

		case FSM_Enter(TRYING):				// state TRYING
		case FSM_Exit(TRYING): 
		case FSM_Enter(PROCEEDING):			// state PROCEEDING
			break;

		case FSM_Exit(PROCEEDING):
			handleRequestStateProceeding();
			break;

		case FSM_Enter(COMPLETED):			// state COMPLETED
			break;

		case FSM_Exit(COMPLETED):
			handleRequestStateCompleted( request);
			break;

		case FSM_Enter(CONFIRMED):			// state CONFIRMED
		case FSM_Exit(CONFIRMED): 
		case FSM_Enter(TERMINATED):			// state TERMINATED
		case FSM_Exit(TERMINATED):
			break;
	}
};



/*
	handles received responses
	check the state of the transaction and call specific method
*/
void sipServerTransaction::handleResponse( sipResponse *response ) {

	FSM_Switch(_fsm) {
		case FSM_Enter(INIT):				// state INIT
		case FSM_Exit(INIT):
			sendResponseToSocket( response->dup() );
			FSM_Goto( _fsm, RESPOND_ONLY );
			break;

		case FSM_Enter(TRYING):				// state TRYING
			break;

		case FSM_Exit(TRYING):
			handleResponseStateTrying( response );
			break;

		case FSM_Enter(PROCEEDING):			// state PROCEEDING
			break;

		case FSM_Exit(PROCEEDING):
			handleResponseStateProceeding( response );
			break;

		case FSM_Enter(COMPLETED):			// state COMPLETED
		case FSM_Exit(COMPLETED):
		case FSM_Enter(CONFIRMED):			// state CONFIRMED
		case FSM_Exit(CONFIRMED):
		case FSM_Enter(TERMINATED):			// state TERMINATED
		case FSM_Exit(TERMINATED):
			break;
	}
};



/*
	Handles received requests when the transaction is in state INIT. The behavior
	for INVITE and non-INVITE request is different. A '100 Trying' response has to
	be sent for INVITE requests because of the long response time. For non-INVITE 
	request a provisional response (class 1xx) should not be sent. 
	If an INVITE request was received the transaction goes to state PROCEEDING,
	if non-INVITE it goes to state TRYING. 
*/
void sipServerTransaction::handleRequestStateInit( sipRequest *request) {

	// decapsulate received sipInterfacePacket from message
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();
	
	// needed variable
	sipResponse *response = NULL;


	// send the request to serverModule
	sipIntPckt->request( request->dup());
	sendDirect( sipIntPckt, 0.0, parentModule(), "fromServerTransaction");


	// set state and send provisional response if the method of the request is 
	// INVITE
	if( request->getMethod() == INVITE) {
		// INVITE
		response = new sipResponse(100, "Trying");
		response->header()->setCallId( request->header()->getCallId());
		response->header()->setTo( request->header()->getTo());
		response->header()->setFrom( request->header()->getFrom());
		response->header()->setUri( request->header()->getUri());
		response->header()->setViaList( request->header()->getViaList());
		response->header()->setCSeq( request->header()->getCSeq());

		lastResponseToSocket = response->dup();
		sendResponseToSocket( response );

		FSM_Goto(_fsm, PROCEEDING);
	}
	else if(request->getMethod() != ACK) { 
		// non-INVITE
		FSM_Goto(_fsm, TRYING);
	}
};



/*
	Handles received requests when the transaction is in state PROCEEDING. 
	In this state the last response sent to the socket layer is sent once more. 
*/
void sipServerTransaction::handleRequestStateProceeding() {

	sendLastResponseOnceMore();
};



/*
	Handles received requests when the transaction is in state COMPLETED. If the
	method of the request is INVITE the last response sent to the socket layer is
	sent once more. If the method is ACK, the transaction goes to state CONFIRMED
	and timer I is set. 
*/
void sipServerTransaction::handleRequestStateCompleted( sipRequest *request) {

	if(request->getMethod() == ACK) {
		// method is ACK
		FSM_Goto(_fsm, CONFIRMED);
		if(timeout_I_msg == NULL) timeout_I_msg = new cMessage("timer I");
		scheduleAt(simTime() + _timerI, timeout_I_msg);
	}
	
	else sendLastResponseOnceMore(); 	// method is not ACK
};



/*
	Handles received responses when the transaction is in state TRYING. This state
	is only possible for non-INVITE requests. The given response is sent to the 
	socket layer. 
	The status code of the response is checked: if a provisonal response (class 
	1xx) is received, the transaction goes to state PROCEEDING, else it goes to 
	state COMPLETED.
*/
void sipServerTransaction::handleResponseStateTrying( sipResponse *response) {

	if(! (( response->getMethod() == INVITE) || 
		( response->getMethod() == ACK))) {
		// non-INVITE
		// send response to socket
		sendResponseToSocket( response->dup() );

		// set state
		if(response->getStatusCode() < 200) 		// provisional response
			FSM_Goto(_fsm, PROCEEDING);

		else {						// final response
			FSM_Goto(_fsm, COMPLETED);
			if(timeout_J_msg == NULL) timeout_J_msg = new cMessage("timer J");
			scheduleAt(simTime() + _timerJ, timeout_J_msg);
		}
	}
};



/*
	Handles received responses when the transaction is in state PROCEEDING. The 
	response is sent to the socket layer. 
	If the status code of the response is 2xx or greater, the state of the 
	transaction changes. 
		- non-INVITE: 
			PROCEEDING -> COMPLETED and set timer J 
		- INVITE    :
			2xx      : PROCEEDING -> TERMINATED
			300 - 699: PROCEEDING -> COMPLETED and set timer G and H
*/
void sipServerTransaction::handleResponseStateProceeding(sipResponse *response){

	// needed variables: method and status code of response
	sipRequestMethod _meth = response->getMethod();
	int statuscode = response->getStatusCode();
        
	// send response to socket layer
	sendResponseToSocket( response->dup() );

	// set new state
	if( _meth != ACK) {
		if( _meth == INVITE) {
			if( statuscode > 199 && statuscode < 300) {
				goToStateTerminated();
                        }
			if( statuscode > 299)
				FSM_Goto( _fsm, COMPLETED);
		}
		else {
			if( statuscode > 199)
				FSM_Goto( _fsm, COMPLETED);
		}
	}

	// store response as last response sent to socket layer
	if( lastResponseToSocket != NULL) delete lastResponseToSocket;
	lastResponseToSocket = ( sipResponse *)response->dup();

	// set Timer
	if( _meth == INVITE && statuscode > 299) {
		// INVITE and final response
		if( timeout_G_msg != NULL) delete timeout_G_msg;
		if( timeout_H_msg != NULL) delete timeout_H_msg;
			timeout_G_msg = new cMessage("timer G");
			timeout_H_msg = new cMessage("timer H");
	
		scheduleAt( simTime() + _timerG, timeout_G_msg);
		scheduleAt( simTime() + _timerH, timeout_H_msg);
	}
	if( _meth != INVITE && _meth != ACK) {
		// non-INVITE
		if( timeout_J_msg == NULL) timeout_J_msg = new cMessage("timer J");
		scheduleAt( simTime() + _timerJ, timeout_J_msg);
	}
};



/*
	sends the last response sent to socketLayer once more 
	used if the same request is received more than once in certain states
*/
void sipServerTransaction::sendLastResponseOnceMore() {

	// copy last response sent to socket layer
	sipResponse *sResponse = ( sipResponse *)lastResponseToSocket->dup();

	// send response
	SocketInterfacePacket *sockIntPckt;
	sockIntPckt->write( _socketFd, sResponse);
	send( sockIntPckt, "toSocketLayer");
};



/*
	sends the give response to the socket layer
*/
void sipServerTransaction::sendResponseToSocket( sipResponse *response) {

	// needed variable
	SocketInterfacePacket *sockIntPckt;

	// if the socket is bound send response now
	// else store response in the send-queue and send response when the socket 
	// is bound
	if( _socketFd != Socket::FILEDESC_UNDEF) {
		sockIntPckt = new SocketInterfacePacket("writeCall");
		sockIntPckt->write( _socketFd, response );
		send( sockIntPckt, "toSocketLayer");
	} 
	
	else {
		_sendQueue.insert( response );
	}
	
};



/*
	Handles fired timer G. The last response sent to the socket is sent once more.
	Timer G ist re-set. Timer G is only important for state CONFIRMED.
*/
void sipServerTransaction::handleTimerG() {

	if(_debug) ev << "serverTransaction: Timer G fired." << endl;

	timeout_G_msg = NULL;


	// check state of the transaction
	FSM_Switch(_fsm) {
	
		case CONFIRMED:
			timeout_G_msg = new cMessage("timer G"); 
        
			// send last response to socketLayer once more to socketLayer
			sendLastResponseOnceMore();

			// set timer G again
			_timerG = min(2*_timerG, _T2);
			scheduleAt(simTime() + _timerG, timeout_G_msg);

			break;
		
		default:
			if(_debug) 
				ev << "transaction not in state CONFIRMED - timer G ignored." << endl;
			break;
	}
};



/*
	Handles fired timer H. A timeout took place. The TU (sipApplication) has to be
	informed. Send timeoutResponse() message to sipApplication. 
*/
void sipServerTransaction::handleTimerH() {

	// needed variable
	sipInterfacePacket *sipIntPckt = new sipInterfacePacket();

	if(_debug) ev << "serverTransaction: Timer H fired." << endl;
	timeout_H_msg = NULL;

	// send timeout message to sipApplication
	sipIntPckt->timeoutResponse( lastResponseToSocket->dup() );
	sendDirect( sipIntPckt, 0.0, parentModule(), "fromServerTransaction");
};



/*
	Handles fired timer I. If timer I fires the transaction goes to state
	TERMINATED. 
*/
void sipServerTransaction::handleTimerI() {

	if(_debug) ev << "serverTransaction: Timer I fired." << endl;

	timeout_I_msg = NULL;
	
	// go to state TERMINATED
	if( _fsm.state() != TERMINATED)
		goToStateTerminated();
};



/*
	Handles fired timer J. If timer J fires the transaction goes to state
	TERMINATED. 
*/
void sipServerTransaction::handleTimerJ() {

	if(_debug) ev << "serverTransaction: Timer J fired." << endl;

	timeout_J_msg = NULL;

	// go to state TERMINATED
	if( _fsm.state() != TERMINATED) 
		goToStateTerminated();
};



/*
	The transaction goes to state TERMINATED. The socket is closed and a
	closeSession() request is sent to the server module which will delete the 
	transaction. 
*/
void sipServerTransaction::goToStateTerminated() {

	// go to state TERMINATED
	FSM_Goto( _fsm, TERMINATED);
        
	// close socket
	if( _socketFd != Socket::FILEDESC_UNDEF ) {
		SocketInterfacePacket *sockipack = new SocketInterfacePacket("closeSocket");
		sockipack->close( _socketFd );
		send( sockipack, "toSocketLayer" );
	}
	
	// delete last response sent to socket
	if( lastResponseToSocket != NULL ) {
		delete lastResponseToSocket;
		lastResponseToSocket = NULL;
	}
	
	// send closeSession() to server module
        sipInterfacePacket *sipIntPckt = new sipInterfacePacket("closeSession");
        sipIntPckt->closeSession();
        sendDirect( sipIntPckt, 0.0, parentModule(), "fromServerTransaction");
};
