/*
	simulation of SIP
	author: Susann Geisler
	last modified from Susann Geisler on 2003-08-19


	\file sipServerTransaction.h
	\author Susann Geisler
*/

#ifndef SIPSERVERTRANSACTION_H
#define SIPSERVERTRANSACTION_H


// C++ stuff
#include <math.h>

// OMNeT++ stuff
#include <omnetpp.h>

// SIP stuff
#include "../Messages/sipInterfacePacket.h"
#include "../Messages/sipHeader.h" 
#include "../Messages/sipConstants.h"
#include "../Messages/sipResponse.h"
#include "../Messages/sipRequest.h"

// IP-Suite stuff
#include "SocketInterfacePacket.h" 


/*!
	\class sipServerTransaction sipServerTransaction.h
	This module represents a SIP server transaction. It is realized by
	a finite state machine including the states INIT, TRYING, PROCEEDING,
	COMPLETED, CONFIRMED, TERMINATED and RESPOND_ONLY. 
	
	\author Susann Geisler
	\sa sipServerModule
*/

class sipServerTransaction: public cSimpleModule {

	Module_Class_Members(sipServerTransaction, cSimpleModule, 0);
	
	public:
		
		//! defined states INIT, TRYING, PROCEEDING, COMPLETED, CONFIRMED, 
		//! TERMINATED and RESPOND_ONLY
		enum FSM_STATES {
			INIT         = 0,
			TRYING       = FSM_Steady(1),
			PROCEEDING   = FSM_Steady(2), 
			COMPLETED    = FSM_Steady(3), 
			CONFIRMED    = FSM_Steady(4),
			TERMINATED   = FSM_Steady(5),
			RESPOND_ONLY = FSM_Steady(6)
		};  
		
		
		
		
		// OMNeT++ methods
		
		//! initializes variables
		void initialize();
		
		//! handles received messages
		/*!
			\param msg received message
		*/
		void handleMessage(cMessage *msg);
		
		//! cleaning up
		void finish();
        
	private:
	
		// parameters
	
		//! finite state machine variable
		cFSM _fsm;
		
		//! socket variable
		Socket::Filedesc _socketFd;


		//! stores last response sent to socket layer 
		sipResponse *lastResponseToSocket; 


		//! indicates debug mode
		bool _debug;

		//! stores responses which have to be sent to the socket layer until 
		//! the socket is bound
		cQueue _sendQueue;
	
	
		//! messages for timer
		cMessage *timeout_G_msg;	// timer G
		cMessage *timeout_H_msg; 	// timer H
		cMessage *timeout_I_msg;	// timer I
		cMessage *timeout_J_msg;	// timer J
		
		//! values of timer
 		simtime_t _timerG;		// timer G
		simtime_t _timerH;		// timer H
		simtime_t _timerI;		// timer I
		simtime_t _timerJ;		// timer J
		simtime_t _T2;			// timer T2
		simtime_t _T4;			// timer T4

                
		// methods
		
		
		//! handles received messages from the socket layer
		/*!
			the action of the message (SocketInterfacePacket) is checked:
			\li SOCKET_RET: a response to a socket() request is received and a 
				connect() request is sent to the socket layer
			\li CONNECT_RET: the socket is bound now and all responses which have 
				not been sent are sent now
			\li READ_RET: a request from an other host is received, 
				call handleRequest()
			
			\param msg received message
		*/
		void handleMessageFromSocketLayer(cMessage *msg);
		
		
		//! handles received messages from the server module
		/*!
			the action of the message (sipInterfacePacket) is checked:
			\li SA_REQUEST: a request from an other host was received - 
				call handleRequest() \n
				if socket has not been bound, send socket() call
			\li SA_RESPONSE: a response from the application is received - 
				call handleResponse()
				
			\param msg received message
		*/
		void handleMessageFromServerModule(cMessage *msg);
		
		
		//! handles received self messages (timer)
		/*!
			\param msg received self message
		*/
		void handleSelfMessage(cMessage *msg);
                
		
		
		//! handles received request from the network
		/*!
			\param request received request
		*/
		void handleRequest(sipRequest *request);
		
		
		//! handles received responses from the application
		/*!
			\param response received response
		*/
		void handleResponse(sipResponse *response);
                
		
		
		//! handles received requests when the transaction is in state INIT
		/*!
			Handles received requests when the transaction is in state INIT. The 
			behavior for INVITE and non-INVITE request is different. A '100 Trying' 
			response has to be sent for INVITE requests because of the long 
			response time. For non-INVITE request a provisional response (class 1xx) 
			should not be sent. \n
			If an INVITE request was received the transaction goes to state 
			PROCEEDING, if non-INVITE it goes to state TRYING. 
			\param request received request
		*/
		void handleRequestStateInit(sipRequest *request);
		
		
		//! handles received requests when the transaction is in state PROCEEDING
		/*!
			Handles received requests when the transaction is in state PROCEEDING. 
			In this state the last response sent to the socket layer is sent once 
			more.
			\param request received request
		*/
		void handleRequestStateProceeding();
		
		
		//! handles received requests when the transaction is in state COMPLETED
		/*!
			Handles received requests when the transaction is in state COMPLETED. 
			If the method of the request is INVITE the last response sent to the 
			socket layer is sent once more. If the method is ACK, the transaction 
			goes to state CONFIRMED and timer I is set. 
			\param request received request
		*/
		void handleRequestStateCompleted(sipRequest *request);
                
		
		
		//! handles received responses when the transaction is in state TRYING
		/*!
			Handles received responses when the transaction is in state TRYING. 
			This state is only possible for non-INVITE requests. The given response 
			is sent to the socket layer. \n
			The status code of the response is checked: if a provisonal response 
			(class 1xx) is received, the transaction goes to state PROCEEDING, else 
			it goes to state COMPLETED.
			\param response received response
		*/
		void handleResponseStateTrying(sipResponse *response);
		
		
		//! handles received responses when the transaction is in state PROCEEDING
		/*!
			Handles received responses when the transaction is in state PROCEEDING. 
			The response is sent to the socket layer. \n
			If the status code of the response is 2xx or greater, the state of the 
			transaction changes. If a response to a non-INVITE request is received 
			the transaction goes to state COMPLETED. Timer J is set. If a response 
			to an INVITE request is received the new state depends on the status 
			code of the response. If the status code is of class 2, the new state 
			is TERMINATED. Else the transaction goes to state COMPLETED and timer 
			G and H are set.
			\param response received response
		*/
		void handleResponseStateProceeding(sipResponse *response);
		
		
		
		//! handles fired timer G
		/*!
			Handles fired timer G. The last response sent to the socket is sent 
			once more. Timer G ist re-set. Timer G is only important for state 
			CONFIRMED.
		*/
		void handleTimerG();
		
		
		//! handles fired timer H
		/*!
			Handles fired timer H. A timeout took place. The TU (sipApplication) 
			has to be informed. Send timeoutResponse() message to sipApplication. 
		*/
		void handleTimerH();
		
		
		//! handles fired timer I
		/*!
			Handles fired timer I. If timer I fires the transaction goes to state
			TERMINATED. 
		*/
		void handleTimerI();
		
		
		//! handles fired timer J
		/*!
			Handles fired timer J. If timer J fires the transaction goes to state
			TERMINATED. 
		*/
		void handleTimerJ();
                
		
		
		//! manages actions when the transaction goes to state TERMINATED
		/*!
				The transaction goes to state TERMINATED. The socket is closed and a
				closeSession() request is sent to the server module which will 
				delete the transaction. 
		*/
		void goToStateTerminated();
                
		
		
		//! sends the last reponse sent to the socket layer once more
		void sendLastResponseOnceMore();
		
		
		//! sends the given response to the socket layer
		/*!
			\param response response which has to be sent
		*/
		void sendResponseToSocket(sipResponse *response);
};

#endif
