/*!
	\file	sipHeader.h
	\author	jAN bURGHARD
	\author Susann Geisler
	\version 0.2
	
*/
#ifndef __SIPHEADER_H
#define __SIPHEADER_H

// for gcc3.x
using namespace std;

// OMNeT++ stuff
#include <omnetpp.h>

// C++ stuff
#include <list>
#include <vector>

// SIP stuff
#include "sipTags.h"

//! SIP-Header
/*!
	\class sipHeader sipHeader.h
	\author Jan Burghard
	\author Susann Geisler
	
	A basic message header class
*/


class sipHeader {

	private:
		//! value of URI-field
		sipAddress	_uri;
		//! value of From-field
		sipAddress	_from;
		//! value of To-field
		sipAddress	_to;
		//! value of Contact
		vector<sipAddress *> _contact;

		//! value of CallId-field
		stCallId	_callId;
		//! value of CSeq-field
		stCSeq		_cSeq;

		//! ordered list of Via-field values
		list<sipTagVia *> _via;
		//! ordered list of Route-field values
		list<sipTagRoute *> _route;
//		//! ordered list of RecordRoute-field values
//		list<sipTagRecordRoute *> _recordRoute;

		//! value of ContentType-field
		char		*_contentType;
		//! value of ContentLength-field
		unsigned int	_contentLength;
		//! value of Required-field
		char		*_require;
		//! array of Allow-field values
		int *_allow;
		//! array of Supported-field values
		int *_supported;
		//! array of Accept-field values
		int *_accept;
		//! array of ContentDisposition-field values
		int *_contentDisposition;
		//! array of ContentLanguage-field values
		int *_contentLanguage;
		//! array of ContentEncoding-field values
		int *_contentEncoding;

		//! value of Expires-field
		int _expires;

		//! value of Max-Forwards-field
		/*!
		Max-Forwards - headerfield. If unset it's value is "-1", if
		set it's greater.
		*/
		int		_maxForwards;

		/*!
		initializes some stuff
		*/
		void _init();
	public:
		/*!
		default constructor
		*/
		sipHeader();

		/*!
		copy-constructor
		*/
		sipHeader(const sipHeader& h);

		/*!
		destructor
		*/
		virtual ~sipHeader();

		/*!
		assignement operator
		*/
		sipHeader& operator=(const sipHeader& h);

		// --------------
		// set - Methoden
		// --------------

		/*! set Uri-value                                            */
		void setUri(sipAddress address);

		/*! set From-value                                           */
		void setFrom(sipAddress address);

		/*! set To-value                                             */
		void setTo(sipAddress address);

		/*! add a Contact-value                                      */
		void addContact(sipAddress *address);

		/*! add a Contact-Vector                                     */
		void setContacts(vector<sipAddress *> contacts);



		/*! sets a CallId-structure                                  */
		void setCallId(stCallId s);
		/*! set  CallId-value                                        */
		void setCallId(const char *id);

		/*! sets a CSeq-structure                                    */
		void setCSeq(stCSeq s);
		/*! set  CSeq-value                                          */
		void setCSeq(sipRequestMethod method, const int seq);



		/*! adds a "Via"-headerfield structure to the top of the
		    via-list                                                 */
		void addVia(sipTagVia *v);
		/*!
		creates a new "Via"-headerfield structure from the given parameters
		and adds it to the top of the via-list
		*/
		void addVia(const Socket::Protocol p, const char *ip, const int port);
		/*! sets a complete Via-field list                           */
		void setViaList(list<sipTagVia *> via);

		/*! \sa addVia                                               */
		void addRoute(sipTagRoute *r);
		/*! \sa addVia                                               */
		void addRoute(char *ip, int port);
		/*! \sa setViaList                                           */
		void setRouteList(list<sipTagRoute *> route);

//		/*! \sa addVia                                               */
//		void addRecordRoute(sipTagRecordRoute *r);
//		/*! \sa addVia                                               */
//		void addRecordRoute(char *ip, int port);
//		/*! \sa setViaList                                           */
//		void setRecordRouteList(const list<sipTagRecordRoute *> rr);



		/*! set ContentType-value                                    */
		void setContentType(char *s);

		/*! set ContentLength-value                                         */
		void setContentLength(int i);

		/*! set Require-value                                               */
		void setRequire(char *s);

		/*! sets complete Allow-field array				*/
		void setAllow(int *allow);

		/*! sets complete Supported-field array				*/
		void setSupported(int *supported);

		/*! sets complete Accept-field array				*/
		void setAccept(int *accept);

		/*! sets complete ContentDisposition-field array 		*/
		void setContentDisposition(int *disposition);

		/*! sets complete ContentLanguage-field array                */
		void setContentLanguage(int *language);

		/*! sets complete ContentEncoding-field array                */
		void setContentEncoding(int *encoding);



		/*! sets Expires-value                                       */
		void setExpires(int expires);



		/*! set Max-Forwards-value                                   */
		void setMaxForwards(const int i);




		// --------------
		// get - Methoden
		// --------------


		/*! returns Uri-value                                        */
		sipAddress getUri() const;

		/*! returns From-value                                       */
		sipAddress getFrom() const;

		/*! returns To-value                                         */
		sipAddress getTo() const;

		/*! returns the number of stored Contact-values              */
		unsigned int getContactCount() const;

		/*! returns a Contact-value by index                         */
		sipAddress *getContact(int i) const;
		


		/*! returns CallId-structure                                 */
		stCallId getCallId() const;

		/*! returns CSeq-structure                                   */
		stCSeq getCSeq() const;



		/*! returns topmost Via-field*/
		sipTagVia *getVia() const;
		/*! returns topmost Via-field and removes it from list       */
		sipTagVia *popVia();
		/*! returns the complete Via-field list                      */
		list<sipTagVia *> getViaList() const;

		/*! returns topmost Route-field                              */
		sipTagRoute *getRoute() const;
		/*! returns topmost Route-field and removes it from list     */
		sipTagRoute *popRoute();
		/*! returns the complete Route-field list                    */
		list<sipTagRoute *> getRouteList() const;

//		/*! returns topmost RecordRoute-field                        */
//		sipTagRecordRoute *getRecordRoute() const;
//		/*! returns topmost RecordRoute-field and removes it from 
//		    list                                                     */
//		sipTagRecordRoute *popRecordRoute();
//		/*! \sa getViaList                                           */
//		list<sipTagRecordRoute *> getRecordRouteList() const;



		/*! returns ContentType-value                                */
		char *getContentType() const;

		/*! returns ContentLength-value                              */
		int getContentLength() const;

		/*! returns Require-value                                    */
		char *getRequire() const;

		/*! returns Allow-field array                                */
		int *getAllow();

		/*! returnsSupported-field array                             */
		int *getSupported();

		/*! returns Accept-field array                               */
		int *getAccept();

		/*! returns ContentDisposition-field array                   */
		int *getContentDisposition();

		 /*! returns ContentLanguage-field array                     */
		 int *getContentLanguage();

		/*! returns ContentEncoding-field array                      */
		int *getContentEncoding();



		/*! returns Expires-value                                    */
		int getExpires() const;



		/*! returns Max-Forwards-value                               */
		int getMaxForwards() const;
};

#endif
