/*!
	\file sipInterfacePacket.h
	\author jAN bURGHARD
*/

#ifndef __SIPINTERFACEPACKET_H__
#define __SIPINTERFACEPACKET_H__

#include <omnetpp.h>
#include "sockets.h"

#include "../Messages/sipRequest.h"
#include "../Messages/sipResponse.h"

/*  -------------------------------------------------
    Type Definitions
    -------------------------------------------------   */

//! Interfacepacket for communication between SIP-elements.
/*!
	\class sipInterfacePacket sipInterfacePacket.h
	\author Jan Burghard
	\author Susann Geisler
	
	Interface message for communication between SIP-Elements
*/
class sipInterfacePacket: public cMessage {

public:
	/*!
	enumeration of defined sip-actions
	*/
	enum sipAction {
		// SA = SipAction
		SA_UNDEF,
		SA_REQUEST,
		SA_RESPONSE,
		SA_TERMINATE,
		SA_OPEN_PORT,
		SA_TIMEOUT_REQUEST,
		SA_TIMEOUT_RESPONSE,
		SA_FORWARD,
		SA_CLOSE_SESSION,
		SA_FROM_SOCKET,
		SA_ACK
	};

private:

	//! identifies the current action
	sipAction _action;   
      
	//! initializes the sipInterfacePacket
	void _init();
	
	//! clears all contents and resets _action
	void _clear();
	
	//! raises an opp_error if already initilized
	void _check();


public:
	//
	// Constructors, destructors,...
	//
	
	/*!
	default constructor
	*/
	sipInterfacePacket();
	
	/*!
	constructor for named initialization
	\param name ...guess...
	*/
	sipInterfacePacket(const char* name );
	
	/*!
	copy constructor
	\param sIfPkt packet to be copied
	*/
	sipInterfacePacket(const  sipInterfacePacket& sIfPkt );

	/*!
	assignement operator
	*/
	sipInterfacePacket& operator=(const sipInterfacePacket& sIfPkt );
	
	/*!
	duplicate method
	\return a copy of this
	*/
	virtual sipInterfacePacket *dup() const {
		return new sipInterfacePacket(*this); 
	}

	//! OMNeT++ method - returns the classname
	virtual const char *className() const { return "sipInterfacePacket"; }
	
	//! OMNeT++ method - adds infos of this to buf
	virtual void info( char *buf );
	
	//! OMNeT++ method
	virtual void writeContents( ostream& os );





	//
	// API  
	// sipApplication  <--> sipClientModule
	// sipClientModule <--> sipClientTransaction
	// sipApplication  <--> sipServerModule
	// sipServerModule <--> sipServerTransaction
	//

	
	//! orders a module to open a port
	/*!
	instruct sipServerModule to connect a server socket to the given port \n
	source: sipApplication --> destination: sipServerModule \n
	\param port portnumber
	*/
	void openPort( IN_Port port );

	//! informs about a request timeout
	/*!
	message-type to inform about send/receive-timeouts on requests \n
	source: sipClientTransaction --> destination: sipApplication \n
	source: sipClientTransaction --> destination: sipClientModule \n
	*/
	void timeoutRequest( sipRequest *request );
	
	//! informs about a response timeout
	/*!
	message-type to inform about send/receive-timeouts on responses \n
	source: sipServerTransaction --> destination: sipApplication \n
	source: sipServerTransaction --> destination: sipServerModule \n
	*/
	void timeoutResponse( sipResponse *response );
	
	//! orders termination
	/*!
	instruct a sipClientTransaction to terminate \n
	source: sipClientModule --> destination: sipClientTransaction \n
	source: sipApplication  --> destination: sipClientTransaction \n
	*/
	void terminate();
	
	//! to forward a message
	/*! 
	the Packet consists a message to be forwarded
	*/
	void forward( cMessage *msg );
	
	//! "ready to be terminated"-message
	/*!
	instruct the sipClientModule to terminate the sending sipClientTransaction \n
	source: sipClientTransaction --> destination: sipClientModule \n
	*/
	void closeSession();
	
	//! inter-sip-request forward
	/*!
	used to send sipRequests between sip-elements\n
	source: sipApplication  --> destination: sipClientModule \n
	source: sipClientModule --> destination: sipClientTransaction \n
	*/
	void request( sipRequest *request );
	
	//! counterpart to request and timeoutRequest
	/*!
	returns the encapsulated sipRequest
	\return sipRequest 
	*/
	sipRequest *getRequest();
	
	//! inter-sip-response forward
	/*!
	used to send sipResponses between sip-elements \n
	source: sipApplication  --> destination: sipServerModule \n
	source: sipClientModule --> destination: sipServerTransaction \n
	*/
	void respond( sipResponse *response );
	
	//! counterpart to respond and timeoutResponse
	/*!
	returns the encapsulated sipResponse
	\return sipResponse
	*/
	sipResponse *getResponse();
	
	//! forwards a sipRequest, received by the socketLayer
	/*!
	source: sipServerModule --> destination: sipServerTransaction \n
	*/
	void fromSocket( sipRequest *request );
	
	//! special case of "inter-sip-request forward"
	/*!
	used to send transactionless sip-ACK-Requests between sip-elements\n
	source: sipApplication  --> destination: sipClientModule \n
	source: sipClientModule --> destination: sipClientTransaction \n
	*/
	void ack( sipRequest *request );
	
	//
	// general
	// 
	// accessing private data
	//! returns the action or type of the sipInterfacePacket.
	sipAction action() const { return _action; }

};

#endif // __SIPINTERFACEPACKET_H__
