/*!
	\file	sipRequest.cc
	\author	jAN bURGHARD
	\version 0.2
	
*/
#include "sipRequest.h"



// constructor
sipRequest::sipRequest()
	: cMessage("sipRequest") 
{
	_debug = true;
	_init();
	_body.clear();
};

// destructor
sipRequest::~sipRequest() {
	delete _sipVersion;
};

// copy constructor
sipRequest::sipRequest(const sipRequest& r) 
	: cMessage()
{
	setName( r.name() );
	operator=( r );
};

// info method
void sipRequest::info(char *buf)
{
	cMessage::info(buf);
	sprintf( buf+strlen(buf), "%s %s %s", 
		sipRequestMethodStr[_method], _requestURI.toString(), _sipVersion );
};

// writeContents method
void sipRequest::writeContents(ostream& os)
{
	os << sipRequestMethodStr[_method] << " " << _requestURI.toString() 
		<< " " << _sipVersion << endl;
};

// operator=
sipRequest& sipRequest::operator=(const sipRequest& r)
{
	cMessage::operator=( r );
	
	_method      = r._method;
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
	_requestURI  = r._requestURI;
	_sipVersion  = opp_strdup( r._sipVersion );
	_destination = r._destination;
	_header      = r._header;
	_body        = r._body;

	return *this;
};

// init() method
void sipRequest::_init() {
	_method = UNDEF;
	if( _debug ) addPar("Method") = sipRequestMethodStr[_method];
	_sipVersion = opp_strdup( "2.0" );
	setKind( MK_SIPREQUEST );
};


// checks the method of the request
// must be UNDEF
void sipRequest::_check() {
	if ( _method != UNDEF ) {
		opp_error("sipRequest: has already a request-method");
	}
};


// creates an INVITE request from the given parameter
void sipRequest::invite(sipAddress from, sipAddress to, char *call_id, 
	int cseq, sipTagVia *via) {
	_check();
	
        // set method
	_method = INVITE;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
	_requestURI   = to;
	
        // set known header fields
	_header.setTo( to );
	_header.setFrom( from );
	_header.setCallId( call_id );
	_header.setCSeq( INVITE, cseq );
	_header.addVia( via );
};


/*
to create an incomplete INVITE-request. Used to schedule an 
INVITE-request-command.
The missing values have to be added later.
*/
void sipRequest::invite(sipAddress to) {
	_check();
	
	_method		= INVITE;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
	header()->setCSeq( INVITE, 0 );
	header()->setTo( to );
	_requestURI	= to;
	
};

//
void sipRequest::completeInvite(sipAddress from, char *call_id, int cseq, 
	sipTagVia *via) {
	if( _method != INVITE ) {
		opp_error("unexpected call of 'completeInvite' on sipRequest!");
		return;
	};
	
	_header.setFrom( from );
	_header.setCallId( call_id );
	_header.setCSeq( INVITE, cseq );
	_header.addVia( via );
};

/*
   The ACK request constructed by the client transaction MUST contain
   values for the Call-ID, From, and Request-URI that are equal to the
   values of those header fields in the request passed to the transport
   by the client transaction (call this the "original request").  The To
   header field in the ACK MUST equal the To header field in the
   response being acknowledged, and therefore will usually differ from
   the To header field in the original request by the addition of the
   tag parameter.  The ACK MUST contain a single Via header field, and
   this MUST be equal to the top Via header field of the original
   request.  The CSeq header field in the ACK MUST contain the same
   value for the sequence number as was present in the original request,
   but the method parameter MUST be equal to "ACK".
*/
void sipRequest::ack(sipRequest *request, sipAddress *to) {
	_check();
	
	if ( request->getMethod() != INVITE ) {
		opp_error("can only create ACKs from INVITE-requests!");
		return;
	}
	
	// set request method
	_method = ACK;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
	
	// copy Call-ID, From and Request-URI
	_header.setCallId( request->header()->getCallId() );
	_header.setFrom( request->header()->getFrom() );
	_requestURI = *request->getRequestURI()->dup();
	
	// copy To from the response
	_header.setTo( *to );
	
	// Via filed from request
	_header.addVia( request->header()->getVia()->dup() );
	
	// empty message-body
	_body.clear();
	
	_header.setCSeq( ACK, 
	                 request->header()->getCSeq().seqNumber );

};

// does not make sense - at the moment
void sipRequest::ack(sipAddress to) {
	_check();

	// set request method
	_method = ACK;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
};

// creates a CANCEL request from the given (INVITE) request
void sipRequest::cancel( sipRequest *request ) {
	_check();
	
	// first: copy Request
	operator=( *request );
	
	// second: change values
	if ( request->getMethod() != INVITE ) {
		opp_error("can only create CANCELS from INVITE-requests!");
		return;
	}
	_method = CANCEL;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
	

	// edit headerfields
	_header.setCSeq( CANCEL,
	                 request->header()->getCSeq().seqNumber );

	// empty message-body
	_body.clear();	
};

// creates a CANCEL request with the given ip address
void sipRequest::cancel( char *ip ) {
	_check();
	
	_method = CANCEL;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];

	// empty message-body
	_body.clear();

	header()->setTo( sipAddress(ip, 5060) );
};

// creates a BYE request from the given (INVITE) request
void sipRequest::bye(sipRequest *request) {
	_check();
	
	operator=(*request);
	_method = BYE;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
	// empty message body
	_body.clear();

	_header.setCSeq(BYE, request->header()->getCSeq().seqNumber);
};

// creates a BYE request with the given ip address
void sipRequest::bye( char *ip) {
	_check();
	
	_method = BYE;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
	_body.clear();
	

	sipAddress to = _header.getTo();
	to.setIp( ip );
	_header.setTo( to );
};




// create a REGISTER request
/*
being a reseverd expression, this function is called "sipRegister" in
spite of "register". Params are the proxy on which the registration 
shall be done and the address which shall be mapped (from).
The addresses "from" shall be mapped to have to be added later as
"contact"-addresses (header()->addContact( sipAddress * )).
*/
void sipRequest::sipRegister(sipAddress proxy, sipAddress from) {
	_check();
	
	_method = REGISTER;
	setName( sipRequestMethodStr[_method] );
	if( _debug ) par("Method") = sipRequestMethodStr[_method];
	header()->setCSeq( _method, 0 );
	
	header()->setUri( proxy );
	header()->setTo( from );
	header()->setFrom( from );
};

// returns the key of the request consisting of branch-id and method
char *sipRequest::requestKey() const {
	if ( _header.getVia() != NULL ) {
		return opp_concat( _header.getVia()->getBranch(),
		                   sipRequestMethodStr[ _method ] );
	}
	else {
		return sipRequestMethodStr[ _method ];
	}
};

// returns the method
sipRequestMethod sipRequest::getMethod() {
	return _method;
};

// returns the addressee of the request
sipAddress sipRequest::getTo() {
	return _header.getTo();
};

// returns the destination of the request
void sipRequest::setDestination(const char *ip, const int port) {
	_destination.setIp( ip );
	_destination.setPort( port );
};

// returns the destination ip of the request
char *sipRequest::getDestIp() const {
	return _destination.getIp();
};

// returns the destination port of the request
int sipRequest::getDestPort() const {
	return _destination.getPort();
};

// returns a pointer to the request header
sipRequestHeader *sipRequest::header() {
	return &_header;
};

// returns a pointer to the request message body
sipMessageBody *sipRequest::body() {
	return &_body;
};

// returns the SIP version
char *sipRequest::getVersion() {
	return _sipVersion;
};

// returns the Request URI 
sipAddress *sipRequest::getRequestURI() {
	return &_requestURI;
};

// sets the Request URI
void sipRequest::setRequestURI(sipAddress *uri) {
	_requestURI = sipAddress( *uri );
};

// writes content of request
void sipRequest::writeMessageState() {
	ev << "--------------------------------------------------------" << endl;
	ev << "--------------- REQUEST MESSAGE STATE ------------------" << endl;
	ev << "--------------------------------------------------------" << endl;
	ev << "msg of class " << className() << endl;
	ev << "method: " << getMethod() << endl;
	ev << "INVITE request" << endl;
	ev << "callID: " << header()->getCallId().id << endl;
	ev << "localTag: " << header()->getFrom().getTag();
	ev << " , remoteTag: " << header()->getTo().getTag() << endl;
	ev << "from: " << header()->getFrom().getIp();
	ev << " , to: " << header()->getTo().getIp()<< endl;
	ev << "request URI: " << header()->getUri().getIp() << endl;
	ev << "cseq.number: " << header()->getCSeq().seqNumber;
	ev << " , cseq.method: " << header()->getCSeq().method << endl;
	ev << "max forwards: " << header()->getMaxForwards() << endl;
	ev << "number of contact fields: " << header()->getContactCount() << endl;
	for(unsigned int i = 0; i < header()->getContactCount(); i++) {
		ev << "contact: " << header()->getContact( i )->getIp() << endl;
	}
	ev << "---------------------------------------------------------" << endl;
};
