/*!
	simulation of SIP
	author: Jan Burghard
	last modified from Susann Geisler on 2003-08-19


	\file	sipRequest.h
	\author	jAN bURGHARD
	\version 0.2
	
*/
#ifndef __SIPREQUEST_H
#define __SIPREQUEST_H

#include <omnetpp.h>

#include "sipRequestHeader.h"
#include "sipMessageBody.h"
#include "sipConstants.h"
#include "sipTags.h"


//! SIP-Request
/*! 
\class sipRequest sipRequest.h
\author Jan Burghard
This class represents a request message of the SIP 
*/
class sipRequest : public cMessage {

private:
	//! indicates the debug mode
	bool			_debug;
	
	//! method parameter of the request
	sipRequestMethod	_method;
	
	//! address to where the request will be sent
	sipAddress		_requestURI;
	
        //! used SIP version
	char			*_sipVersion;
        
        //! the header of the request
	sipRequestHeader	_header;
        
        //! the message body of the request
	sipMessageBody		_body;
	
	//! next-hop destination address
	sipAddress _destination;


        //! initializes variables
	void _init();
        
        //! checks state of the request - must be UNDEF
        /*!
                is used to guarantee that no request is used
                more than once
        */
	void _check();
	
public:
	//! default constructor
	sipRequest();
	
	/*! destructor                                                        */
	~sipRequest();
	
	//! copy-constructor
	sipRequest(const sipRequest& r);
	
	//! returns class name string
	virtual const char *className() const { return "sipRequest"; }
        
	//! duplicates the object by creating an exact copy (uses copy constructor)
	virtual sipRequest *dup() const { return new sipRequest(*this); }
        
  //! returns a one-line info about object contents
	virtual void info(char *buf);
        
  //! write a more detailed report about the object into a file
	virtual void writeContents(ostream& os);
        
  //! the assignement operator (copies object contents from another object)
	sipRequest& operator=(const sipRequest& r);
	
	//! set next-hop destination
	void setDestination(const char *ip, const int port);

	//! get next-hop IP
	char *getDestIp() const;
	
	//! get next-hop Port
	int getDestPort() const;


	//
	// Methods to set up a correct SIP-Request
	//
	//! create an INVITE request
	/*!
	used for direct INVITE-request creation.
	Makes necessary changes to header values.
	*/
	void invite(sipAddress from, sipAddress to, char *call_id, 
		int cseq, sipTagVia *via);
	
	//!
	/*!
	to create an incomplete INVITE-request. Used to schedule an 
	INVITE-request-command.
	The missing values have to be added later.
	*/
	void invite(sipAddress to);
	
	//!
	/*!
	add missing information to a request created with "void invite(sipAddress to)"
	*/
	void completeInvite(sipAddress from, char *call_id, int cseq, sipTagVia *via);

	//! create an ACK request
	/*!
	makes necessary changes to header values. Most values will be taken from
	the original request
	*/
	void ack(sipRequest *request, sipAddress *to);
	
	//!
	/*!
	designed for being called from commandfile-reader
	*/
	void ack(sipAddress to);

	//! create a CANCEL request
	/*!
	copies most values of the given request and makes necessary changes to
	header values
	*/
	void cancel(sipRequest *request);
	
	void cancel( char *ip);

	//! create a BYE request
	/*!
	copies most values of given request and makes necessary changes to 
	header values
	*/
	void bye(sipRequest *request);
	
	void bye(char *ip);

	//! create a REGISTER request
	/*!
	being a reseverd expression, this function is called "sipRegister" in
	spite of "register". Params are the proxy on which the registration 
	shall be done and the address which shall be mapped (from).
	The addresses "from" shall be mapped to have to be added later as
	"contact"-addresses (header()->addContact( sipAddress * )).
	
	This method can directly be called from commandfile-reader
	*/
	void sipRegister(sipAddress proxy, sipAddress from);


	//! unique request-identifier
	/*!
	\return a string of the via-tags branch-parameter and the request
	method
	*/
	char *requestKey() const;

	//! returns a pointer to the sipRequestHeader
	sipRequestHeader *header();
	
	//! returns a pointer to the sipMessageBody
	sipMessageBody *body();

	//! returns the SIP version of the request
	char *getVersion();
	
	//! returns the request-URI
	sipAddress *getRequestURI();
	
	//! sets the request-URI
	void setRequestURI(sipAddress *uri);
	
	//! returns the method of the request
	sipRequestMethod	getMethod();
	
	//! returns the addressee of the request
	sipAddress		getTo();

        //! writes the state of the request with the parameter values	
        void writeMessageState();
};


#endif
