/*!
	\file	sipTags.h
	\author Susann Geisler
	\author	jAN bURGHARD
	\version 0.1
	
*/
#ifndef __SIPTAGS_H
#define __SIPTAGS_H

// for gcc3.x
using namespace std;

// C++-stuff
// #include <string>

// OMNeT++ stuff
#include <omnetpp.h>

// IP-Suite stuff
#include "sockets.h"

// SIP stuff
#include "sipConstants.h"


/* ----------------------------------------------------------------------
   several tag-classes
   ---------------------------------------------------------------------- */

//! sipAddress
/*!
	\class sipAddress sipTags.h
	\author Susann Geisler
	
	this class represents a sip-url
		
*/
class sipAddress {
public:
	enum userParam {
		UNDEF	= 0,
		phone	= 1,
		ip	= 2
	};

private:
	char	*_user;
	char	*_password;
	int	_telesub;
	char	*_domainlabel;
	char	*_toplabel;
	char	*_ip;
	char	*_host;
	int	_port;
        char    *_tag;
	Socket::Protocol _transportProt;
	userParam _userparam;
	int	_requestmethod; // should be "sipRequest::sip_request_method"
	int	_ttlparam;
	char	*_maddrparam;
	
	//! initialization method
	void _init();
	
public:
	//! default constructor
	sipAddress();
	
	//! copy constructor
	sipAddress(const sipAddress& a);
	
	//! constructor
	/*!
	\param user name of the addressed user
	\param host name of the addressed host
	*/
	sipAddress(const char *user, const char *host);

	//! constructor
	/*!
	\param ip ip-address of the addressed host
	\param port portnumber of the addressed host
	*/
	sipAddress(const char *ip, const int port);

	//! constructor
	/*!
	\param user name of the addressed user
	\param ip ip-address of the addressed host
	\param port portnumber of the addressed host
	*/
	sipAddress(const char *user, const char *ip, const int port);
	
	//! default destructor
	virtual ~sipAddress();
	
	//! duplicate method
	virtual sipAddress *dup() const { return new sipAddress(*this); }
	
	//! the assignement operator (copies object contents from another object)
	sipAddress& operator=(const sipAddress& a);
	
	//! address comparison
	bool equals(const sipAddress& a);
	
	//! returns the whole address ass a char*
	char *toChar() const;
	
	//! set username
	void setUser(const char *s);
	//! set Password
	void setPassword(const char *s);
	//! set Telesub
	void setTelesub(const int i);
	//! set Domainlabel
	void setDomainlabel(const char *s);
	//! set Toplabel
	void setToplabel(const char *s);
	//! set Hostname
	void setHost(const char *s);
	//! set ip-address
	void setIp(const char *s);
	//! set portnumber
	void setPort(const int i);
	//! set TransportProtocol
	void setTransportProtocol(const Socket::Protocol s);
	//! set UserParam
	void setUserParam(const sipAddress::userParam userp);
	//! set MethodParam
	void setMethodParam(const int method);
	//! set TtlParam
	void setTtlParam(const int i);
	//! set MaddrParam
	void setMaddrParam(const char *ip);
	//! set Tag
        void setTag(char *tag);
	
	//! returns the username
	char *getUser() const;
	//! returns the password
	char *getPassword() const;
	//! returns Telesub
	int getTelesub() const;
	//! returns Domainlabel
	char *getDomainlabel() const;
	//! returns Toplabel
	char *getToplabel() const;
	//! returns Hostname
	char *getHost() const;
	//! returns ip-address
	char *getIp() const;
	//! returns Portnumber
	int getPort() const;
	//! returns TransportProtocol
	Socket::Protocol getTransportProtocol() const;
	//! returns UserParam
	sipAddress::userParam getUserParam() const;
	//! returns MethodParam
	int getMethodParam() const;
	//! returns TtlParam
	int getTtlParam() const;
	//! returns MaddrParam
	char *getMaddrParam() const;
	//! returns Tag
        char *getTag();	

	//! returns the address as string
	char *toString() const;
};


//! Via: SIP-Tag
/*!
	\class sipTagVia sipTags.h
	\author Jan Burghard
	
	this represents the "Via"-headerfield of a sip-header
*/
class sipTagVia {
private:
	Socket::Protocol _protocol;
	char		*_ip;
	char		*_hostname;
	int		_port;
	bool		_paramHidden;
	char		_paramTtl;
	char		*_paramMaddr;
	sipAddress	*_paramReceived;
	char		*_paramBranch;
	
	//! initialization method
	void _init();
public:
	/*! 
	default constructor
	*/
	sipTagVia( const char *ip = "0.0.0.0" );
	
	/*! 
	copy constructor
	\param v another sipTagVia-object
	*/
	sipTagVia(const sipTagVia& v);
	
	/*!
	another usefull constructor
	\param p transport protocol
	\param ip ip-field
	\param port port-field
	*/
	sipTagVia(const Socket::Protocol p, const char *ip, const int port);
	
	/*!
	destructor
	*/
	virtual ~sipTagVia();
	
	/*!
	assignement operator
	\param via another sipTagVia-object
	
	copies all values of the given via-object to this
	*/
	sipTagVia& operator=(const sipTagVia& via);
	
	/*!
	duplication method
	*/
	virtual sipTagVia *dup() const { return new sipTagVia( *this ); };
		
	//! sets the transport-protocol-tag to p
	void setProtocol(const Socket::Protocol p);
	
	//! sets the ip-address to s
	void setIp(const char *s);
	
	//! sets the port-tag to i
	void setPort(const int i);
	
	//! sets the branch-parameter to s
	void setBranch(const char *s);
	
	//! sets received-tag
	void setReceived(const char *ip, const int port);
	
	//! returns the protocol-tag
	Socket::Protocol getProtocol() const;
	
	//! returns the ip-address
	char *getIp() const;
	
	//! returns the port-value
	int getPort() const;
	
	//! returns the branch-value
	char *getBranch() const;
	
	//! returns received-tag
	sipAddress *getReceived() const;
	
	//! returns the whole headerfield as string
	char *toString() const;
};


//! Route: SIP-Tag
/*!
	\class sipTagRoute sipTags.h
	\author Jan Burghard
	
	class that represents the "Route"-headerfield of a sip-header
*/
class sipTagRoute {
private:
	char		*_ip;
	char		*_hostname;
	int		_port;
public:
	//! default constructor
	sipTagRoute();
	
	//! another usefull constructor
	/*!
	\param ip ip-address for the tag-entry
	\param port portnumber of the tag-entry
	*/
	sipTagRoute(const char *ip, const int port);

	/*! 
	copy constructor
	\param r another sipTagRoute-object
	*/
	sipTagRoute(const sipTagRoute& r);

	//! default destructor
	virtual ~sipTagRoute();
	
	/*!
	assignement operator
	\param route another sipTagRoute-object
	
	copies all values of the given Route-object to this
	*/
	sipTagRoute& operator=(const sipTagRoute& route);
	
	//! set the ip-address
	void setIp(char *s);
	//! set the portnumber
	void setPort(int i);
	
	//! returns the ip-address
	char *getIp() const;
	//! returns the portnumber
	int getPort() const;
	
	//! returns the route-tag as string
	char *toString() const;
};


//! RecordRoute: SIP-Tag
/*!
	\class sipTagRecordRoute sipTags.h
	\author Jan Burghard
	
	class that represents the "Record-Route"-headerfield of a sip-header
*/
class sipTagRecordRoute {
private:
	char 	*_ip;
	char	*_hostname;
	int	_port;
	bool	_lrFlag;
public:
	//! default constructor
	sipTagRecordRoute();
	
	//! another usefull constructor
	sipTagRecordRoute(const char *ip, const int port);
	
	/*! 
	copy constructor
	\param r another sipTagRoute-object
	*/
	sipTagRecordRoute(const sipTagRecordRoute& r);

	//! default destructor
	virtual ~sipTagRecordRoute(); 
	
	/*!
	assignement operator
	\param rroute another sipTagRecordRoute-object
	
	copies all values of the given RecordRoute-object to this
	*/
	sipTagRecordRoute& operator=(const sipTagRecordRoute& rroute);
	
	//! set the ip-address
	void setIp(char *s);
	//! set the portnmber
	void setPort(int i);
	//! set the loose-routing flag (should be true)
	void setLr(bool b);
	
	//! returns the ip-address
	char *getIp() const;
	//! returns the portnumber
	int getPort() const;
	//! returns the loose-routing flag
	bool getLr() const;
	
	//! returns the Tag as string
	char *toString() const;
};



#endif
