/*!
	\file	sipLocationService.h
	\author	jAN bURGHARD
	\version 0.1
	
*/
#ifndef __SIPLOCATIONSERVICE_H
#define __SIPLOCATIONSERVICE_H

// for gcc3.x
using namespace std;

// C++ stuff
#include <vector>

// OMNeT++ stuff
#include <omnetpp.h>

// SIP stuff
#include "../Messages/sipResponse.h"
#include "../Messages/sipRequest.h"
#include "../Messages/sipTags.h"
#include "simpleRouting.h"



//! SIP Location Record
/*!
	\class sipLocationRecord sipLocationService.h
	\author Jan Burghard
	
	A Location Record stores map-addresses to an address of record.
*/
class sipLocationRecord : public cObject {	
protected:
	//! address that is to be mapped
	sipAddress *_addr_of_record;
	
	//! map-addresses
	vector<sipAddress *> _mapAddresses;
public:
	//! default constructor
	sipLocationRecord( const sipAddress *address );
	
	//! default desturctor
	virtual ~sipLocationRecord();
	
	//! adding a map-address for the address of record
	/*!
	\param address pointer to a sipAddress-object. This object will be duplicated!
	\return count of mapaddresses
	*/
	int addMapAddress( sipAddress *address );
	
	//! determine cout of mapaddresses
	unsigned int mapCount();
	
	/*!
	returns a map-address at "index". If no address is stored at "index"
	NULL will be returned.
	\param index position of address to be returned
	\return sipAddress
	*/
	sipAddress *mapAddress( unsigned int index );
	
	/*! \return -1 if address is not mapped - "index" otherwise          */
	int isMapped( sipAddress *address );

	/*! \return sipAddress - the address of record                       */
	sipAddress *addrOfRecord() const;

	/* ----------------------------------------------------------------------------
	   --- OMNeT++ specific functions
	   ----------------------------------------------------------------------------	*/
	//! returns class name string
	virtual const char *className() const { return "sipLocationRecord"; }
	//! duplicates the object by creating an exact copy (uses copy constructor)
	virtual sipLocationRecord *dup() const { return new sipLocationRecord(*this); }
	//! returns a one-line info about object contents
	virtual void info(char *buf);
	//! write a more detailed report about the object into a file
	virtual void writeContents(ostream& os);
	//! the assignement operator (copies object contents from another object)
	sipLocationRecord& operator=(const sipLocationRecord& r);
};






//! a simple sip location service
/*!
	\class sipLocationService sipLocationService.h
	\author Jan Burghard
	
	A simple local location service for the SIP.
*/
class sipLocationService : public cObject {

public:
	//! default constructor
	sipLocationService();
protected:
	/*! debugging on / off                                               */
	bool _debug;
	/*! pointer to sipLocationRecords                                    */
	vector<sipLocationRecord *> _entrys;	
	/*! pointer to routing module                                        */
	simpleRouting *_routingModule;
public:
	//! add a map address
	/*!
	add a mapaddress for an address
	\param address of record
	\param map address where the address of record shall be mapped to
	*/
	int addMapAddress(const sipAddress *address, const sipAddress *map);

	//! handles Register requests
	/*!
	processes a REGISTER request as described in RFC3261, section 10.3.
	Steps 1 to 4 are omitted!
	\return sipResponse with the encoded result
	*/
	sipResponse *handleRegister( sipRequest *request );

	//! get a sipLocationRecord
	/*!
	\param address - the address of record whose sipLocationRecord 
	  shall be returned
	\return sipLocationRecord
	*/
	sipLocationRecord *getRecord( const sipAddress *address );
	
	
	
	
	/* ----------------------------------------------------------------------------
	   --- OMNeT++ specific functions
	   ----------------------------------------------------------------------------	*/
	//! returns class name string
	virtual const char *className() const { return "sipLocationService"; }
	//! duplicates the object by creating an exact copy (uses copy constructor)
	virtual sipLocationService *dup() const { return new sipLocationService(*this); }
	//! returns a one-line info about object contents
	virtual void info(char *buf);
	//! write a more detailed report about the object into a file
	virtual void writeContents(ostream& os);
	//! the assignement operator (copies object contents from another object)
	sipLocationService& operator=(const sipLocationService& r);
};



#endif
