/*!
	\file	sipProxyLayer.cc
	\author	jAN bURGHARD
	\version 0.5
	
*/
#include "sipProxyLayer.h"

Define_Module( sipProxyLayer );


/* 
initializes std-vars
*/
void sipProxyLayer::initialize() {
	
	// Debug-mode on/off
	_debug = par("debug").boolValue();
	
	// local ip and port
	// the ip will be gotten from the RoutingTable module od
	// the networklayer. See "_handleSelfMessage".
	_localIp = IN_Addr("127.0.0.1");
	_serverPort = (int)par("serverPort");
	
	// network name
	_nodename = opp_strdup( par("nodeName").stringValue() );
	
	// timer C delay
	_timerCstart = strToSimtime("3min 1s");
	
	// branch index
	_branchCount = 0;
	
	// scheduling initialization message
	scheduleAt( simTime(), new cMessage("INITIALIZE") );
	
	// location service
	_locationService = new sipLocationService();

	// vectors for simulation results
	_incoming_v = new cOutVector( opp_concat("sipProxy-", _nodename, "-incoming") );
	_outgoing_v = new cOutVector( opp_concat("sipProxy-", _nodename, "-outgoing") );
};



/*
this method receives all events and forwards them to their
destinations
*/
void sipProxyLayer::handleMessage(cMessage *msg) {
	if ( _debug ) {
		ev << "received a '" << msg->className() <<
		      "' named '" << msg->name() << "'" << endl;
	}

	if ( msg->isSelfMessage() ) {
		_handleSelfMessage( msg );
	}
	
	else if ( msg->arrivalGateId() == findGate("fromServerModule") ) {
		_handleMessageFromServerModule( msg );
	} 
	
	else if ( msg->arrivalGateId() == findGate("fromClientModule") ) {
		_handleMessageFromClientModule( msg );
	} 
	
	else {
		opp_error("message from unknown gate");
	}
	
	delete msg;		
};



/*
used to free up allocated memory
*/
void sipProxyLayer::finish() {
	for(unsigned int i = 0; i < _processes.size(); i++) {
		_processes[i]->timerC = _stopTimer( _processes[i]->timerC );
		delete _processes[i];
	}

	if ( _debug ) {
		ev << "sipProxyLayer: " << _processes.size() 
		   << " processes deleted" << endl;
	}
	_processes.clear();
	
		
	// delete result-vectors
	delete _incoming_v;
	delete _outgoing_v;
}



/*

*/
cMessage *sipProxyLayer::_startTimer(char *name, simtime_t delay) {
	if ( delay > 0 ) {
		cMessage *timerX = new cMessage( name );
		scheduleAt( simTime() + delay, timerX );
	
		return timerX;
	} else {
		return NULL;
	}
};



/*

*/
cMessage *sipProxyLayer::_stopTimer(cMessage *timer) {
	if ( timer != NULL ) delete cancelEvent( timer );
	return NULL;
};




/*
this module handles self messages, e.g. the INITIALIZE message or timers
*/
void sipProxyLayer::_handleSelfMessage(cMessage *msg) {

	if ( opp_strcmp(msg->name(), "INITIALIZE") == 0 ) {

		// determine the local ip-address(es) from RoutingTable-Module
		RoutingTable *rt = (RoutingTable *)parentModule()
					->submodule("networkLayers")
					->submodule("routingTable");
		if ( rt == NULL ) {
			opp_error("could not find the RoutingTable!");
			return;
		}
		_localIp = IN_Addr( rt->getInterfaceByIndex(0)->inetAddr->getString() );
		if ( _debug ) myEv << "ip is set to: " << _localIp << endl;
	
		// 
		_routing.setLocalPort( _serverPort );
		_routing.addLocalIp( _localIp );

		// now we need the server socket
		sipInterfacePacket *sIfPkt = new sipInterfacePacket("GET_SOCKET");
		sIfPkt->openPort( _serverPort );
		_send( sIfPkt, "toServerModule" );
	}
	
	else if ( opp_strcmp(msg->name(), "timer C") == 0 ) {
		_handleTimerC( msg );
	}

};



/*
this module handles messages from the server module
*/
void sipProxyLayer::_handleMessageFromServerModule(cMessage *msg) {
	sipInterfacePacket *sIfPkt = (sipInterfacePacket *)msg;
	sipRequest *request;
	
	switch ( sIfPkt->action() ) {
	case sipInterfacePacket::SA_REQUEST:
		request = sIfPkt->getRequest();
		
		// store simtime of arrival
		_incoming_v->record( (double)request->getMethod() );

		// register requests have to be treated specially
		if ( request->getMethod() == REGISTER ) {
			_handleRegister( request );
		}
		else {
			_handleNewRequest( request );
		}
		break;
		
	default:
		ev << name() << ": got unknown IfPkt" << endl;
		break;
	}
};



/*
this module handles messages from the Client module
*/
void sipProxyLayer::_handleMessageFromClientModule(cMessage *msg) {
	sipInterfacePacket *sIfPkt = (sipInterfacePacket *)msg;
	sipResponse *response = NULL;
	
	switch ( sIfPkt->action() ) {
	case sipInterfacePacket::SA_RESPONSE:
		response = sIfPkt->getResponse();
		// store simtime of arrival
		_incoming_v->record( (double)response->getStatusCode() );

		// handle the response
		_handleResponse( response );
		break;
		
	case sipInterfacePacket::SA_TIMEOUT_REQUEST:
		_handleRequestTimeout( sIfPkt->getRequest() );
		break;
		
	default:
		ev << name() << ": got unknown IfPkt" << endl;
		break;
	}
};



/*
this module handles new income sipRequests
*/
void sipProxyLayer::_handleNewRequest(sipRequest *r) {

	// ------------------------------------------------
	// request validation
	// ------------------------------------------------
	if ( _debug ) myEv << "request validation" << endl;
	// 0. Version check
	if ( 0 != opp_strcmp(r->getVersion(), "2.0") ) {
		_sendError( r, 505, "Version Not Supported" );
		return;
	}
	
	
	// 1. Reasonable syntax check
	// 2. URI scheme check
	// 3. Max_Forwards
	if ( r->header()->getMaxForwards() == -1 ) {
		r->header()->setMaxForwards( 70 );
		if ( _debug ) myEv << "MaxForwards - headerfield missing!" << endl;
	}
	else if ( r->header()->getMaxForwards() == 0 ) {
		// too many hops error
		_sendError( r, 483, "Too Many Hops" );
		return;
	}
	// 4. Optional Loop Detection check
	// 5. Proxy-Require check
	// 6. Proxy-Authorization check

	proxySubProcess *p = new proxySubProcess( r );
	_processes.push_back( p );
	
	
	// ------------------------------------------------
	// Route Information Preprocessing
	// ------------------------------------------------
	if (_debug) myEv << "Route Information Preprocessing" << endl;
	// 1. maddr check

	// ------------------------------------------------
	// Determining Request Targets
	// ------------------------------------------------
	proxySubProcess::requestSet *rset;
	// 1. maddr check
	if ( false ) {
	
	}

	// 2. responsibility check for domain ...
	sipLocationRecord *slrec = _locationService->getRecord( r->getRequestURI() );
	if ( slrec != NULL ) {
		for(unsigned int i = 0; i < slrec->mapCount(); i++) {
			rset = new proxySubProcess::requestSet();
			rset->branch = NULL;
			rset->target = slrec->mapAddress( i )->dup();
			rset->request = NULL;
			rset->responseCode = -1;
			rset->response = NULL;
			p->addRequestSet( rset );	
		}
	}

	// request-URI is one target set
	rset = new proxySubProcess::requestSet();
	rset->branch = NULL;
	rset->target = p->getRequest()->getRequestURI()->dup();
	rset->request = NULL;
	rset->responseCode = -1;
	rset->response = NULL;
	p->addRequestSet( rset );
	
	// any targets found?
	if ( p->targetCount() < 1 ) {
		_sendError( r, 404, "Not Found" );
		
		// removing this subprocess
		delete p;
		_processes.pop_back();
		
		return;
	}
	

	// ------------------------------------------------
	// Request Forwarding
	// ------------------------------------------------
	if (_debug) myEv << "Request Forwarding" << endl;
	sipRequest *out;
	sipInterfacePacket *sIfPkt;
	char buf[255];	// for branch pararmeter generation

	for (int count = 0; count < p->targetCount(); count++ ) {
	
	// 1. Make a copy of the received request
	out = r->dup();
	
	// 2. Update the Request-URI
	out->setRequestURI( p->getRequestSet(count)->target );
	
	// 3. Update the Max-Forwards header field
	out->header()->setMaxForwards( out->header()->getMaxForwards() - 1 );
	
	// 4. Optionally add a Record-route header field value
	// 5. Optionally add additional header fields
	// 6. Postprocess routing information
	
	// 7. Determine the next-hop address, port, and transport
	out->setDestination( _routing.nextHopIp(p->getRequestSet(count)->target->getIp()),
	                     _routing.nextHopPort(p->getRequestSet(count)->target->getIp()) );
	if ( _debug ) myEv << "to-addr:'" 
		<< out->getDestIp() << "' to-port:'" 
		<< out->getDestPort() << "'" 
		<< endl;


	// -----------------------------------------------------
	// at this point we must generate a the branch parameter
	// -----------------------------------------------------
	sprintf( buf, "%d%s%d", count, _nodename, _branchCount++ );
	char *branch = opp_concat( SIP_MAGIC_COOKIE, buf );
	p->getRequestSet(count)->branch = opp_strdup( branch );
	if ( _debug ) myEv << "new branch is: '" << branch << "'" << endl;
	
	// 8. Add a Via header field value
	sipTagVia *via = new sipTagVia( _localIp );
	via->setBranch( branch );
	out->header()->addVia( via );
	
	// 9. Add a Content-Length header field if necessary
	
	//10. Forward the new request
	//    previously a copy of  the request will be stored to the 
	//    requestSet
	p->getRequestSet(count)->request = out->dup();
	sIfPkt = new sipInterfacePacket("REQUEST");
	sIfPkt->request( out );
	_send( sIfPkt, "toClientModule" );

	//11. Set timer C
	p->timerC = _startTimer( "timer C", _timerCstart );
	
	} // for
	
	if (_debug) myEv << "done ..." << endl;
};




/*
this module processes incoming responses
*/
void sipProxyLayer::_handleResponse(sipResponse *r) {

	if (_debug) {
		ev << "Got " << r->getStatusCode() << " Response." << endl;
	}
	
	proxySubProcess *p = NULL;
	char *branch;

	// ------------------------------------------------
	// Response Processing
	// ------------------------------------------------
	if ( !_responseSyntaxCheck( r ) ) {
		ev << "response did not pass syntax check - discarding it!" << endl;
		delete r;
		return;
	}
	
	branch = r->header()->getVia()->getBranch();


	// 1. Find the appropriate response context
	//    in our case we have to find the matching proxySubProcess
	for (unsigned int i = 0; i < _processes.size(); i++) {
		if ( _processes[i]->matchRequest( r ) >= 0 ) {
			p = _processes[i];
		}
	}
	if ( p == NULL ) {
		// no context found -> discard response
		delete r;
		return;
	}
	
	
	// 2. Update timer C for provisional responses
	//    ( for INVITE requests and 100 < status < 200 )
//	if ( (p->getRequest()->getMethod() == sipRequest::INVITE)
	if ( (p->getRequest()->getMethod() == INVITE)
	     && (r->getStatusCode() > 100) 
	     && (r->getStatusCode() < 200) ) 
	{
		// Reset timer C
		p->timerC = _stopTimer( p->timerC );
		p->timerC = _startTimer( "timer C", _timerCstart );
	}
	
	// 3. Remove the topmost Via
	r->header()->popVia();
	
	// for compatibility issues we have to store the localTag of the 
	// response - US's use it to match CANCEL-Requests to their 
	// dialogs
	if( r->getStatusCode() > 100 ) {
		sipRequest *request = p->getRequestSet(branch)->request;
		sipAddress to = request->header()->getTo();
		to.setTag( r->header()->getTo().getTag() );
		request->header()->setTo( to );
	}

	// Store response code
	if ( (int)r->getStatusCode() > p->getRequestSet(branch)->responseCode ) {
		p->getRequestSet(branch)->responseCode = r->getStatusCode();
	}
	
	// Store response itself if statuscode is greater than 199
	if ( r->getStatusCode() >= 200 ) {
		p->getRequestSet(branch)->response = r->dup();
	}
	
	// Foward response if (100 < statusCode < 300)
	if ( (100 < r->getStatusCode()) && (r->getStatusCode() < 300) ) {
		_forwardResponse( r, p );
	}
	
	// check if a final response has already been forwarded
	// if not -> check if all requests are answered finally
	// if yes -> choose best response and forward it
	if ( !p->isFinallyAnswered() ) {
	
		// check if all requests are answered with a final response
		int count = 0;
		for (int i = 0; i < p->targetCount(); i++) {
			if ( p->getRequestSet(i)->responseCode >= 200 ) {
				count++;
			}
		}
		if ( count == p->targetCount() ) {
			_forwardResponse( _chooseBestResponse(p), p );
		}
	}
	
	// if the responses statuscode is out of 2xx CANCEL other pending
	// requests
	if ( (200 <= r->getStatusCode()) &&
	     (r->getStatusCode() < 300)  &&
	     (!p->isCancellingRequests())
	     ) 
	{
		for (int i = 0; i < p->targetCount(); i++) {
			if ( (p->getRequestSet(i)->responseCode < 200) &&
			     (p->getRequestSet(i)->responseCode > 100) )
			{
				// this request has to be canceled
				sipRequest *cancel = new sipRequest();
				cancel->cancel( p->getRequestSet(i)->request );
				
				sipInterfacePacket *sIfPkt = 
				  new sipInterfacePacket("CANCEL REQUEST");
				sIfPkt->request( cancel );
				_send( sIfPkt, "toClientModule" );
			}
		}
	}
	
/*	// 4. Add the response to the response context (only for final)
	if ( (p->getRequestSet(branch)->responseCode > r->getStatusCode())
	     && (r->getStatusCode() >= 200) ) 
	{
		p->getRequestSet(branch)->responseCode = r->getStatusCode();
		p->getRequestSet(branch)->response = (sipResponse *)r->dup();
	}
	
	sipResponse *toForward = NULL;
	
	// 5. Check to see if this response should be forwarded immediately
	if ( (r->getStatusCode() > 100) && (r->getStatusCode() < 300) ) {
		sipResponse *toForward = r;
	}
	
	// 6. When necessary, choose the best final response from the
	//    response context
	// That is when all requests are answered
	if ( toForward == NULL ) {
		sipResponse *tmp;
		int count = 0;
		
		// check if all requests are answered with a final response
		for (int i = 0; i < p->targetCount(); i++) {
			tmp = p->getRequestSet(i)->response;
			if ( !((tmp == NULL) &&
			       (tmp->getStatusCode() > 200 )) ) {
//			       (tmp->getStatusCode() < 200 )) ) {
				count++;
			}
		}
		
		// "count" is the number of answered requests
		if ( count == p->targetCount() ) {
			toForward = _chooseBestResponse( p );
		}
	}
	
	// if toForward points to a response that has to be forwarded
	// if not, we can stop here
	if ( toForward == NULL ) { return; }
	
	// 7. 8. 9.
	_forwardResponse( toForward, p );
	
	//10. Generate any necessary CANCEL requests
	// this has to be done for all requestSets the already received a
	// provisional but NO final response!
*/
	// A copy has been made - the response can be deleted
	delete r;
};


/*
this function returns the best response of the given subProcess. 
It implements step 6 of RFC3261 chapter 16.7
*/
sipResponse *sipProxyLayer::_chooseBestResponse(proxySubProcess *proc) {
	// 6. When necessary, choose the best final response from the
	//    response context
	sipResponse *response = NULL;	
	int best = 999;
	sipResponse *tmp;
	
	for (int i = 0; i < proc->targetCount(); i++) {
	
		// a response has to exist and it's state-code has
		// to be less than "best"
		tmp = proc->getRequestSet(i)->response;
		if ( (tmp != NULL) && ((int)tmp->getStatusCode() < best) ) {

			// if it's provisional, it's only the "best" response
			// if no final response exists
			// (btw. provisionals should not be in requestSet)
			if ( (best == 999) || (tmp->getStatusCode() >= 200) ) {
				best = tmp->getStatusCode();
				response = tmp;
			}
		}
	}

	return response;
};

	
/*
this module forwards the best response of the given subProcess
*/
void sipProxyLayer::_forwardResponse(sipResponse *response, proxySubProcess *proc) {
	if ( (response == NULL) || (proc == NULL) ) {
		opp_error("::_forwardResponse - got NULL parameter!");
		return;
	}
	
	// 7. Aggregate authorization header field values if necessary
	
	// 8. Optionally rewrite Record-Route header field values
	
	// 9. Forward the response
	// we take the destination from the via-tag
//	sipTagVia *via = proc->getRequest()->header()->getVia();
//	response->setDestination( via->getIp(), via->getPort() );
	// we take the destination from the via-received-tag
	sipAddress *received = proc->getRequest()->header()->getVia()->getReceived();
	response->setDestination( received->getIp(), received->getPort() );

	sipInterfacePacket *sIfPkt = new sipInterfacePacket("UPSTREAM");
	sIfPkt->respond( (sipResponse *)response->dup() );
	_send( sIfPkt, "toServerModule" );
	
	// set the "_answered_final"-flag
	proc->setFinallyAnswered();
};


/*
this module processes the timer's subprocess
*/
void sipProxyLayer::_handleTimerC(cMessage *msg) {
	if ( _debug ) myEv << "timer C fired" << endl;
	
	// the timer may be restartet
	// but in our case we only send a 408 (request timeout)
	// to the client and CANCELs to all requestSets that 
	// already send a provisional answer
	
	// this timer fires if no final response has been received
	// and the last provisional response can in at
	// "simTime() - _timerCstart"
	
	// 1. find corresponding proxySubProcess
	proxySubProcess *proc = NULL;
	for (unsigned int i = 0; i < _processes.size(); i++) {
		if ( _processes[i]->timerC == msg ) {
			proc = _processes[i];
		}
	}
	if ( proc == NULL ) { 
		// something went wrong - ignoring it
		ev << "sipProxyLayer: timerC did not match any subProcess!" << endl;
		return; 
	}
	
	// 2. cancel all provisional responded requests
	for (int i = 0; i < proc->targetCount(); i++) {
		if ( (proc->getRequestSet(i)->responseCode < 200) &&
		     (proc->getRequestSet(i)->responseCode >= 100) )
		{
			// this request has to be canceled
			sipRequest *cancel = new sipRequest();
			cancel->cancel( proc->getRequestSet(i)->request );
			
			sipInterfacePacket *sIfPkt = 
			  new sipInterfacePacket("CANCEL REQUEST");
			sIfPkt->request( cancel );
			_send( sIfPkt, "toClientModule" );
		}
	}
	
	// 3. send timeout message
	_sendError( proc->getRequest(), 408, "request timeout" );
	
	// 4. remove timer from subprocess
	proc->timerC = NULL;
};


/*
this module calls the OMNeT++-send-method. Additionally it records 
the sending-process to the result-vector. 
*/
void sipProxyLayer::_send(sipInterfacePacket *sIfPkt, const char *gate) {
	// First we take a copy of the sIfPkt
	sipInterfacePacket *copy = sIfPkt->dup();
	
	// from the copy we generate the result-entry
	sipRequest *request = copy->getRequest();
	sipResponse *response = copy->getResponse();
	
	if( request != NULL ) {
		// sIfPkt contained a request => recording Method
		_outgoing_v->record( (double)request->getMethod() );
	}
	else if( response != NULL ) {
		// sIfPkt contained a response => recording statuscode
		_outgoing_v->record( (double)response->getStatusCode() );	
	}
	else {
		// sIfPkt contained noting => recording action of 
		// sipInterfacePacket
		_outgoing_v->record( (double)REC_SIPINTERFACE_ACTION 
		                             + copy->action() );	
	}
	
	// now the copys are useless
	delete copy;
	delete request;
	delete response;
	
	// the original packet will be sent out
	send( sIfPkt, gate );
};
	

/*
this module produces error responses
*/
void sipProxyLayer::_sendError(sipRequest *r, int errCode, char *errText) {
	sipResponse *err = new sipResponse( errCode, errText );
	
	// copy header
	err->header()->setViaList( r->header()->getViaList() );
	err->header()->setCSeq( r->header()->getCSeq() );
	err->header()->setFrom( r->header()->getFrom() );
	err->header()->setTo( r->header()->getTo() );
	
	// set Destination
	sipAddress *received = r->header()->getVia()->getReceived();
	err->setDestination( received->getIp(), received->getPort() );
		
	
	// encapsulate response and send it
	sipInterfacePacket *sIfPkt =
	  new sipInterfacePacket( opp_concat(err->name(), " error") );
	sIfPkt->respond( err );
	_send( sIfPkt, "toServerModule" );
};



/*

*/
bool sipProxyLayer::_responseSyntaxCheck(sipResponse *response) {
	bool result = true;
	
	// Via-Tag ok?
	if ( response->header()->getVia() == NULL ) { 
		result = false;
		ev << "responseSyntaxCheck: missing via-tag." << endl;
	}
	
	// XXX ok?
	if ( false ) {
		result = false;
	}
	
	return result;
};





/*
handles request timeouts
*/
void sipProxyLayer::_handleRequestTimeout(sipRequest *request) {
	// inform subProcess about final answer
	proxySubProcess *p = NULL;
	for (unsigned int i = 0; i < _processes.size(); i++) {
		if ( _processes[i]->isInBranchSet( request->header()->getVia()->getBranch() ) >= 0 ) {
			p = _processes[i];
		}
	}
	if ( (p != NULL) && (!p->isFinallyAnswered()) ) { 
		p->setFinallyAnswered();
		_sendError( p->getRequest(), 408, "request timeout" );
	}
};






/*
handles REGISTER requests
*/
void sipProxyLayer::_handleRegister(sipRequest *request) {
	sipResponse *response = _locationService->handleRegister( request );

	sipInterfacePacket *sIfPkt = new sipInterfacePacket("UPSTREAM");
	sIfPkt->respond( response );
	_send( sIfPkt, "toServerModule" );	

	delete request;
};





/* -----------------------------------------------------------------------------
 
 p r o x y S u b P r o c e s s

----------------------------------------------------------------------------- */

proxySubProcess::proxySubProcess(sipRequest *request) {
	_request = request;
	_answered_final = false;
	_cancelRequests = false;
};

proxySubProcess::~proxySubProcess() {
	for (int i = 0; i < targetCount(); i++) {
		delete getRequestSet(i)->branch;
		delete getRequestSet(i)->target;
		delete getRequestSet(i)->request;
		delete getRequestSet(i)->response;
	}

	delete _request;
};

void proxySubProcess::addRequestSet(requestSet *t) {
	_targets.push_back( t );
};

proxySubProcess::requestSet *proxySubProcess::getRequestSet(int i) {
	if ( (i >= 0) && (i < (int)_targets.size()) ) {
		return _targets[i];
	} else {
		return NULL;
	}
};

proxySubProcess::requestSet *proxySubProcess::getRequestSet(char *branch) {
	return getRequestSet( isInBranchSet(branch) );
};

int proxySubProcess::targetCount() {
	return _targets.size();
};

int proxySubProcess::isInBranchSet(const char *b) {
	for (int i = 0; i < (int)_targets.size(); i++) {
		if ( opp_strcmp(_targets[i]->branch, b) == 0 ) {
			return i;
		}
	}
	return -1;
};

int proxySubProcess::matchRequest(sipResponse *response) {
	for (int i = 0; i < (int)_targets.size(); i++) {

//		if ( (response->header()->getCSeq().method == _targets[i]->request->getMethod()) &&
		if ( (response->getMethod() == _targets[i]->request->getMethod()) &&
		     (0 == opp_strcmp(response->header()->getVia()->getBranch(), _targets[i]->branch)) ) {
			return i;		
		}
	}

	return -1;
};

sipRequest *proxySubProcess::getRequest() const {
	return _request;
};
