/*!
	\file	sipProxyLayer.h
	\author	jAN bURGHARD
	\version 0.5
	
*/
#ifndef __SIPPROXYLAYER_H
#define __SIPPROXYLAYER_H

// for gcc3.x
using namespace std;

// C++ stuff
#include <vector>

// OMNeT stuff
#include <omnetpp.h>

// IP-Suite
#include "sockets.h"
#include "RoutingTable.h"

// SIP-Stuff
#include "../Messages/sipTags.h"
#include "../Messages/sipRequest.h"
#include "../Messages/sipResponse.h"
#include "../Messages/sipConstants.h"
#include "../Messages/sipInterfacePacket.h"
#include "simpleRouting.h"
#include "sipLocationService.h"


//! SIP proxy subprocess
/*!
	\class proxySubProcess sipProxyLayer.h
	\author Jan Burghard
	
	represents a sip-proxys sub process that handles request forwarding
	and response receiving and processing.
*/
class proxySubProcess {

public:
	//! set of variables needed manage one sipRequest
	/*!
	to store all data needed for one request-target
	*/
	struct requestSet {
		/*! branch string                                             */
		char *branch;
		/*! request target                                            */
		sipAddress *target;
		/*! request sent to target (needed for CANCEL-generation)     */
		sipRequest *request;
		/*! response code                                             */
		int responseCode;
		/*! the response itself                                       */
		sipResponse *response;
	};

private:
	/*! the original received request                                     */
	sipRequest *_request;
	/*! vector for all handeled requestSets                               */
	vector<requestSet *> _targets;
	/*! will be set if a final answer is forwarded                        */
	bool _answered_final;
	/*! will be set if pending requests are about to be CANCELed          */
	bool _cancelRequests;

public:
	/*! process timer                                                     */
	cMessage *timerC;

	/*! default constructor                                               */
	proxySubProcess(sipRequest *request);
	/*! default destructor                                                */
	~proxySubProcess();
	
	/*! add a target-set                                                  */
	void addRequestSet(requestSet *t);
	/*! get target-set by index                                           */
	requestSet *getRequestSet(int i);
	/*! get target-set by branch-parameter                                */
	requestSet *getRequestSet(char *branch);
	/*! ammount of target-sets                                            */
	int targetCount();
	
	//!
	/*! checks if a target-set with the given branch exists
	\return index of matching target-set, if non was found "-1"
	*/
	int isInBranchSet(const char *b);
	
	//! return no. of matching requestSet
	/*! compares request-method and branch-parameter                      */
	int matchRequest(sipResponse *response);
	
	/*! returns the original request                                      */
	sipRequest *getRequest() const;

	//! sets the _answered_final - flag
	/*! to indicate that a final answer has been forwarded                */
	void setFinallyAnswered() { _answered_final = true; }
	
	/*! returns the _answered_final - flag                                */
	bool isFinallyAnswered() { return _answered_final; }

	//! sets the _cancelRequests - flag
	/*! to indicate that a final answer has been forwarded                */
	void setCancelRequests() { _cancelRequests = true; }
	
	/*! returns the _cancelRequests - flag                                */
	bool isCancellingRequests() { return _cancelRequests; }
};






/* ************************************************************************** */
/* ************************************************************************** */
/* ************************************************************************** */




//! SIP Proxylayer
/*!
	\class sipProxyLayer sipProxyLayer.h
	\author Jan Burghard
	
	a statefull SIP proxy
*/
class sipProxyLayer : public cSimpleModule {

public:
	// -------------------
	// inherited functions
	// -------------------
	
	//! OMNeT++ macro
	/*!
	will create constructors and OMNeT++ specific stuff.
	*/
	Module_Class_Members(sipProxyLayer, cSimpleModule, 0);

	//! inherited from OMNeT++
	/*!
	initializes the class.
	*/
	virtual void initialize();

	//! inherited from OMNeT++
	/*!
	handles incoming messages.
	*/
	virtual void handleMessage(cMessage *msg);
	
	//! inherited from OMNeT++
	/*!
	used to free up allocated memory.
	*/
	virtual void finish();
	
private:
	/*! debugmode                                                         */
	bool	_debug;
	/*! network-name of the proxy                                         */
	char	*_nodename;
	/*! the proxy's ip-address                                            */
	IN_Addr _localIp;
	/*! the proxy'S server-port                                           */
	IN_Port _serverPort;
	/*! pointer to the location service module                            */
	sipLocationService *_locationService;
	
	/*! Routing-Module that holds the local "addresses"                   */
	simpleRouting _routing;
	
	/*! collection of subprocesses                                        */
	vector<proxySubProcess *> _processes;
	
	/*! delay for transaction timer                                       */
	simtime_t _timerCstart;
	/*! transaction timer                                                 */
	cMessage *_timerC;
	
	/*! counter for "branch" calcualtion                                  */
	int _branchCount;

	/*! vector to record simulationtime of incoming messages              */
	cOutVector *_incoming_v;
	/*! vector to record simulationtime of outgoing messages              */
	cOutVector *_outgoing_v;

	// Timer methods	
	/*! \sa sipClientTransaction::_startTimer                             */
	cMessage *_startTimer(char *name, simtime_t delay);
	/*! \sa sipClientTransaction::_stopTimer                              */
	cMessage *_stopTimer(cMessage *timer);
	
	/*!
	this module handles self messages, e.g. the INITIALIZE message or timers
	*/
	void _handleSelfMessage(cMessage *msg);
	
	/*!
	this module handles messages from the server module
	*/
	void _handleMessageFromServerModule(cMessage *msg);

	/*!
	this module handles messages from the Client module
	*/
	void _handleMessageFromClientModule(cMessage *msg);
	
	/*!
	this module handles new income sipRequests
	*/
	void _handleNewRequest(sipRequest *r);

	/*!
	this module processes incoming responses.
	It implements steps 1 to 5 of RFC3261 chapter 16.7
	*/
	void _handleResponse(sipResponse *r);
	
	/*!
	this function returns the best response of the given subProcess. 
	It implements step 6 of RFC3261 chapter 16.7
	*/
	sipResponse *_chooseBestResponse(proxySubProcess *proc);
	
	/*!
	this module forwards the best response of the given subProcess. 
	It implements steps 7 to 9 of RFC3261 chapter 16.7
	\param response the response to be forwarded
	\param proc subprocess from which steps 7 and 8 take their data and
	       step 9 takes the destination address
	*/
	void _forwardResponse(sipResponse *response, proxySubProcess *proc);
	
	/*!
	this module processes the timer's subprocess
	*/
	void _handleTimerC(cMessage *msg);
	
	/*!
	this module calls the OMNeT++-send-method. Additionally it records 
	the sending-process to the result-vector. 
	*/
	void _send(sipInterfacePacket *sIfPkt, const char *gate);
	
	/*!
	this module produces error responses
	*/
	void _sendError(sipRequest *r, int errCode, char *errText);
	
	/*!
	checks a response's syntax
	*/
	bool _responseSyntaxCheck(sipResponse *response);

	/*!
	handles request timeouts
	*/
	void _handleRequestTimeout(sipRequest *request);
	
	/*!
	handles REGISTER requests
	*/
	void _handleRegister(sipRequest *request);
	
};


#endif
